/*************************************************************************/ /*!
@File
@Title          fp_switch_host_ageing.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    implements switch ageing functionality
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/
 #include <fp_library.h>
 #include <fp_switch.h>
 #include <hw_2field_hash_table.h>
 #include <linux/spinlock.h>

/* Delete secondary vlan entries corresponding to primary vlan */
void hw_2field_pvlan_entries_delete( uint8_t *mac, uint16_t vlanId, uint32_t action_entry ){
	uint32_t i = 0, pvid = 0, fwd_list = 0, svid = 0;
    int32_t portNo = -1;
    struct port_s *port = NULL;

	pvid = vlanId;

	fwd_list = action_entry & BRENTRY_FWD_PORT_LIST_MASK;

	for( i = 0; i < FP_MAX_SWITCH_PORTS; i++ ){
		if( ( FWD_PORT_LIST_MASK >> i ) & 0x1 ){
			if( ( fwd_list >> i ) & 0x1 ){
				portNo = i;
				break;
			}
		}
	}

	if( portNo == -1 ){
	   return;
	}

	port = &fp_port[ portNo ];

	/* Return if port primary vlan is not same as entry vlan */
	if( !port->pvlan_enable ){
		return;
	}

	if( port->pvlan_pri != pvid ){
		return;
	}

	/* Delete all the secondary vlan entries corresponding to primary vlan */
	switch( port->pvlan_type ){
		case ISOLATED:
		case COMMUNITY:
			svid = port->pvlan_sec[ 0 ];
			hw_hash_table_delete( mac, svid );
			fp_rc_stats.hw_hash_entry_count--;
	    	break;

		case PROMISCUOUS:
		case INTERSWITCH:
			for( i = 0; i < port->num_sec; i++ ){
				svid = port->pvlan_sec[ i ];
				hw_hash_table_delete( mac, svid );
				fp_rc_stats.hw_hash_entry_count--;
			}
	    	break;
		default:
			return;
	}
}

/**********************************************************************
 *                              Extern Declarations
***********************************************************************/

/********************************************************************
 * Function Name  : fp_switch_host_ageing
 * Description    : handles hw mac table update for ageing.
 *                  read each entry from hash table,
			delete hash entry if fresh bit is set(1)
			change fresh bit to set(1) if it is unset(0)
 * Inputs         :
 * Parameters     : void
 * Outputs        :
 * Parameters     : -
 * Returns        : void
 ********************************************************************/
void fp_switch_host_ageing( void ){
    uint32_t data[ 4 ], hashidx, coll_ptr, pvlan_prim = 0;
    uint16_t vlanid;
    uint8_t mac[ 6 ];

    static uint32_t hashBase;

	if (fp_baseAddr == FP_NULL){
		FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_EMERG, "pci base address null\n" );
		return;
	}

    /* 2-field Mac table format: Table Depth - 1024(hash 512 + collison 512)
    *============================================================================*
    * valid bits | col_ptr | portNO | field_valids | action entry | vlanid|  MAC *
    *============================================================================*
    * 4-bit      | 16-bit  | 4-bit  | 8-bit(mac4)  | 31-bit (mac3)| 13-bit|48-bit*
    *============================================================================*/

    /*=============================================================================*
    * MAC1:MAC[31:0]                                                               *
    *==============================================================================*
    * MAC2:ACT_ENTRY[31:29] + VLANID[28:16] + MAC[15:0]                            *
    *==============================================================================*
    * MAC3:FIELD_VALIDS[31:28] + ACT_ENTRY[27:0]                                   *
    *==============================================================================*
    *MAC4:RSVD[31:28]+VALIDBITS[27:24]+COLPTR[23:8]+ PORTNO[7:4] + FIELDVALIDS[3:0]*
    *=============================================================================*/

	FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_DEBUG, "HASH UPDATE is Called\r\n" );

    for( hashidx = 0; hashidx <  HASH_CNT_PER_ITERATION; hashidx++ ){
start:
        coll_ptr = 0;
        hw_hash_table_read_entry( data, ( hashBase + hashidx ) );

		if( data[ REG_MAC4 ] ){
			FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "mac4:%x\r\n", data[ REG_MAC4 ] );
		}

        if( data[ REG_MAC4 ] & HASH_ENTRY_VALID_FLAG ){                                                       /* If entry is valid */
			FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "HASH idx:%x\r\n", ( hashBase + hashidx ) );
			if ( data[ REG_MAC4 ] & HASH_ENTRY_COLL_FLAG ){
			    coll_ptr = ( data[ REG_MAC4 ] >> 8 ) & 0xFFFF;
			}

			FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "mac3:%x\r\n", data[ REG_MAC3 ] );

            if( data[ REG_MAC3 ] & HASH_ENTRY_FRESH_FLAG ){
                /* FRESH bit set, so delete entry */
				if( data[ REG_MAC3 ] & HASH_ENTRY_PVLAN_FLAG ){
					pvlan_prim = 1;
				}

                FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "HASH Del:%x data[0] = %x, data[1] = %x\r\n", ( hashBase + hashidx ), data[ REG_MAC1 ], data[ REG_MAC2 ] );
                mac[ 0 ] = ( data[ REG_MAC1 ] >> 24 ) & 0xFF;
                mac[ 1 ] = ( data[ REG_MAC1 ] >> 16 ) & 0xFF;
                mac[ 2 ] = ( data[ REG_MAC1 ] >>  8 ) & 0xFF;
                mac[ 3 ] = ( data[ REG_MAC1 ] >>  0 ) & 0xFF;
                mac[ 4 ] = ( data[ REG_MAC2 ] >>  8 ) & 0xFF;
                mac[ 5 ] = ( data[ REG_MAC2 ] >>  0 ) & 0xFF;
                vlanid   = ( data[ REG_MAC2 ] >> 16 ) & 0xFFF;

                hw_hash_table_delete( mac, vlanid );
                fp_rc_stats.hw_hash_entry_count--;

                if( coll_ptr ){
					if( pvlan_prim ){
						hw_2field_pvlan_entries_delete( mac, vlanid, ( ( data[ REG_MAC3 ] << 4 ) | ( data[ REG_MAC2 ] >> 29 ) ) );
						pvlan_prim = 0;
					}
					goto start;
				}
			}else{          /* Set FRESH bit, if entry isn't static */
				if( !( data[ REG_MAC3 ] & HASH_ENTRY_STATIC_FLAG ) ){
				    data[ REG_MAC3 ] |= HASH_ENTRY_FRESH_FLAG;
                    hw_hash_table_write_entry( data, ( hashBase + hashidx ) );
				    FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "HASH Update:%x\r\n", ( hashBase + hashidx ) );
				}else{
					continue;
				}
			}
            if( pvlan_prim ){
				hw_2field_pvlan_entries_delete( mac, vlanid, ( ( data[ REG_MAC3 ] << 4 ) | ( data[ REG_MAC2 ] >> 29 ) ) );
				pvlan_prim = 0;
			}
		}

        while( coll_ptr ){                                                                            /* collision space entries */
            hw_hash_table_read_entry( data, coll_ptr );
			if( data[ REG_MAC4 ] & HASH_ENTRY_VALID_FLAG ){                                                    /* check entry's validity */
                if( data[ REG_MAC3 ] & HASH_ENTRY_FRESH_FLAG ){                                                /* FRESH bit set, so delete entry */
                
                    if( HASH_ENTRY_PVLAN_FLAG & data[ REG_MAC3 ] ){
						pvlan_prim = 1;
                    }

                    mac[0] = ( data[ REG_MAC1 ] >> 24 ) & 0xFF;
                    mac[1] = ( data[ REG_MAC1 ] >> 16 ) & 0xFF;
                    mac[2] = ( data[ REG_MAC1 ] >>  8 ) & 0xFF;
                    mac[3] = ( data[ REG_MAC1 ] >>  0 ) & 0xFF;
                    mac[4] = ( data[ REG_MAC2 ] >>  8 ) & 0xFF;
                    mac[5] = ( data[ REG_MAC2 ] >>  0 ) & 0xFF;
                    vlanid = ( data[ REG_MAC2 ] >> 16 ) & 0xFFFF;

                    hw_hash_table_delete( mac, vlanid );

					fp_rc_stats.hw_hash_entry_count--;
					FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "Coll HASH Del:%x\r\n", coll_ptr );
				}else{                                        /* Set FRESH bit, if entry isn't static */
					if( !( data[ REG_MAC3 ] & HASH_ENTRY_STATIC_FLAG ) ){
					    data[ REG_MAC3 ] |= HASH_ENTRY_FRESH_FLAG;
                        hw_hash_table_write_entry( data, coll_ptr );
					    FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO,	"COLL HASH UPDATE:%x\r\n", coll_ptr );
					}else{
						coll_ptr = 0;
						if( data[ REG_MAC4 ] & HASH_ENTRY_COLL_FLAG ){
							coll_ptr = ( data[ REG_MAC4 ] >> 8 ) & 0xFFFF;
						}
						continue;
					}
				}

				coll_ptr = 0;
				if( data[ REG_MAC4 ] & HASH_ENTRY_COLL_FLAG ){
					coll_ptr = ( data[ REG_MAC4 ] >> 8 ) & 0xFFFF;
				}

			    if ( pvlan_prim ){
				    hw_2field_pvlan_entries_delete( mac, vlanid, ( ( data[ REG_MAC3 ] << 4 ) | ( data[ REG_MAC2 ] >> 29 ) ) );
				    pvlan_prim = 0;
			    }
            }else{
                coll_ptr = 0;
                FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_EMERG, "Hash entry at collision pointer invalid!\n" );
            }
		}
	}

	/*update hashBase */
	hashBase += hashidx;

	/* updating entries in blocks */
	if ( hashBase >= 4095 ){
		hashBase = 0;
	}
}
