/*************************************************************************/ /*!
@File
@Title          fp_switch_init.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    Initializes port, global & hwmac table paramaters
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/

#include <linux/string.h>
#include <linux/slab.h>
#include <hw_vlan_hash_table.h>
#include "../include/fp_global_addr_map.h"
#include "fp_mac_ops.h"
#include "sfp_copper_setup.h"

/**********************************************************************
 *                                      #defines
 **********************************************************************/


/**********************************************************************
 *                              Global Declarations
***********************************************************************/
struct port_s fp_port[FP_MAX_SWITCH_PORTS];
struct global_s fp_global;
struct rc_stats_s fp_rc_stats;
/**********************************************************************
 *                              Extern Declarations
***********************************************************************/

/******************************************************************n
 * Function Name  : fp_hw_sw_init
 * Description    : initializes switch hw & mac table elements
 * Inputs         :
 * Parameters     :
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
uint32_t fp_hw_sw_init(void)
{
	uint32_t i, addr1 = 0, addr2 = 0, tmp;
	uint64_t gBrEntry = 0;

	if (fp_baseAddr == NULL) {
		FP_DEBUG(DBGP_FEAT_BRIDGE, FP_LOG_EMERG, "%s: pci_mem base is NULL\n", __func__);
		return 0;
	}

	memset(fp_port, 0, ( sizeof(struct port_s) * FP_MAX_SWITCH_PORTS ) );

	for( i = 0; i < NUM_OF_OPTIC_PORTS; i++ ){
		//fp_port[ i ].optic_link.module_descr = FP_NULL;
		fp_port[ i ].optic_link.module_descr = kmalloc( SIZE_OF_SFP_READ_DATA_IN_BYTES, GFP_KERNEL );
		if( fp_port[ i ].optic_link.module_descr == FP_NULL ){
			FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_EMERG, "%s: can not allocate memory for optic port %d\n", __func__, i );
		}else{
			memset( fp_port[ i ].optic_link.module_descr, 0, SIZE_OF_SFP_READ_DATA_IN_BYTES );
		}
	}

	/* init port data structure */
	for (i = 0; i < FP_MAX_SWITCH_PORTS; i++)
    {
		addr1 = addr2 = 0;
		if ((FWD_PORT_LIST_MASK >> i) & 0x1)
        {
#ifndef HW_SWITCH_SOC
			fp_port[i].shutdown              = 0;
			fp_port[i].tpid                  = VLAN_TAGGED_FRAME;
			fp_port[i].aft                   = ANY_TAGGING;
			fp_port[i].fallback_bd_id        = DEF_PORT_FALLBACK_BD_ID;
			fp_port[i].block_state           = FORWARDING;
			fp_port[i].csr_untag_from_btable = 0;

#ifdef HW_SW_FA_EN
			/* HOST_PORT of switch should always has to send
				tagged frames to FA */
			if (i == HOST_PORT_NO)
				fp_port[i].aft              = TAGGED_ONLY;
#endif                                                                          /* HW_SW_FA_EN */
#endif                                                                          /* !HW_SWITCH_SOC */
			fp_get_port_addr(0, i, &addr1, &addr2);

			/* class hw port csr reg:
			*======================================================*
			*res |block_state |aft |shutdown |fallback_bd_id |tpid *
			*======================================================*
			*20-bit| 4-bit   |4-bit| 1-bit   | 16-bit      |16-bit *
			*=====================================================*/
			if (addr1 && addr2)
            {
				fp_reg_write( (uint32_t)addr1, 
                    (uint32_t)( ((fp_port[i].fallback_bd_id << PORT_FALLBACK_BDID_START_POS) & PORT_FALLBACK_BDID_MASK)|
                    ((fp_port[i].tpid << PORT_TPID_START_POS) & PORT_TPID_MASK)) );

				tmp = fp_reg_read( (uint32_t)addr1);
				//FP_DEBUG(DBGP_FEAT_BRIDGE, FP_LOG_EMERG,"%s: addr 0x%x, write value 0x%08X\n", __func__, addr1, tmp);

				fp_reg_write( (uint32_t)addr2,
                    (uint32_t)( ((fp_port[i].csr_untag_from_btable << PORT_UNTAG_FROM_BTABLE_START_POS) & PORT_UNTAG_FROM_BTABLE_MASK) |
                    ((fp_port[i].block_state << PORT_BLOCKSTATE_START_POS) & PORT_BLOCKSTATE_MASK) |
                    ((fp_port[i].aft <<  PORT_AFT_START_POS) & PORT_AFT_MASK) |
                    ((fp_port[i].shutdown << PORT_SHUTDOWN_START_POS) & PORT_SHUTDOWN_MASK) ) );

				tmp = fp_reg_read( (uint32_t)addr2);
				//FP_DEBUG(DBGP_FEAT_BRIDGE, FP_LOG_EMERG,"%s: addr 0x%x, write value 0x%08X\n", __func__, addr2, tmp);

			}
			fp_get_port_addr(1, i, &addr1, &addr2);

			/* class hw port csr reg:
			*======================================================*
			*res |block_state |aft |shutdown |fallback_bd_id |tpid *
			*======================================================*
			*20-bit| 4-bit   |4-bit| 1-bit   | 16-bit      |16-bit *
			*=====================================================*/
			if (addr1 && addr2)
            {
				fp_reg_write( (uint32_t)addr1,
                    (uint32_t)( ((fp_port[i].fallback_bd_id << PORT_FALLBACK_BDID_START_POS) &
                    PORT_FALLBACK_BDID_MASK) | ((fp_port[i].tpid << PORT_TPID_START_POS) & PORT_TPID_MASK)) );

				tmp = fp_reg_read( (uint32_t)addr1);
				FP_DEBUG(DBGP_FEAT_BRIDGE, FP_LOG_EMERG,"%s: addr 0x%x, write value 0x%08X\n", __func__, addr1, tmp);

				fp_reg_write( (uint32_t)addr2,
					(uint32_t)( ((fp_port[i].csr_untag_from_btable << PORT_UNTAG_FROM_BTABLE_START_POS) & PORT_UNTAG_FROM_BTABLE_MASK) |
					((fp_port[i].block_state << PORT_BLOCKSTATE_START_POS) & PORT_BLOCKSTATE_MASK) |
					((fp_port[i].aft << PORT_AFT_START_POS) & PORT_AFT_MASK) |
					((fp_port[i].shutdown << PORT_SHUTDOWN_START_POS) & PORT_SHUTDOWN_MASK)) );

				tmp = fp_reg_read( (uint32_t)addr2);
				FP_DEBUG(DBGP_FEAT_BRIDGE, FP_LOG_EMERG,"%s: addr 0x%x, write value 0x%08X\n", __func__, addr2, tmp);
			}
		}
	}

	/* init fp_global fallback bd entry  */
//#ifndef HW_SWITCH_SOC
	fp_global.fallback_bd_entry.ucast_hit_action = ACT_FORWARD;
	fp_global.fallback_bd_entry.ucast_miss_action = ACT_FORWARD;
	fp_global.fallback_bd_entry.mcast_hit_action = ACT_FORWARD;
	fp_global.fallback_bd_entry.mcast_miss_action = ACT_FORWARD;
//	fp_global.fallback_bd_entry.forward_list = GLOBAL_BR_ENTRY_PORTLIST;
	fp_global.fallback_bd_entry.forward_list = 0x10000;
	fp_global.fallback_bd_entry.untag_list = GLOBAL_BR_ENTRY_UNTAGLIST;
	fp_global.fallback_bd_entry.mstp_action = ACT_FORWARD;
	fp_global.l2_special_punt_enable = 1;
//#endif /* !HW_SWITCH_SOC */

/* GLOBAL ENTRY:
*========================================================================*
* fallback_bd_entry                            |  l2_special_punt_enable *
*========================================================================*
* 31-bit                                       |  1-bit                  *
*========================================================================*/
	gBrEntry = ( (((uint64_t)fp_global.fallback_bd_entry.ucast_hit_action << BRENTRY_UCAST_HIT_ACT_START_POS) & BRENTRY_UCAST_HIT_ACT_MASK) |
				 (((uint64_t)fp_global.fallback_bd_entry.ucast_miss_action << BRENTRY_UCAST_MISS_ACT_START_POS) & BRENTRY_UCAST_MISS_ACT_MASK) |
				 (((uint64_t)fp_global.fallback_bd_entry.mcast_hit_action << BRENTRY_MCAST_HIT_ACT_START_POS) & BRENTRY_MCAST_HIT_ACT_MASK) |
				 (((uint64_t)fp_global.fallback_bd_entry.mcast_miss_action << BRENTRY_MCAST_MISS_ACT_START_POS) & BRENTRY_MCAST_MISS_ACT_MASK)  |
				 (((uint64_t)fp_global.fallback_bd_entry.untag_list << BRENTRY_UNTAG_LIST_START_POS) & BRENTRY_UNTAG_LIST_MASK) |
				 ((fp_global.fallback_bd_entry.forward_list) & BRENTRY_FWD_PORT_LIST_MASK) );

	gBrEntry = ((gBrEntry << 1) | (fp_global.l2_special_punt_enable & 0x1));

    fp_reg_write( (CLASS_HW_BASE_ADDR + CLASS_GLOBAL_CFG), (gBrEntry & 0xFFFFFFFF) );
    fp_reg_write( (CLASS_HW_BASE_ADDR + CLASS_GLOBAL_CFG1), ((gBrEntry >> 32) & 0xFFFFFFFF) );

	fp_reg_write( (CLASS2_HW_BASE_ADDR + CLASS_GLOBAL_CFG), (gBrEntry & 0xFFFFFFFF) );
    fp_reg_write( (CLASS2_HW_BASE_ADDR + CLASS_GLOBAL_CFG1), ((gBrEntry >> 32) & 0xFFFFFFFF) );

	/* Configure STP MAC details */
	fp_reg_write( (uint32_t)CLASS_SNOOP_CTRL, STP_CTRL_REG_VAL);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_ADDR1_LSB, STP_MAC_ADDR1_LSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_ADDR1_MSB, STP_MAC_ADDR1_MSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_ADDR2_LSB, STP_MAC_ADDR2_LSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_ADDR2_MSB, STP_MAC_ADDR1_MSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_MASK1_LSB, STP_MAC_MASK1_LSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_MASK1_MSB, STP_MAC_MASK1_MSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_MASK2_LSB, STP_MAC_MASK2_LSB);
	fp_reg_write( (uint32_t)CLASS_SNOOP_SPL_MCAST_MASK2_MSB, STP_MAC_MASK2_MSB);

/* Configure STP MAC details */
	fp_reg_write( (uint32_t)CLASS2_SNOOP_CTRL, STP_CTRL_REG_VAL);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_ADDR1_LSB, STP_MAC_ADDR1_LSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_ADDR1_MSB, STP_MAC_ADDR1_MSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_ADDR2_LSB, STP_MAC_ADDR2_LSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_ADDR2_MSB, STP_MAC_ADDR1_MSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_MASK1_LSB, STP_MAC_MASK1_LSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_MASK1_MSB, STP_MAC_MASK1_MSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_MASK2_LSB, STP_MAC_MASK2_LSB);
	fp_reg_write( (uint32_t)CLASS2_SNOOP_SPL_MCAST_MASK2_MSB, STP_MAC_MASK2_MSB);

	/* reset rc codes stats for cli */
	fp_rc_stats.hw_hash_entry_count = 0;
	fp_rc_stats.l2_spl_cnt = 0;
	fp_rc_stats.sa_miss_cnt = 0;
	fp_rc_stats.sa_relearn_cnt = 0;
	fp_rc_stats.sa_is_active_cnt = 0;
	fp_rc_stats.snoop_upper_cnt = 0;
	fp_rc_stats.requested_cnt = 0;
	fp_rc_stats.mgmt_cnt = 0;
	fp_rc_stats.igmp_cnt = 0;
	fp_rc_stats.flood_cnt = 0;
	fp_rc_stats.parse_cnt = 0;
	fp_rc_stats.fwd_cnt = 0;
	fp_rc_stats.drop_cnt = 0;

	fp_reg_write( (uint32_t)CLASS_HW_GLOBAL_CUTTHRU_REG, CUT_THRU_PORTLIST | (STPID1 << STPID1_START_POS));
	fp_reg_write( (uint32_t)CLASS2_HW_GLOBAL_CUTTHRU_REG, CUT_THRU_PORTLIST | (STPID1 << STPID1_START_POS));

	return 0;
}
/******************************************************************
 * Function Name  : fp_hw_vlantable_add
 * Description    : Add default vlan table entry
 * Inputs         :
 * Parameters     :
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_hw_vlantable_add(void)
{
	uint64_t action_entry = 0;

	if (fp_baseAddr == NULL) {
		FP_DEBUG(DBGP_FEAT_BRIDGE, FP_LOG_EMERG, "%s: pci_mem base is NULL\n", __func__);
		return 0;
	}
	action_entry = FWD_PORT_LIST_MASK;
	//action_entry = 0x10000;

	return wsp_vlan_entry_add(1, action_entry);
}


void fp_get_port_addr(uint32_t class_num, uint32_t i, uint32_t *addr1, uint32_t *addr2)
{
    if( ( class_num < FP_CLASS_MAX_NUMBER ) && ( i < FP_MAX_SWITCH_PORTS ) )
    {
        *addr1 = hwPortStruct1[class_num][i];
        *addr2 = hwPortStruct2[class_num][i];
    }
	else
    {
		*addr1 = 0;
		*addr2 = 0;
	}
}

void sfp_disconnected( uint32_t port_num ){
	if( port_num < NUM_OF_OPTIC_PORTS ){
		if( fp_port[ port_num ].optic_link.module_descr ){
			memset( fp_port[ port_num ].optic_link.module_descr, 0, SIZE_OF_SFP_READ_DATA_IN_BYTES );
		}
		fp_port[ port_num ].optic_link.rate = 0;
		fp_port[ port_num ].optic_link.status = 0;
		fp_port[ port_num ].optic_link.pres = 0;
		//FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_DEBUG, "%s: port_%d disconnected\n", __func__, port_num );
	}
}

uint32_t is_sfp_module_connected( uint32_t port_num ){
	if( fp_port[ port_num ].optic_link.pres ){
		return 1;	//sfp module connected to the slot (port)
	}else{
		return 0;	//sfp module disconnected at the slot (port)
	}
}

uint32_t sfp_store_data( uint32_t port_num, uint32_t addr_offset, uint32_t bytes_count, uint32_t *from ){
	uint32_t i;
	
	if( port_num < NUM_OF_OPTIC_PORTS ){
		if( fp_port[ port_num ].optic_link.module_descr ){
			if( ( addr_offset + bytes_count ) <= SIZE_OF_SFP_READ_DATA_IN_BYTES ){
				for( i = 0; i < bytes_count; i++ ){
					fp_port[ port_num ].optic_link.module_descr[ addr_offset + i ] = ( from[ i >> 2 ] >> ( 24 - ( ( i << 3 ) & 0x1F ) ) );
					//FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_DEBUG, "%s: address %d, value 0x%02X\n", __func__, ( addr_offset + i ), ( ( from[ i >> 2 ] >> ( 24 - ( ( i << 3 ) & 0x1F ) ) ) & 0xFF ) );
				}
				return 0;
			}
		}
	}
	return 1;	//error
}

uint32_t sfp_module_connected( uint32_t port_num ){
	if( port_num < NUM_OF_OPTIC_PORTS ){
		if( fp_port[ port_num ].optic_link.module_descr[ 18 ] | ( fp_port[ port_num ].optic_link.module_descr[ 2 ] == 0x21 ) ){
			fp_port[ port_num ].optic_link.pres = COPPER_SFP_MODULE;
			FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "%s: Try to start thread for port_%d\n", __func__, port_num );
			fp_sfp_copper_setup_job_start( ( ( SFP_MODULE_I2C_EPHY_ADDR << 8 )|( port_num ) ) );   //try to setup copper sfp module
		}else{
			fp_port[ port_num ].optic_link.pres = OPTIC_SFP_MODULE;
		}			
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: port_%d connected %s module\n", __func__, port_num, ( fp_port[ port_num ].optic_link.pres == OPTIC_SFP_MODULE )?"OPTIC":"COPPER" );
		fp_port[ port_num ].optic_link.rate = 1000;
		fp_port[ port_num ].optic_link.status |= OPTIC_STATUS_SFP_DATA_READY;
		restart_sgmii_autoneg( port_num );
		return 0;
	}
	return 1;	//error
}

uint32_t sfp_copper_configured( uint32_t port_num ){
	fp_port[ port_num ].optic_link.status |= OPTIC_STATUS_COP_SFP_READY;
	return 0;
}
