/*************************************************************************/ /*!
@File
@Title          fp_bd_api.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    implements pci bus mastering dma operations
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#include<linux/version.h>
#include<linux/pci.h>
#include<linux/interrupt.h>
#include<linux/netdevice.h>

#include <fp_library.h>
#include "fp_net_driver.h"
#include "fp_hif_reg.h"
#include "fp_bd_api.h"

/* holds the total ring size of the entire dma alloc */
UINT  fp_total_ring_size;

/***************************************************************************
 * Function Name  : fp_bd_create_ring
 * Description    : allocate memory for tx/rx/write_bk bd's and tx/rx pktbuf
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 ***************************************************************************/
int fp_bd_create_ring(fp_private_t *fp)
{
	char buff[256];
	UINT *dev_addr = (UINT *)&buff[0];
	void *cpu_addr;
	struct bd_ring *txring, *rxring;
	UINT ch_index;

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
			fp->hif.ch[ch_index].txring =
				(struct bd_ring *)
				kmalloc(sizeof(struct bd_ring), GFP_ATOMIC);
			fp->hif.ch[ch_index].rxring =
				(struct bd_ring *)
				kmalloc(sizeof(struct bd_ring), GFP_ATOMIC);
			if ((fp->hif.ch[ch_index].txring == NULL) ||
			(fp->hif.ch[ch_index].rxring == NULL)) {
				FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG,
						"error kmalloc: rx/tx ring\n");
				FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG,
					"ERROR at ring allocation "
					"ch_index %u\n", ch_index);
				return -1;
			}

			txring = fp->hif.ch[ch_index].txring;
			rxring = fp->hif.ch[ch_index].rxring;

			memset(txring, 0, sizeof(struct bd_ring));
			memset(rxring, 0, sizeof(struct bd_ring));

			/* Size required to be allocated for
					TX & RX BDs, TX & RX buffers */
			txring->size =
				(NUM_TX_WB_DESCR * sizeof(struct wb_bd)) +
				(NUM_TX_DESCR * sizeof(struct bd)) +
				(NUM_TX_DESCR * MAX_BUFF_SIZE);

			rxring->size =
				(NUM_RX_WB_DESCR * sizeof(struct wb_bd)) +
				(NUM_RX_DESCR * sizeof(struct bd)) +
				(NUM_RX_DESCR * MAX_BUFF_SIZE);

			fp_total_ring_size = txring->size + rxring->size +
							(6 * ALIGN_64_BIT_MASK);

			/* The kernel allocates required memory
							for given size */
			cpu_addr = dma_alloc_coherent(&fp->pci_dev->dev,
				fp_total_ring_size, (dma_addr_t *)dev_addr,
				GFP_ATOMIC);
			if (cpu_addr == NULL) {
				FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG,
					"dma_alloc_coherent failed\n");
				FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG,
					"ERROR at DMA allocation"
					"ch_index %u\n", ch_index);
				return -1;
			}

			txring->dev_addr = rxring->dev_addr = *dev_addr;
			txring->cpu_addr = rxring->cpu_addr = cpu_addr;

			/*****************************************************
			*  DMA Memory Mapping Order:                         *
			*   TX WR BACK BD Desc                               *
			*   RX WR BACK BD DESC                               *
			*   TX BD DESC                                       *
			*   RX BD DESC                                       *
			*   TX PKT BUFF                                      *
			*   RX PKT BUFF                                      *
			*****************************************************/

			/* Mapping above DMA memory into
				corresponding BD structure format:*/

			/* Virtual bdp address initialization */
			txring->wb_bd_tbl_va =
				(struct wb_bd *)((((unsigned long)cpu_addr) +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));

			rxring->wb_bd_tbl_va = (struct wb_bd *)
			 	(((unsigned long)txring->wb_bd_tbl_va +
			 	(NUM_TX_WB_DESCR * sizeof(struct wb_bd)) +
			 	ALIGN_64_BIT_MASK - 1) &
			 	~(ALIGN_64_BIT_MASK - 1));

			txring->bd_tbl_va = (struct bd *)
			 	(((unsigned long)rxring->wb_bd_tbl_va +
			 	(NUM_RX_WB_DESCR * sizeof(struct wb_bd)) +
			 	ALIGN_64_BIT_MASK - 1) &
			 	~(ALIGN_64_BIT_MASK - 1));
			
			rxring->bd_tbl_va = (struct bd *)
			 	(((unsigned long)txring->bd_tbl_va +
			 	(NUM_TX_DESCR * sizeof(struct bd)) +
			 	ALIGN_64_BIT_MASK - 1) &
			 	~(ALIGN_64_BIT_MASK - 1));
															
			txring->bd_buff_va   = (struct bd *)
			 	(((unsigned long)rxring->bd_tbl_va +
			 	(NUM_RX_DESCR * sizeof(struct bd)) +
			 	ALIGN_64_BIT_MASK - 1) &
			 	~(ALIGN_64_BIT_MASK - 1));
															
			rxring->bd_buff_va   = (struct bd *)
			 	(((unsigned long)txring->bd_buff_va +
			 	(NUM_TX_DESCR * MAX_BUFF_SIZE) +
			 	ALIGN_64_BIT_MASK - 1) &
			 	~(ALIGN_64_BIT_MASK - 1));
															
			/* Physical bdp address initialization */
			txring->wb_bd_tbl_pa = (dma_addr_t)
				(((unsigned long)txring->dev_addr +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));

			rxring->wb_bd_tbl_pa = (dma_addr_t)
				(((unsigned long)txring->wb_bd_tbl_pa +
				(NUM_TX_WB_DESCR * sizeof(struct wb_bd)) +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));
															
			txring->bd_tbl_pa    = (dma_addr_t)
				(((unsigned long)rxring->wb_bd_tbl_pa +
				(NUM_RX_WB_DESCR * sizeof(struct wb_bd)) +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));
															
			rxring->bd_tbl_pa = (dma_addr_t)
				(((unsigned long)txring->bd_tbl_pa +
				(NUM_TX_DESCR * sizeof(struct bd)) +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));
															
			txring->bd_buff_pa = (dma_addr_t)
				(((unsigned long)rxring->bd_tbl_pa +
				(NUM_RX_DESCR * sizeof(struct bd)) +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));
															
			rxring->bd_buff_pa = (dma_addr_t)
				(((unsigned long)txring->bd_buff_pa +
				(NUM_TX_DESCR * MAX_BUFF_SIZE) +
				ALIGN_64_BIT_MASK - 1) &
				~(ALIGN_64_BIT_MASK - 1));
															
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,
				"va %p %p %p %p %p %p %p\n", (void *)cpu_addr,
				(void *)txring->wb_bd_tbl_va,
				(void *)rxring->wb_bd_tbl_va,
				(void *)txring->bd_tbl_va,
				(void *)rxring->bd_tbl_va,
				(void *)txring->bd_buff_va,
				(void *)rxring->bd_buff_va);
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,
				"pa %u %lx %lx %lx %lx, txring->wb_bd_tbl_pa %lx, rxring->wb_bd_tbl_pa %lx\n",
				*dev_addr, (ULONG)txring->bd_tbl_pa,
				(ULONG)rxring->bd_tbl_pa,
				(ULONG)txring->bd_buff_pa,
				(ULONG)rxring->bd_buff_pa,
				(ULONG)txring->wb_bd_tbl_pa,
				(ULONG)rxring->wb_bd_tbl_pa);

			txring->head = txring->tail = txring->bd_tbl_va;
			txring->wb_read_ptr = txring->wb_bd_tbl_va;
			txring->ring_len = NUM_TX_DESCR;
			txring->wb_seq_num = 1;
			txring->seq_num = 1;
			txring->wb_ring_len = NUM_TX_WB_DESCR;
			txring->flag = TX_BD_INIT_RING;

			rxring->head = rxring->tail = rxring->bd_tbl_va;
			rxring->wb_read_ptr = rxring->wb_bd_tbl_va;
			rxring->ring_len = NUM_RX_DESCR;
			rxring->seq_num = 1;
			rxring->wb_seq_num = 1;
			rxring->wb_ring_len = NUM_RX_WB_DESCR;
			rxring->flag = RX_BD_INIT_RING;
		} /* End of inner for loop */

	return 1;
}

/***************************************************************************
 * Function Name  : fp_bd_init_ring
 * Description    : initialize the tx/tx/wb rings
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 ***************************************************************************/
int fp_bd_init_ring(struct bd_ring *ring)
{
	int desc_index = 0;
	struct bd *bd_va, *bd_pa;
	UINT bd_bufaddr;

	if (ring == NULL) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "%s: null ring ptr\n", __func__);
		return -1;
	}

	bd_va = ring->bd_tbl_va;
	bd_pa = (struct bd *)(ULONG)ring->bd_tbl_pa;
	bd_bufaddr = (UINT)ring->bd_buff_pa;

	for (desc_index = 0; desc_index < ring->ring_len; desc_index++) {

		bd_va->bd_ctrl = 0;
		bd_va->rsvd = 0;
		bd_va->bd_bufaddr = (UINT) (bd_bufaddr + ((desc_index + 0) * MAX_BUFF_SIZE));

		if (ring->flag == RX_BD_INIT_RING) {
			bd_va->bd_ctrl = BD_CTRL_PKT_INT_EN | BD_CTRL_DIR;
			bd_va->bd_buflen = MAX_FRAME_SIZE;
			fp_enque_bd(ring, bd_va);
		}

		if (desc_index == ring->ring_len - 1) {
			bd_va->bd_nextptr = (ULONG)bd_pa;
			bd_va->bd_ctrl |= BD_CTRL_LAST_BD;

		} else {
			bd_va->bd_nextptr = (ULONG)(bd_pa + desc_index + 1);
		}

//		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "va: %p pa %u buf_ptr_pa %u\n", (void *)bd_va, bd_va->bd_nextptr, bd_va->bd_bufaddr);
		bd_va = bd_va + 1;
	}

	return 0;
}

/***************************************************************************
 * Function Name  : fp_dump_wb_bd_info
 * Description    : dump write back bd
 * Inputs
 *   Parameters   : struct wb_bd *
 * Outputs        :
 *   Parameters   :
 *   Returns      : void
 * Changes        :
 ***************************************************************************/
void fp_dump_wb_bd_info(struct wb_bd *dump_bd)
{
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "WBBD:\t%p size %u\n", dump_bd, (unsigned int)sizeof(struct wb_bd) );
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "ctrl\t%x bd_buflen\t%x bd_seqnum\t%x\n", dump_bd->bd_ctrl, dump_bd->bd_buflen, dump_bd->bd_seqnum);
	/*  FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "bd_rsvd1\t%x bd_revd2\t%x\n", dump_bd->rsvd[0], dump_bd->rsvd[1]); */
}

/***************************************************************************
 * Function Name  : fp_dump_bd_info
 * Description    : dump tx/rx bd
 * Inputs
 *   Parameters   : struct bd *
 * Outputs        :
 *   Parameters   :
 *   Returns      : void
 * Changes        :
 ***************************************************************************/
void fp_dump_bd_info(struct bd *dump_bd)
{
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "ctrl    %x seqnum  %x\n",
					dump_bd->bd_ctrl, dump_bd->bd_seqnum);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "buflen  %x rsvd    %x\n",
					dump_bd->bd_buflen, dump_bd->rsvd);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "bufaddr %x nextptr %x\n",
					dump_bd->bd_bufaddr,
					dump_bd->bd_nextptr);
}

/***************************************************************************
 * Function Name  : fp_wb_bd_init_ring
 * Description    : initialize the tx/tx/wb rings
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 ***************************************************************************/
int fp_wb_bd_init_ring(struct bd_ring *ring)
{
	INT desc_index = 0;
	struct wb_bd *curr_bd;

	if (ring == NULL) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "%s: null ring ptr\n",
								__func__);
		return -1;
	}

	memset(ring->wb_bd_tbl_va, 0, sizeof(struct wb_bd));

	curr_bd = ring->wb_bd_tbl_va;
	for (desc_index = 0; desc_index < ring->wb_ring_len; desc_index++) {
		curr_bd->bd_seqnum = 0xff;
		curr_bd++;
	}

	return 0;
}

/***************************************************************************
 * Function Name  : fp_get_next_free_bd
 * Description    : pointer to next free bd in ring
 * Inputs
 *   Parameters   : pointer to struct bd_ring
 * Outputs        :
 *   Parameters   : 
 *   Returns      : int
 * Changes        :
 ***************************************************************************/
struct bd *fp_get_next_free_bd(struct bd_ring *ring)
{
	struct bd *tail;

	if (ring == NULL) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG,
			"%s: get_next_free_bd: bd_ring is NULL\n", __func__);
		return NULL;
	}

	tail = ring->tail;

	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "Free bd %p\n", tail);

	if (tail->bd_ctrl & BD_CTRL_DESC_EN)
		return NULL;
	else
		return tail;
}

/***************************************************************************
 * Function Name  : fp_get_count_of_free_bd
 * Description    : get value of free bd in ring
 * Inputs
 *   Parameters   : pointer to struct bd_ring
 * Outputs        :
 *   Parameters   :
 *   Returns      : int
 * Changes        :
 ***************************************************************************/
int fp_get_count_of_free_bd( struct bd_ring *ring ){
	struct bd *tail;
	int32_t count = 0;

	if (ring == NULL) {
		FP_DEBUG( DBGP_FEAT_ERR, FP_LOG_EMERG,
			"%s: bd_ring is NULL\n", __func__);
		return 0;
	}

	tail = ring->tail;
	while( count < ring->ring_len ){
		if( tail->bd_ctrl & BD_CTRL_DESC_EN ){
			break;
		}
		count += 1;
		tail++;
		if( tail >= ( ring->bd_tbl_va + ring->ring_len ) ){
			tail = ring->bd_tbl_va;
		}
	}
	return count;
}
/**********************************************************************
 * Function Name  : fp_get_curr_bd_index
 * Description    : get the curr bd index from the ring
 * Inputs
 *   Parameters   : struct bd_ring *, struct bd *
 * Outputs        :
 *   Parameters   :
 *   Returns      : int
 * Changes        :
 *********************************************************************/
UINT fp_get_curr_bd_index(struct bd_ring *ring, struct bd *bd)
{
	return (UINT)(bd - ring->bd_tbl_va);
}

/**********************************************************************
 * Function Name  : fp_check_bd_status
 * Description    : check write back bd is avail or not for software
 * Inputs
 *   Parameters   : struct bd_ring *
 * Outputs        :
 *   Parameters   :
 *   Returns      : int
 * Changes        :
 *********************************************************************/
int fp_check_bd_status(struct bd_ring *ring)
{
	struct wb_bd *curr_wb = ring->wb_read_ptr;

	/* Check expected seq num in the bd */
	if ((ring->wb_seq_num == curr_wb->bd_seqnum) &&
			(curr_wb->bd_ctrl && WB_BD_CTRL_DESC_EN)) {

		ring->wb_seq_num = ring->wb_seq_num + 1;
		return 0;

	} else {

		return -1;

	}
}

/**********************************************************************
 * Function Name  : fp_enque_bd
 * Description    : enque bd into the ring
 *                  after enque, bd gets owned by hardware unti it dequeued
 * Inputs
 *   Parameters   : struct bd_ring *, struct bd *
 * Outputs        :
 *   Parameters   :
 *   Returns      : int
 * Changes        :
 *********************************************************************/
int fp_enque_bd(struct bd_ring *ring, struct bd *curr_bd)
{
	if (curr_bd->bd_ctrl & BD_CTRL_DESC_EN) {

		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_INFO, "enque_bd: DESC_EN is already set\n");
		return -1;
	}

	/* Update the bd seq. num */
	curr_bd->bd_seqnum = ring->seq_num;
	ring->seq_num = ring->seq_num + 1;
	curr_bd->bd_ctrl |= BD_CTRL_DESC_EN;

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s:ring->tail is %lx\n", __func__, (unsigned long)ring->tail);

	/* update tail to next free location */
	ring->tail = ring->tail + 1;

	if (ring->tail >= (ring->bd_tbl_va + ring->ring_len)){
		ring->tail = ring->bd_tbl_va;
	}

	return 0;
}

/**********************************************************************
 * Function Name  : fp_deque_bd
 * Description    : deque bd from the ring
 *                  after deque, bd gets owned by software unti it enqueued
 * Inputs
 *   Parameters   : struct bd_ring *, unsigend int *
 * Outputs        :
 *   Parameters   :
 *   Returns      : struct bd *
 * Changes        :
 *********************************************************************/
struct bd *fp_deque_bd(struct bd_ring *ring, UINT *wb_buflen)
{
	struct bd *head;
	struct wb_bd *wrbk = ring->wb_read_ptr;
	struct wb_bd *wrbk_end = ring->wb_bd_tbl_va + ring->wb_ring_len;

	head = ring->head;

	fp_dump_wb_bd_info(wrbk);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "deque_bd: head is %lx, wrbk is %lx\n", (unsigned long)head, (unsigned long)wrbk);

	if ((head->bd_seqnum != wrbk->bd_seqnum) || (!(head->bd_ctrl & BD_CTRL_DESC_EN))) {

		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "head seq num and wb seq num are not equal %d %d\n", head->bd_seqnum, wrbk->bd_seqnum);
		return NULL;
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "seq_no %x, bd_ctrl %x, buf_len %x\n", wrbk->bd_seqnum, wrbk->bd_ctrl, wrbk->bd_buflen);

	*wb_buflen = ( ( wrbk->bd_ctrl & 0x7FF ) << 16 )|( wrbk->bd_buflen & 0xFFFF );

	/* Reset DESC_EN bit, */
	head->bd_ctrl &= (~(BD_CTRL_DESC_EN));

	head->bd_seqnum = 0;

	ring->head = ring->head + 1;

	if (ring->head >= (ring->bd_tbl_va + ring->ring_len)){
		ring->head = ring->bd_tbl_va;
	}

	ring->wb_read_ptr = (((wrbk + 1) >= wrbk_end) ? (ring->wb_bd_tbl_va) : (wrbk + 1));

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "read_ptr is %lx, wrbk_end  %lx\n", (unsigned long)ring->wb_read_ptr, (unsigned long)wrbk_end);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "ring->wb_read_ptr is %lx\n", (unsigned long)ring->wb_read_ptr);

	return head;
}

/**********************************************************************
 * Function Name  : fp_bd_init
 * Description    : initialize the rx bd ring & tx bd ring
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 *********************************************************************/
int fp_bd_init(fp_private_t *fp)
{
	UINT ch_index = 0;

	/* Allocate memory and initialize bd rings */
	if (fp_bd_create_ring(fp) < 0) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "init the txring failed...so return\n");
		return -ENOMEM;
	}

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		if (fp_bd_init_ring(fp->hif.ch[ch_index].txring) < 0)
        {
			FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "channel %u init txring failed\n", ch_index);
			return -ENOMEM;
		}

		if (fp_bd_init_ring(fp->hif.ch[ch_index].rxring) < 0) {
			FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "channel %u init rxring failed\n", ch_index);
			return -ENOMEM;
		}

		if (fp_wb_bd_init_ring(fp->hif.ch[ch_index].txring) < 0) {
			FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "channel %u init wbring failed\n", ch_index);
			return -ENOMEM;
		}

		if (fp_wb_bd_init_ring(fp->hif.ch[ch_index].rxring) < 0) {
			FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "channel %u init wbring failed\n", ch_index);
			return -ENOMEM;
		}
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "END OF BD _INIT\n");

	return 0;
}
