/*************************************************************************/ /*!
@File
@Title          fp_bd_table.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    contains generic bd rx ring & tx ring API's
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <fp_macros.h>
#include <fp_types.h>

#include <fp_bd_table.h>
#include "fp_net_driver.h"
#include "fp_hw_util.h"
#include "fp_hif_reg.h"

/**********************************************************************
 *                                      #defines                                               
 **********************************************************************/
#define SRAM_RETENTION_BUG 1

/**********************************************************************
 *                              Extern Declarations    
***********************************************************************/
#if 0
/**********************************************************************
 * Function Name  : bd_ring_init
 * Description    : initialize the buffer discriptor tables for both 
 *                  rx side & tx side. Chain the descriptors as a circular
 *                  queue, used to store the packets in these descriptors
 *                  via dma controller
 * Inputs
 *   Parameters   : struct fp_private *, int, dma_addr_t *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : struct bd * - base address of buffer descriptor
 * Changes        :
 *********************************************************************/
struct bd *
bd_ring_init(struct fp_private *fp, int num_descr, dma_addr_t *bd_tbl_pa, int xfermode)
{
    int32 i, size;
    struct bd *bd_tbl, *d, *d_pa, *bd_start_pa;
    volatile uint32 next_addr;

    /* Descr queue size should be multiple of 2 */
    ASSERT((num_descr > 0) && ((num_descr & 1) == 0));

    /* Size of the buffer descriptor ring includes dummy
     * descriptor at the end
     */
    size = (num_descr * sizeof(struct bd));

    /* Allocate and map the memory for bd ring */
#ifdef NAKSHA_RAM_DESCRIPTORS
    if (xfermode == NAKSHA_RXSIDE)
        bd_tbl = (struct bd *)(fp->pcimem_base + NAKSHA_RAM_RXDESCS_STARTADDR);
    else
        bd_tbl = (struct bd *)(fp->pcimem_base + NAKSHA_RAM_TXDESCS_STARTADDR);

    *bd_tbl_pa = (dma_addr_t)((unsigned char *)bd_tbl - fp->pcimem_base);
    d_pa = (struct bd *)NAKSHA2NAKSHARAM(*bd_tbl_pa);
    *bd_tbl_pa = (dma_addr_t)d_pa;
#else
    bd_tbl = pci_alloc_consistent(fp->pci_dev, size, bd_tbl_pa);
    d_pa = (struct  bd *)NAKSHA2HOST(*bd_tbl_pa);
#endif
    bd_start_pa = d_pa;

    if (bd_tbl == NULL)
    {
        printk(KERN_CRIT "bd ring memory alloc of size %d failed\n",
               size);
        return(NULL);
    }

    /* Assert 64 bit alignment */
    ASSERT(((uint32)bd_tbl & 63) == 0);
    /* Initialize the descriptor entries by chaining together as
     * a cicrular queue. The host driver always sees the bd ring
     * as array of entries, whereas the device will view it as a
     * entries chained as circular list.
     */
    d_pa++;
#ifdef DEBUG
    if (xfermode == NAKSHA_RXSIDE)
        printk("RxDesc setting\n");
    else
        printk("TxDesc setting\n");
#endif

    for (i = 0; i < num_descr - 1; i++)
    {
        d = bd_tbl + i;
        d->bd_ctrl = d->bd_status = d->bd_bufptr = 0;
        d->bd_next = (struct bd *)n_cpu_to_be32((uint32)d_pa);
#ifdef DEBUG
        printk("Desc setting: Current is 0x%x, next at 0x%x\n", bd_start_pa + i, d_pa);
#endif
        d_pa++;
    }

    /* Chain the last descriptor to first one */
    d = bd_tbl + i;
    d->bd_ctrl = d->bd_status = d->bd_bufptr = 0;
    d->bd_next = (struct bd *)n_cpu_to_be32(bd_start_pa);

#ifdef SRAM_RETENTION_BUG
    d->bd_status = 0x0f0;
    next_addr  = 0;
    next_addr  = (unsigned int) d->bd_next;
#ifdef DEBUG_PCI1
    printk ("Next Address Value returned: 0x%x\n", next_addr);
    printk("Desc setting: Current is 0x%x, next at 0x%x\n", bd_start_pa + i, bd_start_pa);
    printk("Desc setting: bd_end_ptr: 0x%x, bd_next is 0x%x\n", d, d->bd_next);
#endif
#endif

    return (bd_tbl);
}

/**********************************************************************
 * Function Name  : bd_tx_ring_cleanup
 * Description    : free the tx buffer descriptors, release the mapped 
 *                  pci memory
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
void
bd_tx_ring_cleanup(struct fp_private *fp, int32 start, int32 num_descr)
{
    uint32 i, size;
    struct bd *d;
#ifdef DEBUG_PCI1
    printk("txbd_tbl is at 0x%x\n", (unsigned int)fp->rx_bd_tbl);
#endif
    if (fp->tx_bd_tbl == NULL)
        return;
#ifdef NAKSHA_RAM_DESCRIPTORS
    fp->tx_bd_tbl = NULL;
    return;
#endif

    for (i = start; i < num_descr; i++)
    {
        d = fp->tx_bd_tbl + i;

        if (d->bd_bufptr != 0)
        {
            pci_unmap_single(fp->pci_dev,
                             fp->tx_buf_pa[i],
                             n_be32_to_cpu(d->bd_ctrl) & BD_CTRL_BD_BUFLEN_MASK,
                             DMA_TO_DEVICE);

            ASSERT(fp->tx_buf_skb[i]);
            dev_kfree_skb(fp->tx_buf_skb[i]);
            d->bd_bufptr = 0;
            fp->tx_buf_skb[i] = (struct sk_buff*) NULL;
            fp->tx_buf_pa[i] = 0;
        }
    }

    DPRINTK(KERN_DEBUG, "Freed %d buffers in device queue\n", num_descr);

    size = (num_descr * sizeof(struct bd));

    /* Free the tx bd ring */
    pci_free_consistent(fp->pci_dev, size, fp->tx_bd_tbl, fp->tx_bd_tbl_pa);
    fp->tx_bd_tbl = NULL;

    DPRINTK(KERN_DEBUG, "Freed the tx bd ring memory %d bytes\n", size);

    return;
}

/**********************************************************************
 * Function Name  : bd_rx_ring_cleanup
 * Description    : free the rx buffer descriptors, release the mapped 
 *                  pci memory
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
void
bd_rx_ring_cleanup(struct fp_private *fp, int32 start, int32 num_descr)
{
    uint32 i, size;
    struct bd *d;
#ifdef DEBUG_PCI1
    printk("rxbd_tbl is at 0x%x\n",(unsigned int) fp->rx_bd_tbl);
#endif
    if (fp->rx_bd_tbl == NULL)
        return;

#ifdef NAKSHA_RAM_DESCRIPTORS
    fp->rx_bd_tbl = NULL;
    return;
#endif

    for (i = start; i < num_descr; i++)
    {
        d = fp->rx_bd_tbl + i;
#if 0
        printk("rxcleanup: trying to unmap descriptor pa (%d): 0x%x, stored is 0x%x\n",
                                i,(unsigned int*) d->bd_bufptr,(unsigned int*) fp->rx_buf_pa[i]);
#endif

        if (d->bd_bufptr != 0)
        {
            pci_unmap_single(fp->pci_dev,
                             (dma_addr_t)fp->rx_buf_pa[i],
                             MAX_FRAME_SIZE, DMA_FROM_DEVICE);

            ASSERT(fp->rx_buf_skb[i]);
            dev_kfree_skb(fp->rx_buf_skb[i]);
            d->bd_bufptr = 0;
            fp->rx_buf_pa[i] = 0;
        }
    }

    DPRINTK(KERN_DEBUG, "Freed %d buffers in device queue\n", num_descr);

    size = (num_descr * sizeof(struct bd));

    /* Free the rx bd ring */
    pci_free_consistent(fp->pci_dev, size, fp->rx_bd_tbl, fp->rx_bd_tbl_pa);
    fp->rx_bd_tbl = NULL;

    DPRINTK(KERN_DEBUG, "Freed the rx bd ring memory %d bytes\n", size);

    return;
}

/**********************************************************************
 * Function Name  : bd_tx_ring_init
 * Description    : initialize the tx bd ring while module init 
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 *********************************************************************/
int32
bd_tx_ring_init(struct fp_private *fp)
{
    int i;
    struct bd *txd;
    unsigned char *tx_buf_va, *tx_buf_pa;
    /* Allocate and initialize the tx bd ring */
    fp->tx_bd_tbl = bd_ring_init(fp, NUM_TX_DESCR, &fp->tx_bd_tbl_pa, NAKSHA_TXSIDE);
#ifdef DEBUG_PCI1
    printk("txbd_table VA is 0x%x and pa is 0x%x\n",(unsigned int) fp->tx_bd_tbl,(unsigned int) fp->tx_bd_tbl_pa);
#endif
    if (fp->tx_bd_tbl == NULL)
    {
        printk(KERN_CRIT "Out of memory allocating tx bd ring\n");
        return (-1);
    }
#ifdef NAKSHA_RAM_PKTS
    tx_buf_va = (UCHAR *) fp->pcimem_base + NAKSHA_RAM_TXPKTS_STARTADDR;
    tx_buf_pa = (UCHAR *)NAKSHA2NAKSHARAM(tx_buf_va - fp->pcimem_base);
    /* Allocate buffers for the rx frames */
#ifdef DEBUG_PCI1
    printk("%s: Desc Dump, buffer starts at 0x%x, MAX_BUF_ALLOCSIZE is %d \n", __FUNCTION__, tx_buf_pa, MAX_BUF_ALLOCSIZE);
#endif
    for (i = 0; i < NUM_TX_DESCR; i++)
    {
        txd = fp->tx_bd_tbl + i;
        txd->bd_bufptr = (UINT) (tx_buf_pa + (i * MAX_BUF_ALLOCSIZE));
#ifdef DEBUG
        printk("%d:txd is 0x%x, buffer is 0x%x, next is 0x%x \n", i, txd, txd->bd_bufptr, txd->bd_next);
#endif
    }
#endif

    /* Initialize the tx bd write pointer */
    fp->txd_index = 0;

    return (0);
}

/**********************************************************************
 * Function Name  : bd_rx_ring_init
 * Description    : initialize the rx bd ring while module init 
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 *********************************************************************/
int
bd_rx_ring_init(struct fp_private *fp)
{
    int i = 0;
    uint32 rx_buf_pa, rx_buf_va;
    struct bd *rxd;
#ifndef NAKSHA_RAM_PKTS
    struct sk_buff *skb;
#endif

    /* Allocate and initialize the rx bd ring */
    fp->rx_bd_tbl = bd_ring_init(fp, NUM_RX_DESCR, &fp->rx_bd_tbl_pa, NAKSHA_RXSIDE);

    if (fp->rx_bd_tbl == NULL) {
        printk(KERN_CRIT "Out of memory allocating rx bd ring\n");
        return (-1);
    }

#ifdef DEBUG_PCI1
    printk("txbd_table VA is 0x%x and pa is 0x%x\n",(unsigned int) fp->rx_bd_tbl,(unsigned int) fp->rx_bd_tbl_pa);
#endif

#ifdef NAKSHA_RAM_PKTS
    rx_buf_va = (UINT32) (fp->pcimem_base + NAKSHA_RAM_RXPKTS_STARTADDR);
    rx_buf_pa = (uint32)NAKSHA2NAKSHARAM((rx_buf_va - (uint32)fp->pcimem_base));
#endif

#ifdef DEBUG_PCI
    printk("%s: Desc Dump\n", __FUNCTION__);
#endif
    /* Allocate buffers for the rx frames */
    for (i = 0; i < NUM_RX_DESCR - 1; i++)
    {
        rxd = fp->rx_bd_tbl + i;
#ifdef NAKSHA_RAM_PKTS
        rxd->bd_bufptr = rx_buf_pa + (i * MAX_BUF_ALLOCSIZE);
        fp->rx_buf_pa[i] = rxd->bd_bufptr;
#else /* NAKSHA_RAM_PKTS */
        skb = dev_alloc_skb(MAX_FRAME_SIZE);

        if (skb == NULL)
        {
            printk(KERN_CRIT "Out of memory allocating rx buffers\n");
            bd_rx_ring_cleanup(fp, 0, i);
            return (-1);
        }

        rx_buf_pa = pci_map_single(fp->pci_dev, skb->data,
                                   MAX_FRAME_SIZE, DMA_FROM_DEVICE);

        if (rx_buf_pa == 0)
        {
            printk(KERN_ERR "Error mapping memory\n");
            bd_rx_ring_cleanup(fp, 0, i);
            dev_kfree_skb(skb);
            return (-1);
        }

        if (rx_buf_pa & 0x80000000) {
            printk("RX Init: Can't use this packet, as we can't DMA from this Address\n");
            bd_rx_ring_cleanup(fp, 0, i);
            dev_kfree_skb(skb);
            return -1;
        }

        /* Save the virtual and physical addresses */
        fp->rx_buf_skb[i] = skb;
        fp->rx_buf_pa[i] = rx_buf_pa;
        rxd->bd_bufptr = NAKSHA2HOST(rx_buf_pa);
#endif /* NAKSHA_RAM_PKTS */

        /* Init the descr fields */
        /* Set the buffer address and size */
        rxd->bd_status = 0;

        rxd->bd_ctrl = n_cpu_to_be32(MAX_FRAME_SIZE);
        /* Enable per packet receive interrupt */
        rxd->bd_ctrl |= n_cpu_to_be32(BD_CTRL_PKT_INT_EN |
                                    BD_CTRL_LIFM | BD_CTRL_DIR |
                                    BD_CTRL_DESC_EN);
#ifdef DEBUG
        printk("%d:rxd is 0x%x, buffer is 0x%x, next is 0x%x, ctrl is 0x%x \n", i, rxd, rx_buf_pa + (i * MAX_BUF_ALLOCSIZE), rxd->bd_next, rxd->bd_ctrl);
#endif
    }
    {
#ifdef NAKSHA_RAM_PKTS
       /* For the last one fill the ctrl as 0 */
        rxd++;
        rxd->bd_bufptr = rx_buf_pa + (i * MAX_BUF_ALLOCSIZE);
        fp->rx_buf_pa[i] = rxd->bd_bufptr;
        rxd->bd_ctrl = 0;
#endif /* NAKSHA_RAM_PKTS */
    }

    DPRINTK(KERN_DEBUG, "Rx bd ring initialized\n");

    /* Initialize the rx bd read pointer */
    fp->rxd_index = 0;

    return (0);
}

/**********************************************************************
 * Function Name  : fp_bd_init
 * Description    : initialize the rx bd ring & tx bd ring 
 * Inputs
 *   Parameters   : struct fp_private *, int, int
 * Outputs        :
 *   Parameters   : struct fp_private *
 *   Returns      : int
 * Changes        :
 *********************************************************************/
int fp_bd_init(struct fp_private *fp)
{
    /* Allocate memory and initialize bd rings */
    if (bd_tx_ring_init(fp) < 0) {
        printk("init the txring failed...so return\n");
        spin_unlock_bh(&fp->lock);
        return (-ENOMEM);
    }

    if (bd_rx_ring_init(fp) < 0)
    {
        printk("init the rxring failed...so return\n");
        bd_tx_ring_cleanup(fp, 0, NUM_TX_DESCR);
        spin_unlock_bh(&fp->lock);
        return (-ENOMEM);
    }

    return 0;
}

#endif
