/*************************************************************************/ /*!
*  @File
*  @Title          fp_drv_mcu.c
*  @Copyright      Copyright (c) Milandr. All Rights Reserved
*  @Description    Initializes EAC paramaters
*  @License        Dual MIT/GPLv2
*
*  The contents of this file are subject to the MIT license as set out below.
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  The above copyright notice and this permission notice shall be included in
*  all copies or substantial portions of the Software.
*
*  Alternatively, the contents of this file may be used under the terms of
*  the GNU General Public License Version 2 ("GPL") in which case the provisions
*  of GPL are applicable instead of those above.
*
*  If you wish to allow use of your version of this file only under the terms of
*  GPL, and not to allow others to use your version of this file under the terms
*  of the MIT license, indicate your decision by deleting the provisions above
*  and replace them with the notice and other provisions required by GPL as set
*  out in the file called "GPL-COPYING" included in this distribution. If you do
*  not delete the provisions above, a recipient may use your version of this file
*  under the terms of either the MIT license or GPL.
*
*  This License is also included in this distribution in the file called
*  "MIT-COPYING".
*
*  EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
*  PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
*  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
*  PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
*  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
*  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
*  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
*                                      Include Files
**********************************************************************/
#include <linux/timer.h>
#include <linux/netdevice.h>
#include <linux/slab.h>

#include "../include/fp_library.h"

#include "fp_drv_mcu.h"
#include "fp_net_driver.h"

/**********************************************************************
*                                      #defines
**********************************************************************/

//#define DRV_MCU_TRANSFER_TIME_STEP      (HZ / 1000)
#define DRV_MCU_TRANSFER_TIME_STEP      (HZ / 100)

/**********************************************************************
 *                              Global Declarations
 ***********************************************************************/
/* The timer structure used to performe driver_to_mcu_state_machine */
static struct timer_list fp_drv_mcu_ops_timer;
static struct drv_mcu_transfer_s drv_2_mcu;
spinlock_t fp_drv_mcu_ops_lock;

void init_drv_mcu_ops_struct( void ){
   spin_lock_init( &fp_drv_mcu_ops_lock );
   memset( &drv_2_mcu, 0, sizeof( struct drv_mcu_transfer_s ) );
}
/**********************************************************************
 * Function Name  : get_transfer_data
 * Description    : get data of transfer, if the transfer is completed
 * Inputs         : data - pointer, where data will be written
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success, 1 - transfer processing
 ********************************************************************/
uint32_t get_transfer_data( uint32_t *data, uint32_t size_in_words ){
   uint32_t i, result;

   result = 1;
   spin_lock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "get_transfer_data: spin LOCK fp_drv_mcu_ops_lock\n" );
   if( drv_2_mcu.result == 0 ){
      if( size_in_words > 2 ){
        size_in_words = 2;
      }
      for( i = 0; i < size_in_words; i++ ){
         data[ i ] = drv_2_mcu.data[ i ];
      }
      result = 0;
   }
   spin_unlock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "get_transfer_data: spin UNLOCK fp_drv_mcu_ops_lock\n" );
   return result;  
}

/**********************************************************************
 * Function Name  : set_transfer_data
 * Description    : set data of transfer, if it needed
 * Inputs         : data - value
 *                : param_number - data word in buffer, range: 0 - 1
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success, 1 - transfer processing
 ********************************************************************/
uint32_t set_transfer_data( uint32_t value, uint32_t param_number ){
   uint32_t i, result;

   result = 0;
   param_number &= 1;
   i = param_number ? FP_DATA2_REG : FP_DATA1_REG;
   if( set_pci_cfg_word( i, value ) ){
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "set_transfer_data: write data to DATA%d_REG failed\n", i );
      result = 1;
   }
   return result;  
}
/**********************************************************************
 * Function Name  : start_drv_mcu_transfer
 * Description    : start of driver to mcu transfer
 * Inputs         : user_cmd - command to execute
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success, 1 - error (command not supported)
 ********************************************************************/
uint32_t start_drv_mcu_transfer( enum drv_mcu_transfer_cmd_e user_cmd, uint32_t *data ){
   uint32_t result;

   result = 1;
   spin_lock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "start_drv_mcu_transfer: spin LOCK fp_drv_mcu_ops_lock\n" );
   if( ( ( user_cmd & DRV_MCU_CMD_REG_CMD_MASK ) < MAX_DRV_MCU_CMD ) && ( drv_2_mcu.result == 0 ) ){
      if( ( ( user_cmd & DRV_MCU_CMD_REG_CMD_MASK ) == SET_ETH_PHY_REG_I2C )&&( data != FP_NULL ) ){
         set_transfer_data( data[ 0 ], 0 );
      }
      drv_2_mcu.result = 1;
      drv_2_mcu.flags = 0;
      drv_2_mcu.cmd = user_cmd;
      fp_drv_mcu_ops_timer.expires = jiffies + DRV_MCU_TRANSFER_TIME_STEP;
      add_timer( &fp_drv_mcu_ops_timer );
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "start_drv_mcu_transfer: transfer start, cmd 0x%08X\n", user_cmd );
      result = 0;
   }
   spin_unlock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "start_drv_mcu_transfer: spin UNLOCK fp_drv_mcu_ops_lock\n" );
   return result;
}
/**********************************************************************
 * Function Name  : is_transfer_complete
 * Description    : get status of driver to mcu transfer
 * Inputs         : flags - pointer, where flags will be written, if the transfer is completed
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success, 1 - transfer processing
 ********************************************************************/
uint32_t is_transfer_complete( uint32_t *flags ){
   uint32_t result;

   result = 1;
   spin_lock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "is_transfer_complete: spin LOCK fp_drv_mcu_ops_lock\n" );
   if( drv_2_mcu.result == 0 ){
      *flags = drv_2_mcu.flags;
      result = 0;
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "is_transfer_complete: transfer complete\n" );   
   }
   spin_unlock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "is_transfer_complete: spin UNLOCK fp_drv_mcu_ops_lock\n" );
   return result;
}
/**********************************************************************
 * Function Name  : get_last_transfer_cmd
 * Description    : get last command of transfer
 * Inputs         : cmd - pointer, where command will be written
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success, 1 - transfer processing
 ********************************************************************/
// uint32_t get_last_transfer_cmd( uint32_t *cmd ){
//    *cmd = drv_2_mcu.cmd;
//    return 0;
// }
/**********************************************************************
 * Function Name  : fp_drv_mcu_dispatcher
 * Description    : dispatcher of driver to mcu transfer state machine
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
void fp_drv_mcu_dispatcher( unsigned long data ){
   uint32_t local_flags;

   local_flags = 0xFFFFFFFF;
   spin_lock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: spin LOCK fp_drv_mcu_ops_lock\n" );
   switch( drv_2_mcu.state ){
      /*--------------------------------------*/
      /*--------------------------------------*/
      /*--------------------------------------*/
      case STATE_NO_CMD:
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: STATE_NO_CMD\n" );
         /* try to start driver to mcu transfer */

         if( set_pci_cfg_word( FP_CMD_REG, drv_2_mcu.cmd ) ){
            /* write cmd to CMD_REG failed */
            drv_2_mcu.write_counter++;
            FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "fp_drv_mcu_dispatcher: write cmd to CMD_REG failed (step 1), trials %d!\n", drv_2_mcu.write_counter );
         }else{
            /* write cmd to CMD_REG pass */
            drv_2_mcu.write_counter = 0;
            drv_2_mcu.state = STATE_WAIT_STATUS;
         }
         break;
      /*--------------------------------------*/
      /*--------------------------------------*/
      /*--------------------------------------*/
      case STATE_WAIT_STATUS:
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: STATE_WAIT_STATUS\n" );
         if( get_pci_cfg_word( FP_STATUS_REG, &drv_2_mcu.flags ) ){
            drv_2_mcu.read_counter++;
            FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: can not read flags from STATUS_REG (step 2), trials %d\n", drv_2_mcu.read_counter );
         }else{
            if( drv_2_mcu.flags & DRV_MCU_STATUS_REG_MASK ){
               drv_2_mcu.read_counter = 0;
               drv_2_mcu.state = STATE_CLEAR_CMD;
            }
         }
         break;
      /*--------------------------------------*/
      /*--------------------------------------*/
      /*--------------------------------------*/
      case STATE_CLEAR_CMD:
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: STATE_CLEAR_CMD\n" );
         if( set_pci_cfg_word( FP_CMD_REG, NO_CMD ) ){
            /* write NO_CMD to CMD_REG failed */
            drv_2_mcu.write_counter++;
            FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: write NO_CMD to CMD_REG failed (step 3), trials %d!\n", drv_2_mcu.write_counter );
         }else{
            /* write cmd to CMD_REG pass */
            drv_2_mcu.write_counter = 0;
            drv_2_mcu.state = STATE_CLEAR_STATUS;
         }
         break;
      /*--------------------------------------*/
      /*--------------------------------------*/
      /*--------------------------------------*/
      case STATE_CLEAR_STATUS:
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: STATE_CLEAR_STATUS\n" );
         if( get_pci_cfg_word( FP_STATUS_REG, &local_flags ) ){
            drv_2_mcu.read_counter++;
            FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: can not read flags from STATUS_REG (step 4), trials %d\n", drv_2_mcu.read_counter );
         }else{
            if( !( local_flags & DRV_MCU_STATUS_REG_MASK ) ){
               //if( drv_2_mcu.flags & DRV_MCU_STATUS_REG_CMPL ){
                  if( get_pci_cfg_word( FP_DATA1_REG, &drv_2_mcu.data[ 0 ] ) ){
                        FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "fp_drv_mcu_dispatcher: can not read DATA1_REG\n" );
                  }
                  if( get_pci_cfg_word( FP_DATA2_REG, &drv_2_mcu.data[ 1 ] ) ){
                        FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "fp_drv_mcu_dispatcher: can not read DATA2_REG\n" );
                  }
               //}
               drv_2_mcu.read_counter = 0;
               drv_2_mcu.state = STATE_NO_CMD;
               drv_2_mcu.result = 0;
            }
         }
         break;
      default:
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ALERT, "fp_drv_mcu_dispatcher: Oops! NEVER be here!\n" );
   }
   spin_unlock_bh( &fp_drv_mcu_ops_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "fp_drv_mcu_dispatcher: spin UNLOCK fp_drv_mcu_ops_lock\n" );
   /* Update the refresh timer to call this fucntion again */
   if( drv_2_mcu.result ){
      fp_drv_mcu_ops_timer.expires = jiffies + DRV_MCU_TRANSFER_TIME_STEP;
      add_timer( &fp_drv_mcu_ops_timer );
   }else{
      //fp_drv_mcu_ops_init_timer( 0 );
   }

}


/**********************************************************************
 * Function Name  : fp_drv_mcu_ops_init_timer
 * Description    : do driver to mcu data transfers
 * Inputs         :
 * Parameters     : init (1 - enable; 0 - disable);
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_drv_mcu_ops_init_timer( int32_t init ){
   if ( !init ){
      del_timer( ( struct timer_list * )&fp_drv_mcu_ops_timer );
   }else{
      init_timer( &fp_drv_mcu_ops_timer );
      fp_drv_mcu_ops_timer.expires  = jiffies + DRV_MCU_TRANSFER_TIME_STEP;
      fp_drv_mcu_ops_timer.data     = ( unsigned long )&fp_drv_mcu_ops_timer;
      fp_drv_mcu_ops_timer.function = fp_drv_mcu_dispatcher;
      //add_timer( ( struct timer_list * )&fp_drv_mcu_ops_timer );
   }
   return 0;
}
