/*************************************************************************/ /*!
*  @File
*  @Title          fp_drv_mcu.c
*  @Copyright      Copyright (c) Milandr. All Rights Reserved
*  @Description    Initializes EAC paramaters
*  @License        Dual MIT/GPLv2
*
*  The contents of this file are subject to the MIT license as set out below.
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  The above copyright notice and this permission notice shall be included in
*  all copies or substantial portions of the Software.
*
*  Alternatively, the contents of this file may be used under the terms of
*  the GNU General Public License Version 2 ("GPL") in which case the provisions
*  of GPL are applicable instead of those above.
*
*  If you wish to allow use of your version of this file only under the terms of
*  GPL, and not to allow others to use your version of this file under the terms
*  of the MIT license, indicate your decision by deleting the provisions above
*  and replace them with the notice and other provisions required by GPL as set
*  out in the file called "GPL-COPYING" included in this distribution. If you do
*  not delete the provisions above, a recipient may use your version of this file
*  under the terms of either the MIT license or GPL.
*
*  This License is also included in this distribution in the file called
*  "MIT-COPYING".
*
*  EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
*  PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
*  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
*  PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
*  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
*  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
*  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
*                                      Include Files
**********************************************************************/
#include <linux/timer.h>
#include <linux/netdevice.h>
#include <linux/slab.h>

#include "../include/fp_library.h"

#include "fp_drv_mcu.h"
#include "fp_net_driver.h"
#include "fp_drv_mcu_tasks.h"
#include "fp_drv_mcu.h"
#include "sfp_copper_setup.h"

/**********************************************************************
*                                      #defines
**********************************************************************/

#define DRV_MCU_TASKS_TIME_STEP_10MS    (HZ / 100)
#define DRV_MCU_TASKS_TIME_STEP_100MS   (HZ / 10)
#define DRV_MCU_TASKS_TIME_STEP_50MS    (HZ / 20)
#define DRV_MCU_TASKS_TIME_STEP_1S      (HZ / 1)

/**********************************************************************
 *                              Global Declarations
 ***********************************************************************/

/* The timer structure used to performe driver_to_mcu_state_machine */
static struct timer_list fp_drv_mcu_tasks_timer;
struct drv_mcu_all_tasks_s drv_mcu_instance;
spinlock_t fp_drv_mcu_task_lock;

/**********************************************************************
 * Function Name  : init_drv_mcu_all_tasks_struct
 * Description    : init all needed fields and structures
 * Inputs         :
 * Parameters     : init: 1 - init, 0 - deinit
 * Outputs        :
 * Parameters     : -
 * Returns        : int 0 - success; -1 - error (can not allocate memory)
 ********************************************************************/
static uint8_t my_tArray[ TASKS_QUEUE_SIZE ] = { 0, 0, TEST_CMD, GET_FW_ID, GET_TEMP, GET_SFP_EVENTS };
//static uint8_t my_tArray[ TASKS_QUEUE_SIZE ] = { 0, 0, TEST_CMD, GET_FW_ID, 0, GET_SFP_EVENTS }; //for test purpose only
int32_t init_drv_mcu_all_tasks_struct( uint32_t init ){
   uint32_t i;

   if( !init ){
      if( drv_mcu_instance.tasks_memory ){
         kfree( drv_mcu_instance.tasks_memory );
         drv_mcu_instance.tasks_memory = FP_NULL;
      }
   }else{
      spin_lock_init( &fp_drv_mcu_task_lock );
      memset( &drv_mcu_instance, 0, sizeof( struct drv_mcu_all_tasks_s ) );
      drv_mcu_instance.tasks_memory = kmalloc( ( MAX_DRV_MCU_CMD * ( DRV_MCU_TRANSFER_DATA_REGS << 2 ) ), GFP_KERNEL );
      if( !drv_mcu_instance.tasks_memory ){
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ALERT, "init_drv_mcu_all_tasks_struct: can not allocate memory for tasks!\n" );
         return -1;
      }
      drv_mcu_instance.task[ NO_CMD ].data_ptr = &drv_mcu_instance.tasks_memory[ NO_CMD * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ GET_FW_ID ].data_ptr = &drv_mcu_instance.tasks_memory[ GET_FW_ID * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ GET_TEMP ].data_ptr = &drv_mcu_instance.tasks_memory[ GET_TEMP * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ GET_SFP_EVENTS ].data_ptr = &drv_mcu_instance.tasks_memory[ GET_SFP_EVENTS * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ GET_SFP_DATA ].data_ptr = &drv_mcu_instance.tasks_memory[ GET_SFP_DATA * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ SET_PHY_RESET ].data_ptr = &drv_mcu_instance.tasks_memory[ SET_PHY_RESET * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ TEST_CMD ].data_ptr = &drv_mcu_instance.tasks_memory[ TEST_CMD * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ GET_ETH_PHY_REG_I2C ].data_ptr = &drv_mcu_instance.tasks_memory[ GET_ETH_PHY_REG_I2C * ( DRV_MCU_TRANSFER_DATA_REGS ) ];
      drv_mcu_instance.task[ SET_ETH_PHY_REG_I2C ].data_ptr = &drv_mcu_instance.tasks_memory[ SET_ETH_PHY_REG_I2C * ( DRV_MCU_TRANSFER_DATA_REGS ) ];

      drv_mcu_instance.active_task_index = 0;
      for( i = 0; i < TASKS_QUEUE_SIZE; i++ ){
         drv_mcu_instance.task_array[ i ] = my_tArray[ i ];   
      }
      init_drv_mcu_ops_struct();
   }
   return 0;
}

/**********************************************************************
 * Function Name  : clear_task
 * Description    : reset task field in drv_mcu_all_tasks_s structure, memder task
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void clear_task( enum drv_mcu_transfer_cmd_e task_number ){
   spin_lock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "clear_task: spin LOCK fp_drv_mcu_task_lock\n" );
   drv_mcu_instance.task[ task_number ].task = NO_CMD;
   spin_unlock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "clear_task: spin UNLOCK fp_drv_mcu_task_lock\n" );
}

/**********************************************************************
 * Function Name  : clear_task
 * Description    : start task
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success; 1 - error (structure is busy)
 ********************************************************************/
uint32_t start_task( enum drv_mcu_transfer_cmd_e task_number, uint32_t user_cmd ){
   uint32_t result;

   result = 1;
   spin_lock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "start_task: spin LOCK fp_drv_mcu_task_lock\n" );
   if( ( drv_mcu_instance.task[ task_number ].task == 0 ) && 
       ( drv_mcu_instance.task[ task_number ].result & DRV_TASK_RESEULT_PROCESSING ) == 0 ){
      drv_mcu_instance.task[ task_number ].task_cmd = user_cmd;
      drv_mcu_instance.task[ task_number ].task = 1;
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "start_task: task start\n" );
      result = 0;
   }
   spin_unlock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "start_task: spin UNLOCK fp_drv_mcu_task_lock\n" );
   return result;
}

/**********************************************************************
 * Function Name  : is_task_complete
 * Description    : check complete the transfer
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : 0 - success (complete); 1 - error (transfer isn't complete)
 ********************************************************************/
uint32_t is_task_complete( enum drv_mcu_transfer_cmd_e task_number ){
   uint32_t result;

   result = 1;
   spin_lock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "is_task_complete: spin LOCK fp_drv_mcu_task_lock\n" );
   if( !( drv_mcu_instance.task[ task_number ].result & DRV_TASK_RESEULT_PROCESSING ) ){
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "is_task_complete: task complete\n" );
      result = 0;
   }
   spin_unlock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "is_task_complete: spin UNLOCK fp_drv_mcu_task_lock\n" );
   return result;
}

/**********************************************************************
 * Function Name  : do_one_task_step
 * Description    : do one step of the state machin of the transfer
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void do_one_task_step( enum drv_mcu_transfer_cmd_e task_number /*, uint32_t cmd*/ ){
   uint32_t flags;
  
   spin_lock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "do_one_task_step: spin LOCK fp_drv_mcu_task_lock\n" );
   switch( drv_mcu_instance.task[ task_number ].state ){
      case STATE_IDLE:
         if( !start_drv_mcu_transfer( drv_mcu_instance.task[ task_number ].task_cmd, drv_mcu_instance.task[ task_number ].data_ptr ) ){
            drv_mcu_instance.task[ task_number ].result = DRV_TASK_RESEULT_PROCESSING;
            drv_mcu_instance.task[ task_number ].state = STATE_TASK_PROCESSING;
         }
         break;

      case STATE_TASK_PROCESSING:
         if( !is_transfer_complete( &flags ) ){

            if( drv_mcu_instance.task[ task_number ].data_ptr ){
               get_transfer_data( drv_mcu_instance.task[ task_number ].data_ptr, 2 );
            }
            if( ( flags & DRV_MCU_STATUS_REG_MASK ) == DRV_MCU_STATUS_REG_CMPL ){
               drv_mcu_instance.task[ task_number ].result = DRV_TASK_RESEULT_COMPLETE_SUCCESS;
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "do_one_task_step: transfer complete successfully\n" );
            }else{
               drv_mcu_instance.task[ task_number ].result = DRV_TASK_RESEULT_COMPLETE_ERROR;
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "do_one_task_step: transfer complete with error\n" );
            }
            drv_mcu_instance.task[ task_number ].state = STATE_WAIT_CLR_TASK;
         }
         break;

      case STATE_WAIT_CLR_TASK:
         if( drv_mcu_instance.task[ task_number ].task == NO_CMD ){
            drv_mcu_instance.task[ task_number ].state = STATE_IDLE;
         }
         break;
   }
   spin_unlock_bh( &fp_drv_mcu_task_lock );
   FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "do_one_task_step: spin UNLOCK fp_drv_mcu_task_lock\n" );
}

/**********************************************************************
 * Function Name  : create_task
 * Description    : create task and put it in the queue of processing
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
int32_t create_task( enum drv_mcu_transfer_cmd_e cmd, uint32_t param, uint32_t *data ){
   int32_t index;

   index = 0;
   spin_lock_bh( &fp_drv_mcu_task_lock );
   do{
      if( drv_mcu_instance.task_array[ index ] == 0 ){
         if( ( cmd == SET_ETH_PHY_REG_I2C ) && ( data != FP_NULL ) ){
            drv_mcu_instance.task[ index ].data_ptr[ 0 ] = ( data[ 0 ] & 0xFFFF );
            FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "create_task: task data 0x%08x\n", drv_mcu_instance.task[ index ].data_ptr[ 0 ] );
         }
         drv_mcu_instance.task[ index ].task_cmd = ( param << 8 )|( cmd & 0xFF );
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "create_task: task cmd 0x%08x\n", drv_mcu_instance.task[ index ].task_cmd );
         drv_mcu_instance.task_array[ index ] = cmd;
         break;
      }
      index++;
   }while( index < TASKS_QUEUE_SIZE );
   spin_unlock_bh( &fp_drv_mcu_task_lock );

   if( index == TASKS_QUEUE_SIZE ){
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "create_task: task queue is full!\n" );
      return -1;
   }
   return index;
}

/**********************************************************************
 * Function Name  : drv_mcu_tasks_dispatcher
 * Description    : handler of the tasks
 * Inputs         :
 * Parameters     : 
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void drv_mcu_tasks_dispatcher( unsigned long data ){
   uint32_t i, result, tmp, index, port, timeout;

   result = 0;
   //timeout  = DRV_MCU_TASKS_TIME_STEP_100MS;
   timeout  = DRV_MCU_TASKS_TIME_STEP_10MS;
   switch( drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] ){

      case GET_FW_ID:
         /* get firmware number */
         timeout  = DRV_MCU_TASKS_TIME_STEP_50MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case GET_FW_ID\n" );
         start_task( drv_mcu_instance.active_task_index, GET_FW_ID );
         do_one_task_step( drv_mcu_instance.active_task_index /*, GET_FW_ID*/ );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               i = drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ];
               set_mcu_firmware_version( i );
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: MCU firmware: %d.%d\n", ( i & 0xFF ), ( ( i >> 8 ) & 0xFF ) );
            }
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index /*, GET_FW_ID*/ );
            drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
            result = 1;
         }
         break;

      case GET_TEMP:
         /* get temperature */
         timeout  = DRV_MCU_TASKS_TIME_STEP_100MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case GET_TEMP\n" );
         start_task( drv_mcu_instance.active_task_index, GET_TEMP );
         do_one_task_step( drv_mcu_instance.active_task_index /*, GET_TEMP*/ );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               i = drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ];
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case GET_TEMP, task complete, data 0x%08x\n", i );
               set_temperature( i );     
            }
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index /*, GET_TEMP*/ );
            result = 1;
         }
         break;

      case GET_SFP_EVENTS:
         /* GET_SFP_EVENTS */
         timeout  = DRV_MCU_TASKS_TIME_STEP_50MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case GET_SFP_EVENTS\n" );
         start_task( drv_mcu_instance.active_task_index, GET_SFP_EVENTS );
         do_one_task_step( drv_mcu_instance.active_task_index /*, GET_SFP_EVENTS*/ );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               i = drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ];
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case GET_SFP_EVENTS, task complete, data 0x%08x\n", i );
               /* procesing of sfp events */
               for( port = 0; port < NUM_OF_OPTIC_PORTS; port++ ){
                  tmp = is_sfp_module_connected( port );
                  if(   ( i & ( 1 << port ) ) ||
                        ( ( !tmp ) && ( ( i >> ( port + DRV_MCU_GET_SFP_EVENTS_EVENT_PORT1_CONNECT_OFFSET ) ) & 1 ) )  ){

                     if( i & ( 1 << port ) ){
                        FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: port_%d SFP event occure\n", port );
                     }else{
                        FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: port_%d SFP update configuration\n", port );
                     }

                     if( i & ( 1 << ( port + DRV_MCU_GET_SFP_EVENTS_EVENT_PORT1_CONNECT_OFFSET ) ) ){
                        /* sfp module connected, need to read sfp data */
                        create_task( GET_SFP_DATA, ( ( 6 << 4 )|( port ) ), FP_NULL );
                     }else{
                        /* sfp module disconnected, need to reset data in driver */
                        sfp_disconnected( port );
                        FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: port_%d SFP disconnected\n", port );
                     }
                  }
               }
            }
            /* procesing of sfp done */
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index /*, GET_SFP_EVENTS*/ );
            result = 1;
         }        
         break;

      case GET_SFP_DATA:
         /* GET_SFP_DATA */
         timeout  = DRV_MCU_TASKS_TIME_STEP_50MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case GET_SFP_DATA\n" );
         start_task( drv_mcu_instance.active_task_index, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd );
         do_one_task_step( drv_mcu_instance.active_task_index /*, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd*/ );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               //address offset
               index = ( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd >> 16 ) & 0xFFFF;
               //port number
               port = ( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd >> 8 ) & 0x3;
               //size of data in bytes (1 - 7 bytes read)
               i = ( ( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd >> 12 ) & 0xF ) + 1;

               sfp_store_data( port, index, i, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr );
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case GET_SFP_DATA, port_%d, read %d bytes at address %d: 0x%08x 0x%08x\n", port, i, index, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ], drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 1 ] );

               if( ( index + i ) < SIZE_OF_SFP_READ_DATA_IN_BYTES ){
                  // new index
                  index = index + i;
                  // new size of data in bytes (1 - 7 bytes)
                  if( ( index + 7 ) <= SIZE_OF_SFP_READ_DATA_IN_BYTES ){
                     i = 6;   //read 7 bytes
                  }else{
                     i = ( SIZE_OF_SFP_READ_DATA_IN_BYTES - index ) - 1;
                  }
                  drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd = ( index << 16 )|( i << 12 )|( port << 8 )|( GET_SFP_DATA );
                  FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case GET_SFP_DATA, task update, active_task_index %d , offset %d, size %d\n", drv_mcu_instance.active_task_index, index, i );
               }else{
                  /* SIZE_OF_SFP_READ_DATA_IN_BYTES read */
                  sfp_module_connected( port );
                  drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
                  result = 1;
                  FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: port_%d SFP connected\n", port );
                  FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case GET_SFP_DATA, task complete, active_task_index %d\n", drv_mcu_instance.active_task_index );
               }
            }else{
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: task complete with error, return cmd code 0x%08X\n", drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ] );
            }
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index /*, 0*/ );
            FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case GET_SFP_DATA, drv_mcu_instance.task[ %d ].state = %d\n", drv_mcu_instance.active_task_index, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].state );
         }
         break;

      case SET_PHY_RESET:
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case SET_PHY_RESET, cancel this action\n" );
         drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
         result = 1;
         break;

      case TEST_CMD:
         timeout  = DRV_MCU_TASKS_TIME_STEP_100MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case TEST_CMD\n" );
         start_task( drv_mcu_instance.active_task_index, TEST_CMD );
         do_one_task_step( drv_mcu_instance.active_task_index /*, TEST_CMD*/ );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               i = drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ];
               tmp = drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 1 ];
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: case TEST_CMD, task complete, data1_reg 0x%08x, data2_reg 0x%08x\n", i, tmp );
               if( !( ( i == TEST_CMD_DATA1_REG_RESULT ) && ( ( tmp & 0xFFFFFF00 ) == TEST_CMD_DATA2_REG_RESULT ) ) ){
                  /* unavailable exchange between DRIVER and MCU */
                  fp_drv_mcu_ops_init_timer( 0 );
                  fp_drv_mcu_tasks_init_timer( 0 );
                  return;
               }
            }
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index /*, TEST_CMD*/ );
            drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
            result = 1;
         }else{
            drv_mcu_instance.steps_counter++;
            if( drv_mcu_instance.steps_counter == DRV_TASK_MAX_STEPS_LIMIT ){
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: it is an old firmware on the MCU, SFP modules polling disable\n" );
               set_mcu_firmware_version( OLD_MCU_FIRMWARE_VERSION );
               fp_drv_mcu_ops_init_timer( 0 );
               fp_drv_mcu_tasks_init_timer( 0 );
               return;
            }
         }
         break;

      case GET_ETH_PHY_REG_I2C:
         timeout  = DRV_MCU_TASKS_TIME_STEP_10MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case GET_ETH_PHY_REG_I2C\n" );
         
         start_task( drv_mcu_instance.active_task_index, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd );
         do_one_task_step( drv_mcu_instance.active_task_index );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            port = ( ( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd & 0xFF00 ) << 8 );
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: task complete successfully, rd data 0x%08X\n", drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ] );
               get_eth_phy_reg_i2c_callback( ( uint16_t )( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ] ) | ( port ) );
            }else{
               get_eth_phy_reg_i2c_callback( port | 0xFF000000 );
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: task complete with error, return cmd code 0x%08X\n", drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ] );
            }
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index );
            drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
            result = 1;
         }
         break;

      case SET_ETH_PHY_REG_I2C:
         timeout  = DRV_MCU_TASKS_TIME_STEP_10MS;
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: case SET_ETH_PHY_REG_I2C\n" );
         start_task( drv_mcu_instance.active_task_index, drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd );
         do_one_task_step( drv_mcu_instance.active_task_index );
         if( !is_task_complete( drv_mcu_instance.active_task_index ) ){
            port = ( ( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].task_cmd & 0xFF00 ) << 8 );
            if( drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].result == DRV_TASK_RESEULT_COMPLETE_SUCCESS ){
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "drv_mcu_tasks_dispatcher: task complete successfully\n" );
               set_eth_phy_reg_i2c_callback( port );
            }else{
               set_eth_phy_reg_i2c_callback( port | 0xFF000000 );
               FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "drv_mcu_tasks_dispatcher: task complete with error, return cmd code 0x%08X\n", drv_mcu_instance.task[ drv_mcu_instance.active_task_index ].data_ptr[ 0 ] );
            }
            clear_task( drv_mcu_instance.active_task_index );
            do_one_task_step( drv_mcu_instance.active_task_index );
            drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
            result = 1;
         }
         break;

      default:
         drv_mcu_instance.task_array[ drv_mcu_instance.active_task_index ] = 0;
         result = 1;
         break;
   }

   if( result ){ 
      /* if last command complete, go to processing the next command */
      drv_mcu_instance.active_task_index++;
      if( drv_mcu_instance.active_task_index == TASKS_QUEUE_SIZE ){
         drv_mcu_instance.active_task_index = 0;
      }
   }

   fp_drv_mcu_tasks_timer.expires  = jiffies + timeout;
   add_timer( ( struct timer_list * )&fp_drv_mcu_tasks_timer );
}

/**********************************************************************
 * Function Name  : fp_drv_mcu_ops_init_timer
 * Description    : do driver to mcu data transfers
 * Inputs         :
 * Parameters     : init (1 - enable; 0 - disable);
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
uint32_t fp_drv_mcu_tasks_init_timer( uint32_t init ){
   if ( !init ){
      init_drv_mcu_all_tasks_struct( 0 );
      del_timer( ( struct timer_list * )&fp_drv_mcu_tasks_timer );
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "Delete fp_drv_mcu_tasks_timer\n");
   }else{
      if( init_drv_mcu_all_tasks_struct( 1 ) ){
         FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "Init fp_drv_mcu_tasks_init_timer FAILED!\n" );
         return -1;
      }
      init_timer( &fp_drv_mcu_tasks_timer );
      fp_drv_mcu_tasks_timer.expires  = jiffies + DRV_MCU_TASKS_TIME_STEP_100MS;
      fp_drv_mcu_tasks_timer.data     = ( unsigned long )&fp_drv_mcu_tasks_timer;
      fp_drv_mcu_tasks_timer.function = drv_mcu_tasks_dispatcher;
      add_timer( ( struct timer_list * )&fp_drv_mcu_tasks_timer );
      FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "Init fp_drv_mcu_tasks_init_timer PASS\n" );
   }
   return 0;
}


