/*************************************************************************/ /*!
*  @File
*  @Title          fp_mac_ops.c
*  @Copyright      Copyright (c) Milandr. All Rights Reserved
*  @Description    Initializes EAC paramaters
*  @License        Dual MIT/GPLv2
*
*  The contents of this file are subject to the MIT license as set out below.
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  The above copyright notice and this permission notice shall be included in
*  all copies or substantial portions of the Software.
*
*  Alternatively, the contents of this file may be used under the terms of
*  the GNU General Public License Version 2 ("GPL") in which case the provisions
*  of GPL are applicable instead of those above.
*
*  If you wish to allow use of your version of this file only under the terms of
*  GPL, and not to allow others to use your version of this file under the terms
*  of the MIT license, indicate your decision by deleting the provisions above
*  and replace them with the notice and other provisions required by GPL as set
*  out in the file called "GPL-COPYING" included in this distribution. If you do
*  not delete the provisions above, a recipient may use your version of this file
*  under the terms of either the MIT license or GPL.
*
*  This License is also included in this distribution in the file called
*  "MIT-COPYING".
*
*  EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
*  PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
*  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
*  PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
*  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
*  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
*  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
*                                      Include Files
**********************************************************************/
#include <linux/timer.h>
#include <linux/netdevice.h>
#include <linux/slab.h>

#include "../include/fp_library.h"
#include "../include/fp_global_addr_map.h"

#include "fp_mdio_ops.h"
#include "fp_netd.h"
#include "fp_net_driver.h"
#include "fp_stats.h"
#include "fp_mac_ops.h"
#include "fp_drv_mcu.h"
#include "fp_drv_mcu_tasks.h"

/**********************************************************************
*                                      #defines
**********************************************************************/

#define CURRENT_MAC_MODE_SETTINGS_REFRESH    (HZ >> 2)

/**********************************************************************
 *                              Global Declarations
 ***********************************************************************/
//uint32_t sfp_info_hw_addres[ NUM_OF_OPTIC_PORTS ] = { SFP1_INFO_ADDRESS, SFP2_INFO_ADDRESS, SFP3_INFO_ADDRESS, SFP4_INFO_ADDRESS };

fp_stats_t fp_stats;
spinlock_t fp_mac_ops_lock;

/* The timer structure used to see current EMAC mode (speed and duplex) */
static struct timer_list fp_mac_ops_timer;

uint32_t get_mac_register(uint32_t emac_offset, uint32_t reg_offset, uint32_t *result)
{
   uint32_t emac_addr;

   if (fp_baseAddr == NULL)
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
      return(0xFFFFFFFF);
   }

   emac_addr = emac_offset & 0xFF0000;

   if (!((emac_addr > 0x650000) && (emac_addr < 0x770000)))
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: bad emac_offset. 0x%X\n", __func__, emac_addr);
      return(0xFFFFFFFE);
   }

   if (emac_addr == 0x700000)
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: bad emac_offset. 0x%X\n", __func__, emac_addr);
      return(0xFFFFFFFE);
   }

   if (reg_offset > 0x7FC)
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: bad reg_offset. 0x%X\n", __func__, reg_offset);
      return(0xFFFFFFFD);
   }

   *result = fp_reg_read(emac_addr + reg_offset);

   return(0);
}

uint32_t set_mac_register(uint32_t emac_offset, uint32_t reg_offset, uint32_t value)
{
   if (fp_baseAddr == NULL)
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
      return(0xFFFFFFFF);
   }

   if (!((emac_offset > (0x660000 - 1)) && (emac_offset < (0x760000 + 1))))
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: bad emac_offset. 0x%X\n", __func__, emac_offset);
      return(0xFFFFFFFE);
   }

   if (emac_offset == 0x700000)
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: bad emac_offset. 0x%X\n", __func__, emac_offset);
      return(0xFFFFFFFE);
   }

   if (reg_offset > 0x7FC)
   {
      FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: bad emac reg_offset. 0x%X\n", __func__, reg_offset);
      return(0xFFFFFFFD);
   }

   fp_reg_write(emac_offset + reg_offset, value);

   return(0);
}

void restart_sgmii_autoneg( uint32_t emac_num ){
   uint32_t tmp;
   //FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s EMAC_%d restart sgmii auto-negotiation!\n", __func__, emac_num );
   if( get_mac_register( emac[ emac_num ], EMAC_PCS_CONTROL, &tmp ) ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ALERT, "%s can not read EMAC_%d register!\n", __func__, emac_num );
   }

   tmp |= EMAC_PCS_CONTROL_RESTART_AUTONEG_BIT;

   if( set_mac_register( emac[ emac_num ], EMAC_PCS_CONTROL, tmp ) ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ALERT, "%s can not read EMAC_%d register!\n", __func__, emac_num );
   }
}

/* Configures EMAC with new link speed */
void fp_mac_events_handler( uint32_t emac_num ){
   uint32_t events, duplex, speed, link, val, emac_class_config;
   uint64_t bytes;
   uint64_t *stat_field;
   fp_dev_info_t  *logical_priv;
   fp_private_t   *real_priv;
   struct port_s  *port;

   if( fp_logical_dev[ emac_num ] == FP_NULL ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_EMERG, "fp_mac_events_handler: the device isn't OPEN!\n" );
      return;
   }

   logical_priv = netdev_priv( fp_logical_dev[ emac_num ] );

   port = &fp_port[ emac_num ];
   if( port == FP_NULL ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_EMERG, "fp_mac_events_handler: struct fp_port not configured!\n" );
      return;  
   }

   if( get_mac_register( emac[ emac_num ], EMAC_INT_STATUS, &events ) ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "%s FAILED!\n", __func__ );
      return;
   }

   if( events ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "EMAC_%d 0x%X events: 0x%x\n", ( emac_num + 1 ), emac[ emac_num ], events );
   }

   /*-------------------------------*/
   /*--- processing SGMII events ---*/
   /*-------------------------------*/
   if( port->optic_link.pres == OPTIC_SFP_MODULE ){
   //SETTINGS FOR OPTIC SFP MODULE
      if( events & EMAC_INT_STATUS_LINK_CHANGE_BIT ){
         if( get_mac_register( emac[ emac_num ], EMAC_NETWORK_STATUS, &val ) ){
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "%s Can not read EMAC_%d register EMAC_NETWORK_STATUS!\n", __func__, ( emac_num + 1 ) );
            val = 0;
         }
         if( val & EMAC_NETWORK_STATUS_PCS_LINK_STATE_BIT ){
            if( !logical_priv->link_is_up ){
               FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X OPTIC_SFP link up\n", ( emac_num + 1 ), emac[ emac_num ] );
               FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X OPTIC_SFP speed 1000M, full duplex\n", ( emac_num + 1 ), emac[ emac_num ] );
            }
            logical_priv->link_is_up = 1;
            logical_priv->link_full_duplex = 1;
            logical_priv->link_speed = 1000;
         }else{
            if( logical_priv->link_is_up ){
               FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X OPTIC_SFP link down\n", ( emac_num + 1 ), emac[ emac_num ] );
            }
            logical_priv->link_is_up = 0;
            logical_priv->link_speed = 10;
            logical_priv->link_full_duplex = 0;
         }
         if( get_mac_register( emac[ emac_num ], EMAC_NETWORK_CONFIG, &val ) ){
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "%s Can not read EMAC_%d register EMAC_NETWORK_CONFIG!\n", __func__, ( emac_num + 1 ) );
         }else{
            val |= ( EMAC_NETWORK_CONFIG_FULL_DUPLEX_BIT | EMAC_NETWORK_CONFIG_GIGABIT_MODE_ENABLE_BIT );
            if( set_mac_register( emac[ emac_num ], EMAC_NETWORK_CONFIG, val ) ){
               FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "%s EMAC_%d can not write to register EMAC_NETWORK_CONFIG!\n", __func__, ( emac_num + 1 ) );
            }
         }
      }
   //SETTINGS FOR SFP MODULE END
   }else{
   //SETTINGS FOR SGMII
      if( ( events & EMAC_INT_STATUS_PCS_AUTONEG_COMPLETE_BIT ) && ( events & EMAC_INT_STATUS_LINK_CHANGE_BIT ) ){

         if (get_mac_register( emac[ emac_num ], EMAC_PCS_AN_LP_BASE, &val ) ){     // read link, duplex, speed from EMAC register (pcs_an_lp_base), do needed settings in EMAC
            FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "%s Can not read EMAC_%d register 0x214!\n", __func__, emac_num);
         }
         duplex = (val & (1 << 12)) >> 12;                                      // 0 - MDI half duplex; 1 - MDI full duplex
         speed  = (val & (3 << 10)) >> 10;                                      // 0 - MDI 10M, 1 - MDI 100M, 2 - MDI 1000M, 3 - reserved
         link   = (val & (1 << 15)) >> 15;                                      // 0 - MDI link down, 1 - MDI link up
         if( link ){
            logical_priv->link_is_up = 1;
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X %s link up\n", ( emac_num + 1 ), emac[ emac_num ],
                        ( port->optic_link.pres == COPPER_SFP_MODULE )?"COPPER_SFP":"COPPER_SGMII" );
         }else{
            logical_priv->link_is_up = 0;
            speed = 2;
            duplex = 1;
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X %s link down\n", ( emac_num + 1 ), emac[ emac_num ],
                        ( port->optic_link.pres == COPPER_SFP_MODULE )?"COPPER_SFP":"COPPER_SGMII" );
         }
         if( get_mac_register( emac[ emac_num ], EMAC_NETWORK_CONFIG, &val ) ){
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "%s Can not read EMAC_%d 0x%X register 0x04!\n", __func__, ( emac_num + 1 ), emac[ emac_num ] );
         }
         val &= ~( ( 1 << 10 )|( 3 << 0 ) );
         if( duplex ){
            val |= ( 1 << 1 );
            logical_priv->link_full_duplex = 1;
            if( link ){
               FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X %s set full duplex\n", ( emac_num + 1 ), emac[ emac_num ],
                           ( port->optic_link.pres == COPPER_SFP_MODULE )?"COPPER_SFP":"COPPER_SGMII" );
            }
         }else{
            logical_priv->link_full_duplex = 0;
            if( link ){
               FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X %s set half duplex\n", ( emac_num + 1 ), emac[ emac_num ],
                           ( port->optic_link.pres == COPPER_SFP_MODULE )?"COPPER_SFP":"COPPER_SGMII" );
            }
         }
         switch( speed ){
         case 0:  //10M
            logical_priv->link_speed = 10;
            break;
         case 1:  //100M
            val |= 1;
            logical_priv->link_speed = 100;
            break;
         case 2:
            val |= (1 << 10);
            logical_priv->link_speed = 1000;
            break;
         default:
            break;
         }
         if( link ){
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "EMAC_%d 0x%X %s set speed %dM\n", ( emac_num + 1 ), emac[ emac_num ],
                        ( port->optic_link.pres == COPPER_SFP_MODULE )?"COPPER_SFP":"COPPER_SGMII", logical_priv->link_speed );
         }
         if( set_mac_register( emac[ emac_num ], 0x04, val ) ){
            FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "EMAC_%d 0x%X can not write to register 0x04!\n", ( emac_num + 1 ), emac[ emac_num ] );
         }
         /*************************/
         /* FLOW CONTROL SETTINGS */
         /*************************/
         if( link ){   //if port up change flow control parameters
            //port = &fp_port[ emac_num ];
            if( port->fc.setup.update_flag ){ //update parameters of flow control
               FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "update flow control parameters of Port_%d\n", (emac_num + 1) );
               port->fc.setup.update_flag = 0;
               val = phy_dp83867_is_flow_control_available( logical_priv->mdio_port, logical_priv->phy_addr );
               FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: is flow control at Port_%d available: %d\n", ( emac_num + 1 ), val );
               if( (!val) && ( port->fc.setup.enable ) ){   //if flow control unsupported of link partner and flow control enable
                  if(!port->fc.setup.force){
                        FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: flow control at Port_%d unsupported\n", ( emac_num + 1 ) );
                  }else{   //force enable flow control if full duplex mode
                     link = phy_dp83867_get_duplex( logical_priv->mdio_port, logical_priv->phy_addr );
                     if( link == 4){ //if full duplex
                        if( !flow_control_enable( emac_num, &(port->fc) ) ){
                           FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: enable flow control forced at Port_%d, preferred_mode %d\n", ( emac_num + 1 ), port->fc.preferred_mode );
                        }else{
                           FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "fp_mac_events_handler: can not enable flow control forced at Port_%d\n", ( emac_num + 1 ) );
                        }
                     }else{
                        FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: Port_%d in half duplex, flow control unsupported\n", ( emac_num + 1 ) );
                        port->fc.setup.force = 0;
                     }
                  }
               }else if( port->fc.setup.enable ){  //if flow control supported and flow control enable
                  port->fc.preferred_mode = val;
                  if( !flow_control_enable( emac_num, &(port->fc) ) ){
                     FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: enable flow control at Port_%d, mode %d\n", ( emac_num + 1 ), port->fc.preferred_mode );
                  }else{
                     FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "fp_mac_events_handler: can not enable flow control at Port_%d! mode %d\n", ( emac_num + 1 ), val );
                  }
                  switch( val ){
                     /*
                        0 - flow control not supported by Local Host or Link Partner
                        1 - pause flow control supported by Local Host and Link Partner
                        2 - asymmectrical pause flow control supported by Local Host (tx only)
                        3 - asymmectrical pause flow control supported by Local Host (rx only)
                     */
                     case 0:
                        port->fc.preferred_mode = 0;      
                        break;
                     case 1:
                        port->fc.preferred_mode = 3;
                        break;
                     case 2:
                        port->fc.preferred_mode = 2;
                        break;
                     case 3:
                        port->fc.preferred_mode = 1;
                        break;
                  }
               }
               //else{
               //   FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: flow control at Port_%d disable\n", ( emac_num + 1 ) );
               //}
            }
         }else{   //if port down
            if( !flow_control_disable( emac_num ) ){
               port = &fp_port[emac_num];
               port->fc.setup.update_flag = 1;
               //FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "fp_mac_events_handler: disable flow control at Port_%d\n", ( emac_num + 1 ) );
            }else{
               FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "fp_mac_events_handler: can not disable flow control at Port_%d!\n", ( emac_num + 1 ) );
            }
         }
      }
      /* END OF FLOW CONTROL SETTINGS */
   }
      
   

   /*--------------------------*/
   /*--- get MAC statistics ---*/
   /*--------------------------*/
   get_mac_register(emac[emac_num], 0x108, &val);
   logical_priv->phy_stats.tx_ok      += val;
   logical_priv->dev_stats.tx_packets += val;
   get_mac_register(emac[emac_num], 0x10C, &val);
   logical_priv->phy_stats.tx_broadcast += val;
   get_mac_register(emac[emac_num], 0x110, &val);
   logical_priv->phy_stats.tx_multicast += val;
   get_mac_register(emac[emac_num], 0x114, &val);
   logical_priv->phy_stats.tx_pause += val;
   get_mac_register(emac[emac_num], 0x118, &val);
   logical_priv->phy_stats.tx_64 += val;
   get_mac_register(emac[emac_num], 0x11C, &val);
   logical_priv->phy_stats.tx_65_127 += val;
   get_mac_register(emac[emac_num], 0x120, &val);
   logical_priv->phy_stats.tx_128_255 += val;
   get_mac_register(emac[emac_num], 0x124, &val);
   logical_priv->phy_stats.tx_256_511 += val;
   get_mac_register(emac[emac_num], 0x128, &val);
   logical_priv->phy_stats.tx_512_1023 += val;
   get_mac_register(emac[emac_num], 0x12C, &val);
   logical_priv->phy_stats.tx_1024_1518 += val;
   get_mac_register(emac[emac_num], 0x130, &val);
   logical_priv->phy_stats.tx_more_1518 += val;
   get_mac_register(emac[emac_num], 0x134, &val);
   logical_priv->phy_stats.tx_underruns += val;
   get_mac_register(emac[emac_num], 0x138, &val);
   logical_priv->phy_stats.tx_single_collision += val;
   get_mac_register(emac[emac_num], 0x13C, &val);
   logical_priv->phy_stats.tx_2_15_collision += val;
   get_mac_register(emac[emac_num], 0x140, &val);
   logical_priv->phy_stats.tx_more_15_collision += val;
   get_mac_register(emac[emac_num], 0x144, &val);
   logical_priv->phy_stats.tx_late_collision += val;
   get_mac_register(emac[emac_num], 0x148, &val);
   logical_priv->phy_stats.tx_deferred += val;
   logical_priv->dev_stats.collisions   = logical_priv->phy_stats.tx_late_collision
                                          + logical_priv->phy_stats.tx_more_15_collision
                                          + logical_priv->phy_stats.tx_2_15_collision
                                          + logical_priv->phy_stats.tx_single_collision;

   get_mac_register(emac[emac_num], 0x158, &val);
   logical_priv->phy_stats.rx_ok      += val;
   logical_priv->dev_stats.rx_packets += val;
   get_mac_register(emac[emac_num], 0x15C, &val);
   logical_priv->phy_stats.rx_broadcast += val;
   get_mac_register(emac[emac_num], 0x160, &val);
   logical_priv->phy_stats.rx_multicast += val;
   get_mac_register(emac[emac_num], 0x164, &val);
   logical_priv->phy_stats.rx_pause += val;
   get_mac_register(emac[emac_num], 0x168, &val);
   logical_priv->phy_stats.rx_64 += val;
   get_mac_register(emac[emac_num], 0x16C, &val);
   logical_priv->phy_stats.rx_65_127 += val;
   get_mac_register(emac[emac_num], 0x170, &val);
   logical_priv->phy_stats.rx_128_255 += val;
   get_mac_register(emac[emac_num], 0x174, &val);
   logical_priv->phy_stats.rx_256_511 += val;
   get_mac_register(emac[emac_num], 0x178, &val);
   logical_priv->phy_stats.rx_512_1023 += val;
   get_mac_register(emac[emac_num], 0x17C, &val);
   logical_priv->phy_stats.rx_1024_1518 += val;
   get_mac_register(emac[emac_num], 0x180, &val);
   logical_priv->phy_stats.rx_more_1518 += val;
   get_mac_register(emac[emac_num], 0x184, &val);
   logical_priv->phy_stats.rx_less_64 += val;
   get_mac_register(emac[emac_num], 0x188, &val);
   logical_priv->phy_stats.rx_excessive_len += val;
   get_mac_register(emac[emac_num], 0x18C, &val);
   logical_priv->phy_stats.rx_jabbers += val;
   get_mac_register(emac[emac_num], 0x190, &val);
   logical_priv->phy_stats.rx_fcs_err += val;
   get_mac_register(emac[emac_num], 0x194, &val);
   logical_priv->phy_stats.rx_len_err += val;
   get_mac_register(emac[emac_num], 0x198, &val);
   logical_priv->phy_stats.rx_symbol_err += val;
   get_mac_register(emac[emac_num], 0x19C, &val);
   logical_priv->phy_stats.rx_aligment_err += val;
   get_mac_register(emac[emac_num], 0x1A0, &val);
   logical_priv->phy_stats.rx_recource_err += val;
   get_mac_register(emac[emac_num], 0x1A4, &val);
   logical_priv->phy_stats.rx_overruns    += val;
   logical_priv->dev_stats.rx_over_errors += val;
   get_mac_register(emac[emac_num], 0x1A8, &val);
   logical_priv->phy_stats.rx_ip_crc_err += val;
   get_mac_register(emac[emac_num], 0x1AC, &val);
   logical_priv->phy_stats.rx_tcp_crc_err += val;
   get_mac_register(emac[emac_num], 0x1B0, &val);
   logical_priv->phy_stats.rx_udp_crc_err += val;

   get_mac_register(emac[emac_num], 0x100, &val);
   bytes = val;
   get_mac_register(emac[emac_num], 0x104, &val);
   bytes |= ((u64)val & 0xFFFF) << 32;
   logical_priv->dev_stats.tx_bytes += bytes;

   get_mac_register(emac[emac_num], 0x150, &val);
   bytes = val;
   get_mac_register(emac[emac_num], 0x154, &val);
   bytes |= ((u64)val & 0xFFFF) << 32;
   logical_priv->dev_stats.rx_bytes += bytes;

   bytes = fp_reg_read(egpi[emac_num] + 0x6c);
   logical_priv->dev_stats.rx_dropped += bytes;

   emac_class_config = fp_reg_read(WSP_GLOBAL_BASE_ADDR + 0x44);
   stat_field        = &logical_priv->phy_stats.class_rx;

   for (val = 0; val < 10; val++)
   {
      if (emac_class_config & (1 << emac_num))                                  // CLASS_HW2
      {
         *stat_field += fp_reg_read(CLASS2_HW_BASE_ADDR + hwPhyCounters[emac_num][val]);
      }
      else                                                                      // CLASS_HW1
      {
         *stat_field += fp_reg_read(CLASS_HW_BASE_ADDR + hwPhyCounters[emac_num][val]);
      }
      stat_field++;
   }

   if (emac_num == 0)
   {
      /* Get data for HIF */
      real_priv  = netdev_priv(logical_priv->real_dev);
      stat_field = &real_priv->switch_stats.class_rx;
      for (val = 0; val < 10; val++)
      {
         if (emac_class_config & (1 << 16))                                     // CLASS_HW2
         {
            *stat_field += fp_reg_read(CLASS2_HW_BASE_ADDR + hwPhyCounters[16][val]);
         }
         else                                                                   // CLASS_HW1
         {
            *stat_field += fp_reg_read(CLASS_HW_BASE_ADDR + hwPhyCounters[16][val]);
         }
         stat_field++;
      }

      real_priv->switch_stats.bmu1_free = fp_reg_read(BMU1_BASE_ADDR + 0x48);
      real_priv->switch_stats.bmu2_free = fp_reg_read(BMU2_BASE_ADDR + 0x48);
      real_priv->netstats.rx_dropped   += fp_reg_read(HGPI_BASE_ADDR + 0x6c);

      /* Get temperature through MDIO <-- for old version MCU firmware */
      val = get_mcu_firmware_version();
      if( ( val == OLD_MCU_FIRMWARE_VERSION ) ||
          ( val == 0 ) ) {
        
      	fp_mdio_write_phy( MDIO_PORT_1, 0x1F, 6 << 2, 0 );
	     	val = fp_mdio_read_phy( MDIO_PORT_1, 0x1F, 7 << 2 );
   
         if( val != 0xFFFFFFFF ){
            set_temperature( val );
         }
      }
   }
}

/**********************************************************************
 * Function Name  : fp_mac_ops_init_timer
 * Description    : do EMAC speed and duplex settings
 * Inputs         :
 * Parameters     : init (1 - enable; 0 - disable);
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
void fp_mac_ops_event_handler(unsigned long data)
{
   int32_t        i /*, index, j*/;
   fp_dev_info_t *logical_priv;
   fp_stats_t     new_stats;

   if( fp_logical_dev[ 0 ] == FP_NULL ){
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ALERT, "fp_mac_ops_event_handler: the device isn't OPEN!\n" );
      return;
   }

   memset( &new_stats, 0, sizeof(fp_stats_t) );
   spin_lock_bh( &fp_mac_ops_lock );

   /*---------------------------------------------------------*/
   /*----- Update ports statistics and speed, duplex, link ---*/
   /*---------------------------------------------------------*/
   for( i = 0; i < EMAC_NUMS; i++ ){
      fp_mac_events_handler( i );

      logical_priv            = netdev_priv( fp_logical_dev[ i ] );                 // Get data for cli stats
      new_stats.ip4_pkt_cnt  += logical_priv->phy_stats.class_ipv4;
      new_stats.udp_pkt_cnt  += logical_priv->phy_stats.class_udp;
      new_stats.tcp_pkt_cnt  += logical_priv->phy_stats.class_tcp;
      new_stats.icmp_pkt_cnt += logical_priv->phy_stats.class_icmp;
      new_stats.igmp_pkt_cnt += logical_priv->phy_stats.class_igmp;
      new_stats.ip6_pkt_cnt  += logical_priv->phy_stats.class_ipv6;

      new_stats.ip_hdr_len_err_cnt += logical_priv->phy_stats.class_l3_fail;
      new_stats.ip_pkt_ttl_err_cnt += logical_priv->phy_stats.class_ttl_err;
      new_stats.ip_chksum_err_cnt  += logical_priv->phy_stats.class_crc_err;

      new_stats.rx_pkt_cnt += logical_priv->phy_stats.class_rx;
   }
   memcpy( &fp_stats, &new_stats, sizeof( fp_stats_t ) );
   spin_unlock_bh( &fp_mac_ops_lock );

   //test_host_unexpected_stop();

   /* Update the refresh timer to call this fucntion again */
   fp_mac_ops_timer.expires = jiffies + CURRENT_MAC_MODE_SETTINGS_REFRESH;
   add_timer( &fp_mac_ops_timer );
}

/**********************************************************************
 * Function Name  : fp_mac_ops_init_timer
 * Description    : do EMAC speed and duplex settings
 * Inputs         :
 * Parameters     : init (1 - enable; 0 - disable);
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int fp_mac_ops_init_timer(int init)
{
   if ( !init ){
      del_timer( (struct timer_list *)&fp_mac_ops_timer );
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_EMERG, "Delete fp_mac_ops_timer\n" );
      return 0;
   }else{
      spin_lock_init( &fp_mac_ops_lock );
      init_timer( &fp_mac_ops_timer );
      fp_mac_ops_timer.expires  = jiffies + CURRENT_MAC_MODE_SETTINGS_REFRESH;
      fp_mac_ops_timer.data     = ( unsigned long )&fp_mac_ops_timer;
      fp_mac_ops_timer.function = fp_mac_ops_event_handler;
      add_timer( ( struct timer_list *)&fp_mac_ops_timer );
      FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_EMERG, "Init fp_mac_ops_timer\n" );
      return 0;
   }
}
