/*************************************************************************/ /*!
@File
@Title          fp_switch_init.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    Initializes ETHERNET PHY paramaters
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/
#include <linux/netdevice.h>


#include <fp_types.h>
#include "../include/fp_library.h"
#include "../include/fp_global_addr_map.h"
#include "fp_mdio_ops.h"

/*															#defines															*/
 #define MDIO_RD                                     2
 #define MDIO_WR                                     1

 /* 			                                              WRITE0 |  WRITE1 |    OPERATION  |      PHY_ADDR    | REG_ADDR         | WRITE10 | DATA */
 #define MAKE_WR_DATA( PHY_ADDR, REG_ADDR, DATA )		 ((0 << 31)|(1 << 30)|(MDIO_WR << 28)|((PHY_ADDR) << 23)|((REG_ADDR) << 18)|(2 << 16)|(DATA))
 #define MAKE_RD_DATA( PHY_ADDR, REG_ADDR )	         ((0 << 31)|(1 << 30)|(MDIO_RD << 28)|((PHY_ADDR) << 23)|((REG_ADDR) << 18)|(2 << 16)|(0x00))

#define I347AT4_PAGE_SEL	                           22
/**********************************************************************
 *                              Global Declarations
***********************************************************************/

/**********************************************************************
 *                              Extern Declarations
***********************************************************************/
extern UCHAR  *fp_baseAddr;

/******************************************************************
 * Function Name  : fp_mdio_port_enable
 * Description    : Enable or disable SMI interface (MDIO interface)
 * Inputs         :
 * Parameters     : mdio_port_number - number of MDIO port
 *		  : (0 - MDIO1 port (EMAC1),  1 - MDIO2 port (MDIO2))
 * 		  : action - action to do (0 - disable MDIO port,
 *		  : 1 - enable MDIO port)
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void fp_mdio_port_enable(uint32_t mdio_port_number, uint32_t action)
{
	volatile uint32_t *addr;
	uint32_t val;

	if(fp_baseAddr == NULL){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
		return;
	}

	mdio_port_number &= 1;
	action &= 1;

	if(!mdio_port_number){	//emac1
		addr = (uint32_t*)(EMAC1_BASE_ADDR + fp_baseAddr);
	}else{
		addr = (uint32_t*)(EMAC2_BASE_ADDR + fp_baseAddr);
	}

	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"base addr of emac: 0x%p\n", addr);

	val = *addr;
	if(!action){	//disable
		val &= (~0x10);	//disable mdio port
	}
	else{
		val |= 0x10;	//enable mdio port
	}
	*addr = val;

	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"write value 0x%08X\n", val);
}

/******************************************************************
 * Function Name  : fp_mdio_clock_settings
 * Description    : Define MDC frequency of SMI interface (MDIO interface)
 * Inputs         :
 * Parameters     : mdio_port_number - number of MDIO port
 *		  : (0 - MDIO1 port (EMAC1),  1 - MDIO2 port (MDIO2))
 * 		  : divider - divider (range: 0 to 7)
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void fp_mdio_clock_settings(uint32_t mdio_port_number, uint32_t divider)
{
	volatile uint32_t *addr;
	uint32_t val;

	if(fp_baseAddr == NULL){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
		return;
	}

	mdio_port_number &= 1;
	divider &= 7;

	if(!mdio_port_number){	//emac1
		addr = (uint32_t*)(EMAC1_BASE_ADDR + 4 + fp_baseAddr);
	}else{
		addr = (uint32_t*)(EMAC2_BASE_ADDR + 4 + fp_baseAddr);
	}

	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"%s base addr of emac: 0x%p\n", __func__, addr);

	val = *addr;
	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"%s read from network_config reg: 0x%08X\n", __func__, val);
	val = (val & (~(7 << 18)) );
	val |= (divider << 18);
	*addr = val;
	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"%s write to network_config reg: 0x%08X\n", __func__, val);
	//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"MDC clock is 200/%d MHz\n", div[divider] );
}
/******************************************************************
 * Function Name  : mdio_write_start
 * Description    : start mdio write transaction (MDIO interface)
 * Inputs         :
 * Parameters     : emac_base_addr - base address of EMAC with MDIO port
 * 		  : value - code-word for to write in PHY_MANAGEMENT register
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void mdio_write_start(uint32_t emac_base_addr, uint32_t value)
{
	volatile uint32_t *addr;
	uint32_t val;

	if(fp_baseAddr == NULL){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
		return;
	}

	addr = (uint32_t*)(emac_base_addr + 0x34 + fp_baseAddr);
	*addr = value;

	addr = (uint32_t*)(emac_base_addr + 0x8 + fp_baseAddr);
	do{
		val = *addr;
	}while((val & 4) == 0);
}

/******************************************************************
 * Function Name  : mdio_read_start
 * Description    : start mdio read transaction (MDIO interface)
 * Inputs         :
 * Parameters     : emac_base_addr - base address of EMAC with MDIO port
 * 		  : value - code-word for to write in PHY_MANAGEMENT register
 * Outputs        :
 * Parameters     : -
 * Returns        : read data form ethernet PHY
 ********************************************************************/
uint16_t mdio_read_start(uint32_t emac_base_addr, uint32_t value)
{
	volatile uint32_t *addr;
	uint32_t val;

	if(fp_baseAddr == NULL){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,
			"%s: pci_mem base is NULL\n", __func__);
		return 0xFFFF;
	}

	addr = (uint32_t*)(emac_base_addr + 0x34 + fp_baseAddr);
	*addr = value;

	addr = (uint32_t*)(emac_base_addr + 0x8 + fp_baseAddr);
	do{
		val = *addr;
	}while((val & 4) == 0);

	addr = (uint32_t*)(emac_base_addr + 0x34 + fp_baseAddr);
	val = (*addr & 0xFFFF);
	return val;
}


/******************************************************************
 * Function Name  : mdio_write_phy
 * Description    : write data to ethernet PHY register via MDIO interface
 * Inputs         :
 * Parameters     : mdio_port - number of MDIO port (0 - mdio port of EMAC1, 1 - mdio port of EMAC2)
 *                : phy_addr - address of ethernet PHY on the MDIO bus
 *                : reg_addr - write data to this register
 * 		  : value - data to write
 * Outputs        :
 * Parameters     : -
 * Returns        : -
 ********************************************************************/
void fp_mdio_write_phy(uint32_t mdio_port, uint32_t phy_addr, uint32_t reg_addr, uint32_t value)
{
	uint32_t val, emac_base;

	if(!mdio_port){
		emac_base = EMAC1_BASE_ADDR;
	}else{
		emac_base = EMAC2_BASE_ADDR;
	}

	phy_addr &= 0x1F;
	reg_addr &= 0x1F;
	value &= 0xFFFF;

	// FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"MDIO_PORT%d, PHY ADDR %d, REG_ADDR %d, value 0x%04X, emac_base 0x%06X\n", mdio_port, phy_addr, reg_addr, value, emac_base);

	val = MAKE_WR_DATA( phy_addr, reg_addr, value );
	mdio_write_start(emac_base, val);
}

/******************************************************************
 * Function Name  : mdio_read_phy
 * Description    : read data from ethernet PHY register via MDIO interface
 * Inputs         :
 * Parameters     : mdio_port - number of MDIO port (0 - mdio port of EMAC1, 1 - mdio port of EMAC2)
 *                : phy_addr - address of ethernet PHY on the MDIO bus
 *                : reg_addr - read data from this register
 * Outputs        :
 * Parameters     : -
 * Returns        : read data form ethernet PHY register
 ********************************************************************/
uint16_t fp_mdio_read_phy(uint32_t mdio_port, uint32_t phy_addr, uint32_t reg_addr)
{
	uint32_t val, emac_base;

	if(!mdio_port){
		emac_base = EMAC1_BASE_ADDR;
	}else{
		emac_base = EMAC2_BASE_ADDR;
	}

	phy_addr &= 0x1F;
	reg_addr &= 0x1F;

	val = MAKE_RD_DATA( phy_addr, reg_addr );
	val = mdio_read_start(emac_base, val);

  // FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO,"MDIO_PORT%d, PHY ADDR %d, REG_ADDR %d, value 0x%04X, emac_base 0x%06X\n", mdio_port, phy_addr, reg_addr, val, emac_base);

	return (0xFFFF & val);
}


/******************************************************************
 * Function Name  : mdio_read_phy_ID
 * Description    : read ethernet PHY ID code via MDIO interface
 * Inputs         :
 * Parameters     : mdio_port - number of MDIO port (0 - mdio port of EMAC1, 1 - mdio port of EMAC2)
 *                : phy_addr - address of ethernet PHY on the MDIO bus
 * Outputs        :
 * Parameters     : -
 * Returns        : read ID code registers from ethernet PHY
 ********************************************************************/
uint32_t fp_mdio_read_phy_id(uint32_t mdio_port, uint32_t phy_addr){
	uint32_t result;

	fp_mdio_write_phy(mdio_port, phy_addr, I347AT4_PAGE_SEL, 0);

	result = fp_mdio_read_phy(mdio_port, phy_addr, 2);
	result = (result & 0xFFFF) << 16;

	result |= fp_mdio_read_phy(mdio_port, phy_addr, 3);

	return result;
}
