/*************************************************************************/ /*!
@File
@Title          fp_net_driver.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    network driver for wsp module
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                     Include Files
 **********************************************************************/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/if_vlan.h>

#include <fp_library.h>
#include <fp_types.h>
#include <fp_ioctl.h>
#include <fp_netd.h>
#include <fpios.h>
#include "fp_net_driver.h"
#include "fp_bd_api.h"
#include <fp_common_msg_header.h>
#include <fp_common_ctlm.h>
#include <tlite_blocks_ctrl_api.h>
#include <tlite_blocks_ctrlm.h>
#include <switchApi_blocks_ctrlm.h>
#include "fp_hif_reg.h"
#include "fp_hw_util.h"
#include "fp_os_shim.h"

// #include "fp_mdio_ops.h"
#include "fp_phy_setup.h"
#include "fp_mac_ops.h"
#include "fp_drv_mcu_tasks.h"

#include "phy_i347at4.h"
#include "phy_dp83867.h"

#include "fp_global_addr_map.h"
#include "sfp_copper_setup.h"

#ifdef HAVE_PTP_1588_CLOCK
#include <gem_ptp.h>
#endif
/**********************************************************************
 *                              Global Declarations                   *
 **********************************************************************/
 #define MAX_MDIO_PHY_ADDR                8
/* Buffers for Config and PCI memory */
char buff[256];
/* FPPHY: Devices for fp interfaces */
int phyAdrMdio1[ MAX_MDIO_PHY_ADDR ] = { 0, 1, 2,  3,  4,  5,  6,  7 };
int phyAdrMdio2[ MAX_MDIO_PHY_ADDR ] = { 8, 9, 10, 11, 12, 13, 14, 15 };
/* extern fp_dev_info_t  *fp_logical_dev[FP_MAX_PORTS]; */
/* pcm_dev_list_t dev_list[4];// added for tmu */
tlite_dev_list_t fp_tdev_list[TLITE_SWITCH_MAX_DEV_NUM];/* added for tlite */
extern const uint32_t emac[ EMAC_NUMS ];
/**********************************************************************
 *                              Extern Declarations                   *
 **********************************************************************/
 struct fp_net_stats_ethtool
 {
 	char stat_string[ETH_GSTRING_LEN];
 	int stat_offset;
 };

 #define FP_NET_STAT(_name, _stat) { \
 	.stat_string = _name, \
 	.stat_offset = offsetof(struct switch_device_stats, _stat) \
 }

 /* All stats are u64, so we don't need to track the size of the field. */
 static const struct fp_net_stats_ethtool fp_switch_stats_strings[] = {
     FP_NET_STAT("BMU1 mem free",        bmu1_free),
     FP_NET_STAT("BMU2 mem free",        bmu2_free),

     FP_NET_STAT("class received",       class_rx),
     FP_NET_STAT("class L3 fail",        class_l3_fail),
     FP_NET_STAT("class IPV4",           class_ipv4),
     FP_NET_STAT("class IPV6",           class_ipv6),
     FP_NET_STAT("class CRC error",      class_crc_err),
     FP_NET_STAT("class TTL error",      class_ttl_err),
     FP_NET_STAT("class ICMP",           class_icmp),
     FP_NET_STAT("class IGMP",           class_igmp),
     FP_NET_STAT("class TCP",            class_tcp),
     FP_NET_STAT("class UDP",            class_udp),
 };

 #define FP_NET_GLOBAL_STATS_LEN ARRAY_SIZE(fp_switch_stats_strings)
/**********************************************************************
 * Function Name  : getPciBaseAddr
 * Description    : get the pci base address of pci device
 * Inputs
 *   Parameters   : struct net_device *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
char *getPciBaseAddr(void *dev)
{
	fp_private_t *fp = netdev_priv((struct net_device *)dev);
	return (char *)fp->pcimem_base;
}

/**********************************************************************
 * Function Name  : fp_net_open
 * Description    : Fastpath init routine for network driver, once device
 *                  is detected this routine is called immeadiately.
 *                  initialize rx ring buffer descriptors, initialize tx
 *                  ring buffer descriptors, loading fastpath dump, loading
 *                  fastpath global data structures, creating network device,
 *                  initialize network device handlers, irq registration,
 *                  register fp_phy interfaces, etc..
 * Inputs
 *   Parameters   : struct net_device *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
int fp_net_open(struct net_device *dev)
{
	fp_private_t *fp = netdev_priv(dev);
	struct pci_dev *pdev = fp->pci_dev;
    fp_dev_info_t *logical_priv;
	int rc = 0, i = 0;
	
	ASSERT(pdev);

	fp_init_fastpath(dev);

	rc = request_irq(pdev->irq, fp_interrupt, IRQF_SHARED, dev->name, dev);
	if (rc < 0) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "Interrupt registration got failed\n");
		return rc;
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: %d rc %d name %s dev %p\n", __func__,  pdev->irq, rc, dev->name, dev);

	spin_lock_bh(&fp->lock);
	dev->irq = pdev->irq;
	dev->needed_headroom = 32;

	/* Allocate memory and initialize bd rings */
	rc = fp_bd_init(fp);

	if (rc < 0) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "error while initializiing bd table\n");
		spin_unlock_bh(&fp->lock);
		return rc;
	}

	fp_disable_interrupts(fp);
	spin_unlock_bh(&fp->lock);

	/* CLEAR - Program the bdp reg base addresses */
	fp_init_bdp_base_reg(fp);
	fp_enable_interrupts(fp);
	fp_en_hif_rxdma_engine(fp);
	fp_init_rx_tasklet(fp);
	fp_init_tx_tasklet(fp);

	/* Notify the stack of the actual queue counts. */
	rc = netif_set_real_num_tx_queues(dev, NPU_MAX_TX_QUEUES);
	if (rc) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "error setting number of tx queues\n");
		return rc;
	}

	netif_tx_start_all_queues(dev);

	/* create dynamic logcal interfaces based on num of phy ports */
	rc = fp_create_logical_interfaces(dev, FP_MAX_PORTS);
	if (rc < 0) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "fp_net_driver: error in interface creation\n");
		return 0;
	}
	for (i = 0; i < FP_MAX_PORTS; i++) {
		if (((FWD_PORT_LIST_MASK >> i) & 0x1) && (i != HIF_PORT_NUM)) {
			fp_logical_dev[i]->needed_headroom = 32;
		}
	}

//#ifdef HAVE_PTP_1588_CLOCK
//	rc = npu_ptp_init(dev);
//	if (rc < 0) {
//		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "fp_net_driver: npu_ptp_init failed\n");
//		return 0;
//	}
//#endif /* HAVE_PTP_1588_CLOCK */

	/* added for tlite */
	memset(fp_tdev_list, 0, (TLITE_SWITCH_MAX_DEV_NUM * sizeof(tlite_dev_list_t)) );

	pfe_tlite_init(dev);
	pfe_tlite_default_config();

	fp->ageing_timeout = 300;
	/* initialize ageing timer */
	fp_switch_init_timer(1);

	fp_mdio_port_enable(MDIO_PORT_1, 1);
	fp_mdio_port_enable(MDIO_PORT_2, 1);

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_net_open start init phy\n");
  	for(i = 0; i < MAX_MDIO_PHY_ADDR; i++)                                        // Our max PHY MDIO address = 7
	{
		fp_mdio_read_phy_id(MDIO_PORT_1, i);																				// dummy read from D6
		rc = fp_mdio_read_phy_id(MDIO_PORT_1, i);
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_net_open PHY adr = %d ID is 0x%08X\n", i, rc);

		if((rc == I347AT4_PHY_ID) && (phyAdrMdio1[i] != -1))
        {
			phy_i347at4_init(MDIO_PORT_1, i, 0);
            logical_priv = netdev_priv(fp_logical_dev[ phyAdrMdio1[ i ] ]);
            logical_priv->phy_ic_type = PHY_IC_I347AT4;
            logical_priv->phy_addr = i;
            logical_priv->mdio_port = MDIO_PORT_1;
        }
		else if((rc == DP83867_PHY_ID) && (phyAdrMdio1[i] != -1))
        {
			phy_dp83867_init(MDIO_PORT_1, i, 0);
            logical_priv = netdev_priv(fp_logical_dev[ phyAdrMdio1[ i ] ]);
            logical_priv->phy_ic_type = PHY_IC_DP83867;
            logical_priv->phy_addr = i;
            logical_priv->mdio_port = MDIO_PORT_1;
            //FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "ic type: 0x%x, phy addr: %d, mdio port %d\n\n", logical_priv->phy_ic_type, logical_priv->phy_addr, logical_priv->mdio_port);
        }
    }

	for(i = 0; i < MAX_MDIO_PHY_ADDR; i++)
	{
		fp_mdio_read_phy_id(MDIO_PORT_2, i);																					//dummy read from D6
		rc = fp_mdio_read_phy_id(MDIO_PORT_2, i);
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_net_open PHY adr = %d ID is 0x%08X\n", i, rc);

		if((rc == I347AT4_PHY_ID) && (phyAdrMdio2[i] != -1))
        {
			phy_i347at4_init(MDIO_PORT_2, i, 0);
            logical_priv = netdev_priv(fp_logical_dev[phyAdrMdio2[i]]);
            logical_priv->phy_ic_type = PHY_IC_I347AT4;
            logical_priv->phy_addr = i;
            logical_priv->mdio_port = MDIO_PORT_2;
        }
		else if((rc == DP83867_PHY_ID) && (phyAdrMdio2[i] != -1))
        {
			phy_dp83867_init(MDIO_PORT_2, i, 0);
            logical_priv = netdev_priv(fp_logical_dev[phyAdrMdio2[i]]);
            logical_priv->phy_ic_type = PHY_IC_DP83867;
            logical_priv->phy_addr = i;
            logical_priv->mdio_port = MDIO_PORT_2;
            //FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "ic type: 0x%x, phy addr: %d, mdio port %d\n\n", logical_priv->phy_ic_type, logical_priv->phy_addr, logical_priv->mdio_port);
        }
	}

	fp_sfp_copper_setup_thread_init( );
	fp_mac_ops_init_timer( 1 );
	fp_drv_mcu_ops_init_timer( 1 );
	fp_drv_mcu_tasks_init_timer( 1 );

	spin_lock_bh( &fp->lock );
	fp->config_flags |= SWITCH_CONFIG_DONE;
	fp->txq_stopped = 0;
	spin_unlock_bh( &fp->lock );

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_net_open -end\n");
	fp_trace_level = 6;
	return 0;
}

/**********************************************************************
 * Function Name  : fp_net_release
 * Description    : unregister the logical fp_phy interfaces and free the
 *                  ring buffer descriptors
 * Inputs
 *   Parameters   : struct net_device *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
int fp_net_release(struct net_device *dev)
{
	fp_private_t *fp;
	unsigned int portno = 0;
	UINT ch_index = 0;
	uint32_t value;

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_net_release called\n");

	fp_sfp_copper_setup_thread_deinit( );
	fp_drv_mcu_ops_init_timer( 0 );
	fp_drv_mcu_tasks_init_timer( 0 );
	fp_mac_ops_init_timer( 0 );

	for( ch_index = 0; ch_index < NUM_OF_OPTIC_PORTS; ch_index++ ){
		if( fp_port[ ch_index ].optic_link.module_descr ){
			kfree( fp_port[ ch_index ].optic_link.module_descr );
			fp_port[ ch_index ].optic_link.module_descr = FP_NULL;
		}
	}

#if 0
#ifdef CONFIG_HOST_LEARNING
	fp_cleanup_fastpath();
	fp_switch_init_timer(0);
#endif
#endif

	fp = netdev_priv(dev);

	fp_switch_init_timer(0);

	for( portno = 0; portno < EMAC_NUMS; portno++ ){
		fp_reg_write( emac[ portno ], EMAC_NETWORK_CONTROL_TX_ENABLE );
	}

	fp_disable_hif_tx_dma_engine(fp);
	while( fp_reg_read( ( BMU1_BASE_ADDR + BMU_BUF_CNT ) ) & 0xFFFF ){
		//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "waiting for BMU1 count %u\n" , *addr);
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "waiting for BMU1 count\n" );
		mdelay( 100 );
	}

	while( fp_reg_read( ( BMU2_BASE_ADDR + BMU_BUF_CNT ) ) & 0xFFFF ){
		//FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "waiting for BMU2 count %u\n", *addr);
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "waiting for BMU2 count\n" );
		mdelay( 100 );
	}

	mdelay( 10 );
	fp_disable_hif_rx_dma_engine(fp);

	/* soft reset start */
	value = fp_reg_read( ( WSP_GLOBAL_BASE_ADDR + WSP_SYS_GENERIC_CONTROL ) );
	fp_reg_write( ( WSP_GLOBAL_BASE_ADDR + WSP_SYS_GENERIC_CONTROL ), ( value | 0x40000000 ) );

	mdelay( 100 );

	fp_reg_write( ( WSP_GLOBAL_BASE_ADDR + WSP_SYS_GENERIC_CONTROL ), value );
	FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "WSP_SOFT_RESET DONE\n" );

	for( portno = 0; portno < FP_MAX_PORTS; portno++ ){
		if( ( ( FWD_PORT_LIST_MASK >> portno ) & 0x1 ) && ( portno != HIF_PORT_NUM ) ){
			netif_stop_queue( fp_logical_dev[ portno ] );
		}
	}

	netif_tx_stop_all_queues( dev );
	free_irq( dev->irq, dev );

	/* Cleanup the bd rings */
	spin_lock_bh( &fp->lock );

	for( ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++ ){
		dma_free_coherent( &fp->pci_dev->dev, fp_total_ring_size, fp->hif.ch[ ch_index ].txring->cpu_addr, fp->hif.ch[ ch_index ].txring->dev_addr );
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "DMA Free Coherent\n");

	spin_unlock_bh( &fp->lock );

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "unregister logical interfaces %p %p\n", fp_logical_dev[ 0 ], fp_logical_dev[ 1 ]);

	for( portno = 0; portno < FP_MAX_PORTS; portno++ ){
		if( ( ( FWD_PORT_LIST_MASK >> portno ) & 0x1 ) && ( portno != HIF_PORT_NUM ) ){
			unregister_nfp_device( fp_logical_dev[ portno ] );
		}
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "unregister logical interfaces %d\n", portno);
	}

	spin_lock_bh( &fp->lock );
	fp->config_flags &= ( ~SWITCH_CONFIG_DONE );
	spin_unlock_bh( &fp->lock );
#if 0
#ifdef HAVE_PTP_1588_CLOCK
	npu_ptp_stop();
#endif /* HAVE_PTP_1588_CLOCK */
#endif

	return 0;
}

/**********************************************************************
 * Function Name  : fp_net_ioctl
 * Description    : fp device ioctl interface for userspace, user can
 *                  read/write the all pci mem, read/write the PE config space,
 *                  display statastics, read the hardware debug registers to
 *		    user space, read the PE route entry info, will provide
 *			complete statstics abou PE at hardwrae level to user.
 * Inputs
 *   Parameters   : struct net_device *, struct ifreq *, int
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
int
fp_net_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	unsigned int i;
	volatile unsigned int *addr_mem;
	fp_private_t *priv = netdev_priv(dev);
	unsigned char *addr = (unsigned char *)priv->pcimem_base;
	struct pci_dev *pdev = priv->pci_dev;
	fpathcmd_t fpcmd;
	fpathcmd_t *pfpcmd = (fpathcmd_t *)rq->ifr_ifru.ifru_data;
	unsigned int *pBuff = (unsigned int *)&buff[0];
	volatile unsigned int *test_addr;
	unsigned int test_data1 = 0x0;
	volatile unsigned int test_data2 = 0x0;
	unsigned char *byte_addr;
	volatile unsigned char byte_data1;
	volatile unsigned char byte_data2;
	UINT check = 0;
	int result = 0;


	/* copy the user space request to kernel space */
	check = copy_from_user(&fpcmd, /* to */
			pfpcmd, /* from */
			sizeof(fpcmd)); /* how many bytes */

    /* convert the addr to pci address region */
   if((cmd == FPATH_READ_PHY_MEM) || (cmd == FPATH_WRITE_PHY_MEM ))
       addr_mem = (volatile unsigned int *) phys_to_virt(fpcmd.addr);
   else
   {
		if( (fpcmd.addr >= 0x800000)&&(fpcmd.addr < 0x880000) )
			addr_mem = (volatile unsigned int *) ((unsigned char *)priv->pcimem_base_bar4 + fpcmd.addr - 0x800000);
		else
			addr_mem = (volatile unsigned int *) (addr + fpcmd.addr);
   }

#if 0
	if (fpcmd.addr & 0x80000000) {

		addr_mem = (volatile unsigned int *) (fpcmd.addr);
	} else {
		/* convert the addr to pci address region */
		addr_mem = (volatile unsigned int *) (addr + fpcmd.addr);
	}
#endif

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI mem address:%p cmd: %x\n", addr_mem, cmd);

	switch (cmd) {

		case FPATH_READ_CONFIG:
			/* copy_to_user the config space of PCI */
			for (i = 0; i < fpcmd.size; i++) {
				pci_read_config_byte(pdev, (fpcmd.addr + i), &buff[i]);
				FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "\n addr:%lx value:%x", (fpcmd.addr+i), buff[i]&0xff);
			}
			result = copy_to_user(pfpcmd->Buff, &buff[0], fpcmd.size);
			break;

		case FPATH_WRITE_CONFIG:
			pci_write_config_byte(pdev, fpcmd.addr, fpcmd.value);
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, " addr:%lx value:%x", fpcmd.addr, fpcmd.value);
			break;

		case FPATH_READ_PCI_MEM:
			if (fpcmd.size > 128) {
				FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "Too many bytes to read %d(max 128)\n", fpcmd.size);
				break;
			}
			/* copy the PCI memory space to user space */
			for (i = 0; i < (fpcmd.size/4); i++) {
				pBuff[i] = (unsigned int) addr_mem[i];
				FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "\n addr:%p value:%x", &addr_mem[i], pBuff[i]);
			}
			result = copy_to_user(pfpcmd->Buff, pBuff, (fpcmd.size));
			break;

	    case FPATH_READ_PHY_MEM:
	        if (fpcmd.size > 128) {
	            printk("Too many bytes to read %d(max 128)\n", fpcmd.size);
	            break;
	        }
	            /* copy the PCI memory space to user space */
	        for(i =0; i< (fpcmd.size/4); i++){
	            pBuff[i] = (unsigned int) addr_mem[i];
	            printk("\n READ_PHY_MEM addr:%p value:%x",&addr_mem[i],pBuff[i]);
	        }
	        result = copy_to_user(pfpcmd->Buff, /* to */ pBuff, /* from */ (fpcmd.size) ); /* how many bytes */
	        break;


		case FPATH_WRITE_PCI_MEM:
			/* write userspace data to PCI memory space */
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "\n write value :%x at address %p\n", fpcmd.value, addr_mem);
			*addr_mem = fpcmd.value;
			break;

	    case FPATH_WRITE_PHY_MEM:
            /* write userspace data to PCI memory space */
#ifdef DEBUG_PCI
            printk("\n write value :%x at address %p\n", fpcmd.value,addr_mem);
#endif
            *addr_mem = fpcmd.value;
#ifdef DEBUG_PCI
            printk("User Write Data to PCI Mem Space: 0x%p, written: 0x%x, value read back :0x%x\n", addr_mem, test_data1, test_data2);
#endif
            break;

	case FPATH_TX_PCI_DMA:
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "TX PCI DMA :\n");
		test_addr  = (unsigned int *)(0xa00000 + priv->pcimem_base);

		/* Test to see the DDR functionality.. */
		test_data1 = 0x01020304;
		*test_addr = test_data1;

		test_addr  = (unsigned int *)(0xa00000 + priv->pcimem_base);
		test_data2 = (unsigned int) *test_addr;

		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "\n value:%x addr:%p\n", test_data1, test_addr);

		if (test_data2 == 0x01020304) {
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "Basic HW Rd/WR Passed : %x\n", test_data2);
		} else {
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "Basic HW Rd/WR Failed : %x\n", test_data2);
		}
	break;

	case FPATH_RX_PCI_DMA:
		byte_data1 = 0xa5;
		byte_addr  = (unsigned char *)(0xa03001 + priv->pcimem_base);
		*byte_addr = byte_data1;
		byte_data2 = (unsigned char) *byte_addr;
	break;

#if 0
	case PCM_MSG_TMU:
#ifdef TMU_EN
	pcm_tmu_config_ioctl(dev, rq, cmd);
#endif
#endif
	break;
	case PCM_MSG_TLITE:
		if( pcm_tlite_config_ioctl(dev, rq, cmd) ){
			result = -EFAULT;
		}
	break;

	case PCM_MSG_SWITCH:
		if( pcm_switchApi_config_ioctl(dev, rq, cmd) ){
			result = -EFAULT;
		}
	break;

	default:
		result = fp_config_ioctl(dev, rq, cmd);
		break;
	}

	return result;
}

/**********************************************************************
 * Function Name  : fp_get_stats
 * Description    : unregister the logical fp_phy interfaces and free the
 * Inputs
 *   Parameters   : struct net_device *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : struct net_device_stats *
 * Changes        :
 *********************************************************************/
static struct net_device_stats *fp_get_stats(struct net_device *dev)
{
	fp_private_t *fp;

	fp = netdev_priv(dev);
	ASSERT(fp);

	return &fp->netstats;
}

unsigned short fp_select_queue(struct net_device *dev, struct sk_buff *skb, void *accel_priv, select_queue_fallback_t fallback)
{
	USHORT cos = 0;
	UINT hif_queue_map = 0;
	USHORT ch_index = 0;
	fp_private_t *fp = netdev_priv((struct net_device *)dev);

	if( skb_vlan_tag_present( skb ) ){
		cos = ( ( skb_vlan_tag_get( skb ) ) & VLAN_PRIO_MASK ) >> VLAN_PRIO_SHIFT;
		hif_queue_map = fp_reg_read( ( ULONG )&fp->hif.regs->hif_rx_queue_map_ch_no_addr );
		ch_index = ( hif_queue_map >> ( cos << 2 ) ) & HIF_QUEUE_MAP_VALID_MASK;

		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "TX: COS %u, protocol %x, channel %u\n", cos, ntohs( skb->protocol ), ch_index );
	}
	return ch_index;
}

/* ---------------------------- ETHTOOL Support ---------------------------- */
static void
fp_net_get_drvinfo(struct net_device *netdev, struct ethtool_drvinfo *drvinfo)
{
    fp_private_t *fp = netdev_priv(netdev);
    uint8_t string[32];
    uint32_t len;

    FP_DEBUG( DBGP_FEAT_ETHTOOL, FP_LOG_INFO, "Get driver info from fp0" );
    memset( string, 0, 32 );
	len = snprintf( string, 32, "%d.%d", FP_MAJOR_DRIVER_VERSION, FP_MINOR_DRIVER_VERSION );
	strlcpy( drvinfo->driver, "Milandr 1923KX028 driver", 25 );
	strlcpy( drvinfo->version, string, ( len + 1 ) );
	len = snprintf( string, 32, "%d.%d", fp->mcu_firmware_version & 0xFF, ( fp->mcu_firmware_version >> 8 ) & 0xFF );
	strlcpy( drvinfo->fw_version, string, ( len + 1 ) );
	strlcpy( drvinfo->bus_info, pci_name(fp->pci_dev), 32 );

    drvinfo->n_stats = FP_NET_GLOBAL_STATS_LEN;
}

static void
fp_net_get_strings(struct net_device *dev, u32 stringset, u8 *data)
{
	int i;

    FP_DEBUG(DBGP_FEAT_ETHTOOL, FP_LOG_INFO, "Get switch strings, stringset=%d", stringset);

	switch (stringset) {
	case ETH_SS_STATS:
		for (i = 0; i < FP_NET_GLOBAL_STATS_LEN; i++)
        {
			memcpy(data, fp_switch_stats_strings[i].stat_string, ETH_GSTRING_LEN);
			data += ETH_GSTRING_LEN;
		}
		break;
	}
}

static int
fp_net_get_sset_count(struct net_device *dev, int sset)
{
    FP_DEBUG(DBGP_FEAT_ETHTOOL, FP_LOG_INFO, "Get switch sset count, sset=%d", sset);

	switch (sset) {
	case ETH_SS_STATS:
		return FP_NET_GLOBAL_STATS_LEN;
	default:
		return -EOPNOTSUPP;
	}
}

/**
 * fp_get_ethtool_stats - report device statistics
 * @netdev: network interface device structure
 * @stats: ethtool statistics structure
 * @data: pointer to data buffer
 *
 * All statistics are added to the data buffer as an array of u64.
 **/
static void
fp_net_get_ethtool_stats(struct net_device *netdev, struct ethtool_stats *stats, u64 *data)
{
    fp_private_t *adapter = netdev_priv(netdev);
    struct switch_device_stats *switchstats = &adapter->switch_stats;
    unsigned int i;
	char *p;

    FP_DEBUG(DBGP_FEAT_ETHTOOL, FP_LOG_INFO, "Get statistics from switch" );

    for (i = 0; i <FP_NET_GLOBAL_STATS_LEN; i++) {
        p = (char *)switchstats + fp_switch_stats_strings[i].stat_offset;
        data[i] =  *(u64 *)p;
    }
}

const struct net_device_ops fp_netdev_ops = {
	.ndo_start_xmit		= fp_net_xmit,
	.ndo_open			= fp_net_open,
	.ndo_select_queue	= fp_select_queue,
	.ndo_do_ioctl		= fp_net_ioctl,
	.ndo_get_stats		= fp_get_stats,
	.ndo_change_mtu		= NULL
};

struct ethtool_ops fp_ethtool_ops = {
    .get_drvinfo        = fp_net_get_drvinfo,
    .get_strings        = fp_net_get_strings,
    .get_sset_count     = fp_net_get_sset_count,
    .get_ethtool_stats	= fp_net_get_ethtool_stats,
};


void set_mcu_firmware_version( uint32_t value ){
	fp_private_t *fp;
	if( fp_net_dev != FP_NULL ){
		fp = netdev_priv( ( struct net_device *)fp_net_dev );
		fp->mcu_firmware_version = value; //value[7:0] - major number of version; value[15:8] - minor number of version
	}
}


uint32_t get_mcu_firmware_version( void ){
	fp_private_t *fp;
	uint32_t tmp = 0xFFFFFFFF;
	if( fp_net_dev != FP_NULL ){
		fp = netdev_priv( ( struct net_device *)fp_net_dev );
		tmp = fp->mcu_firmware_version; //value[7:0] - major number of version; value[15:8] - minor number of version	
	}
	return tmp;
}

void set_temperature( uint32_t value ){
	fp_private_t *fp;
	if( fp_net_dev != FP_NULL ){
		fp = netdev_priv( ( struct net_device *)fp_net_dev );
		fp->switch_stats.basisTemp = value & 0xFF; 			//value[7:0] - 1923 temp
		fp->switch_stats.boardTemp = (value >> 8 ) & 0xFF;	//value[7:0] - board temp
	}
}

uint32_t get_temperature( void ){
	fp_private_t *fp;
	uint32_t temp = 0xFFFFFFFF;
	if( fp_net_dev != FP_NULL ){
		fp = netdev_priv( ( struct net_device *)fp_net_dev );
		temp = fp->switch_stats.basisTemp | ( fp->switch_stats.boardTemp << 16 );
	}
	return temp;
}


