/*************************************************************************/ /*!
@File
@Title          fp_pci_bm.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    contais interrupt handlers and pci bus master tx/rx routines
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/etherdevice.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/if_vlan.h>

#include <fp_library.h>
#include <fp_macros.h>
#include <fp_types.h>
#include <fp_ioctl.h>
#include <fp_pre_header.h>
#include <hw_vlan_hash_table.h>
#include <hw_2field_hash_table.h>
#include "fp_net_driver.h"
#include <fp_netd.h>
#include "fp_bd_api.h"
#include "fp_hif_reg.h"
#include "fp_os_shim.h"
#include "fp_ltc.h"
#include <fp_switch.h>

static struct tasklet_data fp_rx_info[NUM_HIF_CHANNELS];
static struct tasklet_data fp_tx_info[NUM_HIF_CHANNELS];
static inline void fp_disable_all_ch_interrupts(fp_private_t *fp);

/**********************************************************************
 * Function Name  : fp_rx_tasklet
 * Description    : tasklet handler scheduled under bottom-half context,
 *                  this handler will called when everr tasklet is scheduled
 *                  to process rx packets, copy the packets from rx bd ring
 *                  to the kernel space as sk_buff's, deliver the packet
 *                  fp_phy interfaces based on phyno of the packet received,
 *                  extract necessary data from packet start data to update in
 *                  hooks, from now packet will processed as a regular path
 *                  handle by network stack
 * Inputs
 *   Parameters   : unsigned long - fast path private data structure
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
void
fp_rx_tasklet(ulong data)
{
	struct tasklet_data *rxptr = (struct tasklet_data *)data;
	fp_private_t *fp = (fp_private_t *)rxptr->fp;
	struct bd_ring *rxring;
	struct bd *rxbd;
	struct sk_buff *skb = NULL;
	UCHAR *ptr;
	UINT len = 0, curr_bd_index = 0;
	UINT rx_ctrl = 0, wb_buflen = 0;
	UINT ch_index = rxptr->ch_index;
	INT ret = 0;
	uint8_t jumbo = 0;			//jumbo packet flag
	uint8_t jumbo_last = 0;		//jumbo packet last segment
	uint32_t jumbo_offset = 0;	//processed jumbo packet bytes
	uint16_t jumbo_segment = 0;	//number of segment of cuurent processing jumbo frame

	static UINT rx_count;

	volatile struct hif_regs *regs = fp->hif.regs;

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "RX Tasklet: ch_index %d\n", ch_index);

	rxring = fp->hif.ch[ch_index].rxring;

	if (rxring == NULL) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "%s: null rxring ptr\n", __func__);
		return;
	}

	spin_lock_bh(&fp->rx_lock);

	while( 1 ){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "RX Count: %u\n", ++rx_count);
		rxbd = fp_deque_bd(rxring, &wb_buflen);
		if (rxbd == NULL) {
			FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_DEBUG, "RX wb bd not ready\n");
			break;
		}

		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "bd_ctrl: 0x%04X\n", rxbd->bd_ctrl );
		rx_ctrl = ( wb_buflen >> 20 ) & 0x7FF;
		/* allocate memory for sk_buff */
		if( rx_ctrl & BD_CTRL_LIFM ){
			//packet in one buffer
			if( !jumbo ){
				skb = dev_alloc_skb( MAX_FRAME_SIZE );
				FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "Rcv standard packet\n" );
			}else{
				jumbo_last = 1;
				FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "Rcv last buffer of JUMBO\n" );
			}
		}else{
			//packet in several buffers
			if( !jumbo ){
				FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "Rcv JUMBO packet\n" );
				skb = dev_alloc_skb( MAX_JUMBO_FRAME_SIZE );
				jumbo = 1;
			}
		}

		if (skb == NULL) {
			FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "Out of memory allocating skb RX buffer\n");
			break;
		}

		/* Get the curr rx bd buf length */
		len = wb_buflen & 0xFFFF;
		wb_buflen = 0;

		/* Get the curr rx bd index */
		curr_bd_index = fp_get_curr_bd_index(rxring, rxbd);

		/* Get the virtual address of the frame */
		ptr = (UCHAR *)((unsigned long)rxring->bd_buff_va + (curr_bd_index * MAX_FRAME_SIZE));

		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "LEN: %u index %d rxbd %p rx_buf_ptr %p\n", len, curr_bd_index, rxbd, ptr );

		/* Fetching more will not hurt */
		if( jumbo ){
			memcpy( &skb->data[ jumbo_offset ], ptr, len);
			jumbo_offset += len;
			jumbo_segment++;
			FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "New %d bytes of JUMBO packet, offset %d, segment %d\n", len, jumbo_offset, jumbo_segment );
		}else{
			memcpy(skb->data, ptr, len);
		}
		skb_put(skb, len);
		skb->dev = fp->dev;
		fp->netstats.rx_bytes += skb->len;
		if( !jumbo || jumbo_last ){
			fp_process_rx_header( skb );
			fp->netstats.rx_packets++;
			jumbo = 0;
			jumbo_last = 0;
			jumbo_offset = 0;
			jumbo_segment = 0;
		}

		/* Enable per packet receive interrupt */
		rxbd->bd_ctrl &= ~( BD_CTRL_DESC_EN | BD_CTRL_PKT_INT_EN | BD_CTRL_DIR );

		rxbd->bd_buflen = MAX_FRAME_SIZE;
		rxbd->bd_ctrl |= BD_CTRL_PKT_INT_EN | BD_CTRL_DIR;

		/* enqueue the above rx free bd to hardware freebd list*/
		ret = fp_enque_bd(rxring, rxbd);
		if (ret < 0) {
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "rx bd not avail this should never happen\n");
			break;
		}
	} /* End of while */

	/* init the DMA */
	rx_ctrl = fp_reg_read((ULONG)&regs->ch[ch_index].hif_ctrl_ch);
	rx_ctrl |= HIF_CH_RX_CTRL_DMA_EN;
	fp_reg_write((ULONG)&regs->ch[ch_index].hif_ctrl_ch, rx_ctrl);

	rx_ctrl = fp_reg_read((ULONG)&regs->ch[ch_index].hif_rx_ch_start);
	rx_ctrl |= HIF_CH_RX_START;
	fp_reg_write((ULONG)&regs->ch[ch_index].hif_rx_ch_start, rx_ctrl);

	/* Re-Enable interrupts */
	fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_en, HIF_CH_INT_EN | HIF_CH_RXPKT_INT_EN |	HIF_CH_TXPKT_INT_EN);

	spin_unlock_bh(&fp->rx_lock);

	return;
}

/**********************************************************************
 * Function Name  : fp_net_xmit_
 * Description    : fastpath transmit routine call under softirq context,
 *                  write the every packet received by fp_phy0 and fp_ph1
 *                  logical interfaces to hw tx descripotrs, every packet
 *                  need to convert big-endian before writing into hardware,
 *                  write tx phyport and interface port as start of the packet
 * Inputs
 *   Parameters   : struct net_device *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
int32
fp_net_xmit(struct sk_buff *skb, struct net_device *dev)
{
	fp_private_t *fp;
	INT tx_ctrl, ret = 0;
	volatile struct hif_regs *regs;
	struct bd_ring *txring;
	struct bd *txbd;
	UINT /*pkt_len,*/ curr_bd_index = 0;
	UCHAR *ptr, *ptr1;
	struct tx_header txhdr;
	UINT ch_index = 0;
	char msrp_addr[3] = {0x01, 0x80, 0xc2};

	uint32_t total_pkt_len, bd_needed, bd_use_count;
	USHORT tmp_bd_ctrl;
#ifdef LTC_EN
	unsigned int cur_time = 0, set_time = CLAMP_TIME;
#endif
	fp = netdev_priv(dev);

	spin_lock_bh(&fp->lock);

	ch_index = skb_get_queue_mapping( skb );
	regs = fp->hif.regs;
	txring = fp->hif.ch[ch_index].txring;
	total_pkt_len = 0;
	bd_use_count = 0;
	bd_needed = ( skb->len + ( MAX_FRAME_SIZE - 1 ) ) / MAX_FRAME_SIZE;
	FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "Need %d txbd to send current packet\n", bd_needed );

	/* check for free tx bd avail or not */
	txbd = fp_get_next_free_bd( txring );

	if( ( bd_needed > 1 ) || ( skb->len > 1514 ) ){
		if( !( fp->config_flags & SWITCH_JUMBO_PROCESSING ) ){
			/*  We can't to send this packet, 
				because jumbo disable... return */
			dev_kfree_skb( skb );
			spin_unlock_bh( &fp->lock );
			FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_ERR, "%s: jumbo frame disable\n", __func__ );
			return NETDEV_TX_OK;
		}
		ret = fp_get_count_of_free_bd( txring );
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: free bd %d\n", __func__, ret );
		if( ret < bd_needed ){
			txbd = NULL;
		}
	}

	if( fp->txq_stopped == 1 ){
		//need to update switch settings; please wait
		txbd = NULL;
	}
	if( txbd == NULL ) {
		struct netdev_queue *queue = NULL;
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "Tx Queue %s\n", fp->txq_stopped ? "stopped" : "full" );
		queue = netdev_get_tx_queue( dev, ch_index );
		netif_tx_stop_queue( queue );
		spin_unlock_bh( &fp->lock );
		return NETDEV_TX_BUSY;
	}

	/* Add vlan tag in packet */
	if( !skb_vlan_tagged( skb ) ){
    	FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: insert vlan tag in packet\n", __func__ );
    	skb = vlan_insert_tag( skb, htons( ETH_P_8021Q ), ( fp_port[ HOST_PORT_NO ].fallback_bd_id )|( fp_port[ HOST_PORT_NO ].def_pri << VLAN_PRIO_SHIFT )  ); 
    	if( skb == NULL ){
			FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_EMERG, "%s: can not insert vlan tag in payload\n", __func__ );
			spin_unlock_bh( &fp->lock );
			return NETDEV_TX_OK;
		}
	}
	FP_DEBUG_PRINT_HEX_DUMP( DBGP_FEAT_HIF, FP_LOG_DEBUG, "TX_TO_NET ", DUMP_PREFIX_OFFSET, 16, 1, 
							skb->data, skb->len, true )
	while( bd_use_count < bd_needed ){
		/* get curr bd index */
		curr_bd_index = fp_get_curr_bd_index(txring, txbd);

		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "BD Index %d buff_va %p\n", curr_bd_index, txring->bd_buff_va);

		/* Get the virtual address of the frame */
		ptr1 = (UCHAR *)(txring->bd_buff_va);
		ptr =  ptr1 + (curr_bd_index * MAX_FRAME_SIZE);

		/* ptp packet headroom is failing.. copy both preheader and data
			* If the packet is ptp packet and timestamp is requested */
		if( unlikely( ntohs( skb->protocol ) == 0x88F7 ) ) {
//		if( ntohs( skb->protocol ) == 0x88F7 ) {
			if( !bd_use_count ){
				FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: protocol is: %x (PTP), len: %d\n", __func__, ntohs(skb->protocol), skb->len );
				fp_fill_tx_ptp_header( &txhdr, skb );
				memcpy( ptr, &txhdr, FP_TX_HDR_LEN );
				if( skb->len > ( MAX_FRAME_SIZE - FP_TX_HDR_LEN ) ){
					ret = MAX_FRAME_SIZE - FP_TX_HDR_LEN;
				}else{
					ret = skb->len;
				}
				memcpy( ptr + FP_TX_HDR_LEN, skb->data, ret );
			}else{
				if( ( skb->len - total_pkt_len ) > MAX_FRAME_SIZE ){
					ret = MAX_FRAME_SIZE;
				}else{
					ret = skb->len - total_pkt_len;
				}
				memcpy( ptr, &skb->data[ total_pkt_len ], ret );
			}
		} else {
			if( !bd_use_count ){
				FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: protocol is: %x, len: %d\n", __func__, ntohs( skb->protocol ), skb->len );
				/* Prepend the tx header to transmit packet */
				ret = fp_add_prepend_tx_header( skb );
				if( ret < 0 ) {
					FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_EMERG, "%s: not enough room to tx header in skb\n", __func__ );
					spin_unlock_bh( &fp->lock );
					return NETDEV_TX_OK;
				}
				if( memcmp( msrp_addr, skb->data + sizeof( struct tx_header ), 3 ) == 0 ) {
					int tx_port = skb->cb[ CB_PORT_FOR_INJECTION ];
					struct net_device *dev = NULL;
					struct tx_header *txhdr = ( struct tx_header* )skb->data;

					if( tx_port < FP_MAX_PORTS ){
						txhdr->txport_map = ( 1 << tx_port );
						dev = fp_phy_addr[ tx_port ];
					}else{
						/* We don't know this port,
							cannot handle..return */
						dev_kfree_skb( skb );
						spin_unlock_bh( &fp->lock );
						return NETDEV_TX_OK;
					}
					txhdr->ctrl |= FP_TX_PKT_INJECT_EN;
				}
#ifdef LTC_EN
				if( skb->lt_enable == LTC_ENABLE_MAGICNUM ){
					memset( &txhdr, 0, sizeof( struct tx_header ) );
					memcpy( ( char* )&txhdr, ( char* )skb->data, sizeof( struct tx_header ) );
					cur_time = fp_reg_read( fp_baseAddr, ( ULONG )LTC_TSU_TIME_SEC );

					if( ( int )( ( int )txhdr.launch_time_sec - ( int )cur_time ) > ( int )set_time ){
						txhdr.launch_time_sec = ( int )( ( int )cur_time + ( int )set_time );
					}
					memcpy( ( char* )skb->data, ( char* )&txhdr, sizeof( struct tx_header ) );
				}
#endif
				if( skb->len > ( MAX_FRAME_SIZE - FP_TX_HDR_LEN ) ){
					ret = MAX_FRAME_SIZE - FP_TX_HDR_LEN;
				}else{
					ret = skb->len;
				}
				memcpy( ptr, skb->data, ret );
			}else{
				if( ( skb->len - total_pkt_len ) > MAX_FRAME_SIZE ){
					ret = MAX_FRAME_SIZE;
				}else{
					ret = skb->len - total_pkt_len;
				}
				memcpy( ptr, &skb->data[ total_pkt_len ], ret );
			}
		}
		total_pkt_len += ret;
		bd_use_count++;
		txbd->bd_buflen = ret;
		tmp_bd_ctrl = txbd->bd_ctrl & ~( BD_CTRL_DESC_EN | BD_CTRL_LIFM );
		if( bd_use_count == bd_needed ){
			/* update the ctrl word necessary proper bd flags */
			txbd->bd_ctrl = tmp_bd_ctrl | ( BD_CTRL_PKT_INT_EN | BD_CTRL_LIFM );
		}else{
			txbd->bd_ctrl = ( BD_CTRL_PKT_INT_EN /*| BD_CTRL_CBD_INT_EN |*/ );
		}

		/* enqueue bd to hardware */
		ret = fp_enque_bd( txring, txbd );
		if (ret < 0) {
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "tx bd not avail this should never happen\n");
		}

		txbd = fp_get_next_free_bd( txring );
	}
	dev_kfree_skb( skb );

	tx_ctrl = fp_reg_read((ULONG)&regs->ch[ch_index].hif_ctrl_ch);
	tx_ctrl |= HIF_CH_TX_BDP_POLL_CNTR_EN;
	tx_ctrl |= HIF_CH_TX_CTRL_DMA_EN;
	fp_reg_write((ULONG)&regs->ch[ch_index].hif_ctrl_ch, tx_ctrl);

	tx_ctrl = HIF_CH_TX_START;
	fp_reg_write((ULONG)&regs->ch[ch_index].hif_tx_ch_start, tx_ctrl);
	fp->netstats.tx_packets++;
	fp->netstats.tx_bytes += total_pkt_len;
	// dev->trans_start = jiffies;
	spin_unlock_bh(&fp->lock);

	return NETDEV_TX_OK;
}

void
fp_tx_tasklet(ulong data)
{
	struct tasklet_data *txptr = (struct tasklet_data *)data;
	fp_private_t *fp = (fp_private_t *)txptr->fp;
	struct bd_ring *txring;
	struct bd *txbd;
	UINT wb_buflen = 0;
	UINT ch_index = txptr->ch_index;
	struct netdev_queue *queue = netdev_get_tx_queue(fp->dev, ch_index);

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "TX Tasklet: ch_index %d\n", ch_index);

	/* hif_index = 0; ch_index = 0; */
	txring = fp->hif.ch[ch_index].txring;
	if (txring == NULL) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "%s: null txring ptr\n", __func__);
		return;
	}

	spin_lock_bh(&fp->lock);

	while (1) {
		/* Process tx write back bds upon tx interrupt */
		txbd = fp_deque_bd(txring, &wb_buflen);
		if (txbd == NULL) {
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "TX wb bd not ready\n");
			break;
		}

		if (netif_tx_queue_stopped(queue)){
			netif_tx_wake_queue(queue);
		}

	}

	fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_ch_int_en, HIF_CH_INT_EN | HIF_CH_RXPKT_INT_EN | HIF_CH_TXPKT_INT_EN);

	spin_unlock_bh(&fp->lock);
}

irqreturn_t
fp_interrupt(int irq, void *dev)
{
	fp_private_t *fp = (fp_private_t *)netdev_priv(dev);
	unsigned int reg;

	reg = handle_hif(&fp->hif);

	return IRQ_HANDLED;
}

unsigned int handle_hif(hif_base_t *hif)
{
	UINT ch_index = 0;
	volatile struct hif_regs *regs = hif->regs;
	UINT hif_int_src = fp_reg_read((ULONG)&regs->hif_int_src);

	/* return if interrupt is not ours */
	if (!hif_int_src)
		return 0xFFFFFFFF;

	/* disable all hif channel interrupts */
	fp_disable_hif_ch_interrupts(hif);

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		if (hif_int_src & (HIF_CH_INT << ch_index)){
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: HIF_INT_SRC: 0x%x\n", __func__, hif_int_src);
			handle_channel(hif, ch_index);
		}
	}

	/* enable all hif channel interrupts */
	fp_enable_hif_ch_interrupts(hif);

	fp_reg_write((ULONG)&regs->hif_int_src, hif_int_src);

	return 0;
}

void handle_channel(hif_base_t *hif, UINT ch_index)
{
	volatile struct hif_regs *regs = hif->regs;
	UINT hif_ch_int_src = fp_reg_read((ULONG)&regs->ch[ch_index].hif_ch_int_src);

	if (hif_ch_int_src & HIF_CH_RXPKT_INT) {
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: RX: GOT CALLED, INTSRC IS 0x%x\n", __func__, hif_ch_int_src);
		/* Schedule the deferred intr processing routine */
		tasklet_schedule(&hif->ch[ch_index].rx_tasklet);
	}

	if (hif_ch_int_src & HIF_CH_TXPKT_INT) {
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: TX: GOT CALLED, INTSRC IS 0x%x\n", __func__, hif_ch_int_src);
		tasklet_schedule(&hif->ch[ch_index].tx_tasklet);
	}

	/* Clear ch[ch_index] interrupt register */
	fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_src, hif_ch_int_src);
}


void fp_disable_hif_ch_interrupts(hif_base_t *hif)
{
	UINT ch_index = 0;
	volatile struct hif_regs *regs = hif->regs;

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_en, 0);
	}
}


void fp_enable_hif_ch_interrupts(hif_base_t *hif)
{
	UINT ch_index = 0;
	volatile struct hif_regs *regs = hif->regs;

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_en, (HIF_CH_INT_EN | HIF_CH_RXPKT_INT_EN | HIF_CH_TXPKT_INT_EN));
	}
}

#if 0
/* Sample ISR for hif1 and channel 1. */
void hif1_chan1_isr(int irq, void *dev)
{
	fp_private_t *fp = (fp_private_t *)netdev_priv(dev);
	volatile struct hif_regs *regs = fp->hif[1].regs;

	/* Disable corresponding hif and channel interrupt */
	fp_reg_write(fp_baseAddr, (ULONG)&regs->ch[1].hif_ch_int_en, 0);

	handle_channel(&fp->hif[1], 1);

	/* Enable corresponding hif and channel interrupt */
	fp_reg_write(fp_baseAddr, (ULONG)&regs->ch[1].hif_ch_int_en,
	(HIF_CH_INT_EN | HIF_CH_RXPKT_INT_EN | HIF_CH_TXPKT_INT_EN));
}
#endif

/**********************************************************************
 * Function Name  : fp_disable_all_ch_interrupts
 * Description    : clear all pending interrupts
 * Inputs
 *   Parameters   : fp_private_t *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
static inline void fp_disable_all_ch_interrupts(fp_private_t *fp)
{
	UINT  int_src;
	volatile struct hif_regs *regs;
	UINT ch_index = 0;

	regs = fp->hif.regs;

	/* Clear the pending interrupts */
	int_src = fp_reg_read((ULONG)&regs->hif_int_src);
	fp_reg_write((ULONG)&regs->hif_int_src, int_src);

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		/* disable the channel[ch_index] interrupt enable */
		fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_en, 0);
	}
}

/**********************************************************************
 * Function Name  : fp_disable_interrupts
 * Description    : clear all pending interrupts
 * Inputs
 *   Parameters   : fp_private_t *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *********************************************************************/
void fp_disable_interrupts(fp_private_t *fp)
{
	UINT  int_src;
	volatile struct hif_regs *regs;
	UINT ch_index = 0;

	regs = fp->hif.regs;

	/* Clear the pending interrupts */
	int_src = fp_reg_read((ULONG)&regs->hif_int_src);
	fp_reg_write((ULONG)&regs->hif_int_src, int_src);

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		/* disable the channel[ch_index] interrupt enable */
		fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_en, 0);

		/* Clear the channel[ch_index] pending interrupt sources */
		int_src = fp_reg_read((ULONG)&regs->ch[ch_index].hif_ch_int_src);
		fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_src, int_src);
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: int_src is 0x%x\n", __func__, int_src);
}

void fp_enable_interrupts(fp_private_t *fp)
{
	volatile struct hif_regs *regs;
	UINT ch_index = 0;

	regs = fp->hif.regs;
	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		/* Enable all hif channel interrupts */
		fp_reg_write((ULONG)&regs->ch[ch_index].hif_ch_int_en, (HIF_CH_INT_EN | HIF_CH_RXPKT_INT_EN | HIF_CH_TXPKT_INT_EN));
	}
}

void fp_init_rx_tasklet(fp_private_t *fp)
{
	UINT ch_index = 0;

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		fp_rx_info[ch_index].fp = fp;
		fp_rx_info[ch_index].ch_index = ch_index;
		tasklet_init(&fp->hif.
			ch[ch_index].rx_tasklet, fp_rx_tasklet,
			(ulong)&fp_rx_info[ch_index]);
	}
}

void fp_init_tx_tasklet(fp_private_t *fp)
{
	UINT ch_index = 0;

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		fp_tx_info[ch_index].fp = fp;
		fp_tx_info[ch_index].ch_index = ch_index;
		tasklet_init(&fp->hif.
			ch[ch_index].tx_tasklet, fp_tx_tasklet,
			(ulong)&fp_tx_info[ch_index]);
	}
}

void fp_init_bdp_base_reg(fp_private_t *fp)
{
	UINT ch_index = 0;

	for (ch_index = 0; ch_index < NUM_HIF_CHANNELS; ch_index++) {
		/* Program the RX WRBK BD reg */
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_rx_bdp_wr_low_addr_ch, fp->hif.ch[ch_index].rxring->wb_bd_tbl_pa);
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_rx_bdp_wr_high_addr_ch, 0x0);

		/* Program the RXBD reg */
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_rx_bdp_rd_low_addr_ch, fp->hif.ch[ch_index].rxring->bd_tbl_pa);
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_rx_bdp_rd_high_addr_ch, 0x0);

		/* Program the TX WRBK BD reg */
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_tx_bdp_wr_low_addr_ch, fp->hif.ch[ch_index].txring->wb_bd_tbl_pa);
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_tx_bdp_wr_high_addr_ch, 0x0);

		/* Program the TXBD reg */
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_tx_bdp_rd_low_addr_ch, fp->hif.ch[ch_index].txring->bd_tbl_pa);
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_tx_bdp_rd_high_addr_ch, 0x0);

		/* Program the RX WRBK buff size */
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_rx_wrbk_bd_ch_buffer_size, NUM_RX_WB_DESCR);

		/* Program the TX WRBK buff size */
		fp_reg_write((ULONG)&fp->hif.regs->ch[ch_index].hif_tx_wrbk_bd_ch_buffer_size, NUM_TX_WB_DESCR);
	}

	/* Program the sequence num check  start in MISC reg */
	fp_reg_write((ULONG)&fp->hif.regs->hif_misc, 0x1001f);
}
