/*************************************************************************/ /*!
@File
@Title          fp_pci_driver.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    base pci driver for fastpath module
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

/**********************************************************************
 *                                     Include Files
 **********************************************************************/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/interrupt.h>

#include <fp_types.h>
#include <wsp_vlan_hash_table_L.h>
#include "fp_pci_driver.h"
#include "fp_net_driver.h"
#include "fp_hif_reg.h"
#include "fp_mdio_ops.h"
#include "fp_netd.h"
#include "fp_proc_fs.h"
#include "fp_macros.h"

MODULE_AUTHOR( "Milandr application team <support@milandr.ru>" );
MODULE_LICENSE( "GPL" );
MODULE_DESCRIPTION( "Linux kernel module for ic 1923KX028" );

/**********************************************************************
 *				Global Declarations
 **********************************************************************/
/* Enabling debug prints for a feature can be done by setting the respective
 * flags in the follwoing global variable. */

unsigned char fp_dbgp_arr[FP_MAX_DBGP][FP_MAX_DBGP_LEN] = {
			{"DBGP_FEAT_HIF"}, {"DBGP_FEAT_BRIDGE"},
			{"DBGP_FEAT_ROUTE"}, {"DBGP_FEAT_FWD_PE"},
			{"DBGP_FEAT_TMU"}, {"DBGP_FEAT_CLI"},
			{"DBGP_FEAT_L2CLI"}, {"DBGP_FEAT_L3CLI"},
			{"DBGP_FEAT_L2API"}, {"DBGP_FEAT_L3API"},
			{"DBGP_FEAT_IOCTL"}, {"DBGP_FEAT_LIB"},
			{"DBGP_FEAT_ERR"}, {"DBGP_FEAT_ETHTOOL"},
			{"DBGP_FEAT_PROCFS"},{"DBGP_FEAT_DRV_MCU"},
			{0},{0},
			{0},{0},
			{0},{0},
			{0},{0},
			{0},{0},
			{0},{0},
			{0},{0}
		};

UINT fp_dbgp_flags = 0x5FFF;
UINT fp_trace_level = 7;

UCHAR  *fp_net_dev;
UCHAR  *fp_baseAddr;
UCHAR  *fp_baseAddr_bar4;

uint32_t get_pci_cfg_word( uint32_t offset, uint32_t *ptr ){
	fp_private_t *fppriv;

	if( fp_net_dev != FP_NULL ){
		fppriv = (fp_private_t *)netdev_priv( (struct net_device*)fp_net_dev );
		if( !pci_read_config_dword( fppriv->pci_dev, offset, ptr ) ){

			return 0;
		}
	}
	return 1;
}

uint32_t set_pci_cfg_word( uint32_t offset, uint32_t value ){
	fp_private_t *fppriv;

	if( fp_net_dev != FP_NULL ){
		fppriv = (fp_private_t *)netdev_priv( (struct net_device*)fp_net_dev );
		if( !pci_write_config_dword( fppriv->pci_dev, offset, value ) ){
			return 0;
		}
	}
	return 1;
}



/**********************************************************************
 * Function Name  : fp_pci_probe
 * Description    : pci dev probe routine called when pci dev is
 * Inputs
 *   Parameters   : struct pci_dev *, struct pci_device_id *
 * Outputs        :
 *   Parameters   : -
 *   Returns      : struct net_device_stats *
 * Changes        :
 *********************************************************************/
static int
fp_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	fp_private_t *fppriv;
	struct net_device *dev = NULL;
	UINT mem_start, mem_end, mem_len , io_start, io_end, io_len;
	char macaddr[6];
	int rc = 0;

	dev = alloc_etherdev_mq(sizeof(*fppriv), NPU_MAX_TX_QUEUES);
	if (!dev) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG,
				"Unable to allocate new ethernet device\n");
		return -ENOMEM;
	}


	dev->flags &= ~IFF_UP;
	dev->netdev_ops   = &fp_netdev_ops;
	dev->ethtool_ops = &fp_ethtool_ops;
	dev->rtnl_link_ops = NULL;
	ether_setup(dev);
	dev->mtu = 1500;
	dev->flags &= ~IFF_MULTICAST;

	macaddr[0] = 0x00;
	macaddr[1] = 0x1a;
	macaddr[2] = 0x2b;
	macaddr[3] = 0x3c;
	macaddr[4] = 0x4d;
	macaddr[5] = 0x5e;

	memcpy(dev->dev_addr, macaddr, sizeof(macaddr));

	fppriv = (fp_private_t *)netdev_priv(dev);
	fppriv->dev = dev;
	fppriv->pci_dev = pdev;
	fp_net_dev = (unsigned char *)dev;

	for(rc = 0; rc < FP_MAX_PORTS; rc++){
		fp_logical_dev[rc] = NULL;
	}

	pci_set_drvdata(pdev, dev);
	rc = pci_enable_device(pdev);
	if (rc < 0) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "PCI dev enable failure\n");
		goto err_out_free_dev;
	}

//	mem_start = pci_resource_start(pdev, 0);
	mem_start = pci_resource_start(pdev, 2);
//	mem_end = pci_resource_end(pdev, 0);
	mem_end = pci_resource_end(pdev, 2);

	io_start = pci_resource_start(pdev, 4);
	io_end = pci_resource_end(pdev, 4);

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource BAR2 mem start address: 0x%x\n", mem_start);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource BAR2 mem end address: 0x%x\n", mem_end);

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource BAR4 mem start address: 0x%x\n", io_start);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource BAR4 mem end address: 0x%x\n", io_end);

//	if (!(pci_resource_flags(pdev, 0) & IORESOURCE_MEM)) {
	if (!(pci_resource_flags(pdev, 2) & IORESOURCE_MEM)) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "Cann't read PCI resource flags of BAR2\n");
		goto err_out_disable_pdev;
	}

	if (!(pci_resource_flags(pdev, 4) & IORESOURCE_MEM)) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "Cann't read PCI resource flags of BAR4\n");
		goto err_out_disable_pdev;
	}

//	mem_len = pci_resource_len(pdev, 0);
	mem_len = pci_resource_len(pdev, 2);
	io_len = pci_resource_len(pdev, 4);
//	io_start = pci_resource_start(pdev, 1);
//	io_end = pci_resource_end(pdev, 1);
//	io_len = pci_resource_len(pdev, 1);

//	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource end address: %x\n", mem_len);
//	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource flags read as MEM mapped\n");
//	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource io start address: %x\n", io_start);
//	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource io end address: %x\n", io_end);
//	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource io end address: %x\n", io_len);
//	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI resource flags read as IO mapped\n");

	rc = pci_write_config_dword( pdev, FP_DATA1_REG, 0 );
	if( rc != 0 ){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "fp_pci_probe: can not write FP_DATA1_REG\n" );
	}

	rc = pci_write_config_dword( pdev, FP_DATA2_REG, 0 );
	if( rc != 0 ){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "fp_pci_probe: can not write FP_DATA2_REG\n" );
	}

	rc = pci_write_config_dword( pdev, FP_CMD_REG, 0 );
	if( rc != 0 ){
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_ERR, "fp_pci_probe: can not write FP_CMD_REG\n" );
	}

	rc = pci_request_regions(pdev, DRV_NAME);
	if (rc) {
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "Cann't obtain PCI resources\n");
		goto err_out_disable_pdev;
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "PCI acquired regions\n");

	rc = pci_set_dma_mask(pdev, DMA_BIT_MASK(30));
	if (rc) {
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "No usable DMA configuration\n");
		goto err_out_free_res;
	}

	rc = pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(30));
	if (rc) {
		FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "Consistent DMA not available\n");
		goto err_out_free_res;
	}

	pci_set_master(pdev);
	//pci_enable_msi(pdev);

	SET_NETDEV_DEV(dev, &pdev->dev);

	fppriv->pcimem_base = ioremap_nocache(mem_start, mem_len);
	fppriv->pcimem_base_bar4 = ioremap_nocache(io_start, io_len);

	fp_baseAddr = (UCHAR *)fppriv->pcimem_base;
	fp_baseAddr_bar4 = (UCHAR *)fppriv->pcimem_base_bar4;

	fp_assign_hif_base_addr(fppriv);

	dev->base_addr = mem_start;
	dev->irq = pdev->irq;

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "dev->base_addr: 0x%lx\n", dev->base_addr);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "PCIBASE BAR2: 0x%p\n", fppriv->pcimem_base);
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "PCIBASE BAR4: 0x%p\n", fppriv->pcimem_base_bar4);
	strcpy(dev->name, "fp%d");
	rc = register_netdev(dev);
	if (rc) {
		iounmap((void *)(ULONG)mem_start);
		iounmap((void *)(ULONG)io_start);
		FP_DEBUG(DBGP_FEAT_ERR, FP_LOG_EMERG, "fp_eth: Error %d  initializing card fp_eth card\n", rc);
		return rc;
	}

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "fpath device registered\n");
	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "write and read the PCI memory space\n");
	spin_lock_init(&fppriv->lock);
	spin_lock_init(&fppriv->rx_lock);
	fppriv->config_flags = 0;
	fppriv->ageing_timer_on = 0;
	fppriv->host_system_loopback = 0;
	
	fp_proc_fs_init();
	return 0;

err_out_free_res:
	pci_release_regions(pdev);
err_out_disable_pdev:
	pci_disable_device(pdev);
err_out_free_dev:
	pci_set_drvdata(pdev, NULL);
	free_netdev(dev);

	return 0;
}

void
fp_pci_remove(struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata(pdev);
	fp_private_t *fp = netdev_priv(dev);

	if (dev == NULL)
		return;

	if (fp == NULL)
		return;

	FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_EMERG, "<0> Cleaning Up the Module\n");
	fp_proc_fs_deinit();

	if( fp->config_flags & SWITCH_CONFIG_DONE ){
		fp_net_release(dev);
	}
	
	pci_clear_master(pdev);

	if (fp->pcimem_base)
		iounmap( ( void* )fp->pcimem_base );

	if(fp->pcimem_base_bar4){
		iounmap( ( void* )fp->pcimem_base_bar4 );
	}

	if (dev) {
		unregister_netdev(dev);
		free_netdev(dev);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		pci_set_drvdata(pdev, NULL);
	}

	return;
}

static struct pci_driver fp_pci_driver = {
	.name       = MODULENAME,
	.id_table   = fp_pci_tbl,
	.probe      = fp_pci_probe,
	.remove     = fp_pci_remove,
};

static int __init
fp_pci_init(void)
{
    int result = 0;

	result = pci_register_driver(&fp_pci_driver);

    return result;
}

static void __exit
fp_pci_exit(void)
{
	pci_unregister_driver(&fp_pci_driver);
}

module_init(fp_pci_init);
module_exit(fp_pci_exit);
