/*************************************************************************/ /*!
@File
@Title          fp_pre_header.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    handles rx-tx prepended headers
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/
 #include <linux/version.h>
 #include <linux/module.h>
 #include <linux/netdevice.h>
 #include <linux/etherdevice.h>
 #include <linux/if_vlan.h>

 #include <fp_library.h>
 #include <fp_pre_header.h>
 #include <fp_os_shim.h>
 #include <fp_netd.h>
 #include <fp_ltc.h>
 #include <fp_switch.h>
 #include <hw_vlan_hash_table.h>
 #include <hw_2field_hash_table.h>

#include "fp_net_driver.h"

uint8_t control_block_injection_flag_values[ CB_INJ_TX_FLAG_VALUES_LEN ] = {
	CB_INJ_TX_FLAG1_VALUE,
	CB_INJ_TX_FLAG2_VALUE,
	CB_INJ_TX_FLAG3_VALUE,
	CB_INJ_TX_FLAG4_VALUE,
	CB_INJ_TX_FLAG5_VALUE,
	CB_INJ_TX_FLAG6_VALUE,
	CB_INJ_TX_FLAG7_VALUE,
	CB_INJ_TX_FLAG8_VALUE
};

int fp_fill_tx_ptp_header(struct tx_header *txhdr, struct sk_buff *skb)
{
#ifdef HAVE_PTP_1588_CLOCK
	int tx_port = skb->cb[CB_PORT_FOR_INJECTION];
	struct net_device *dev;
	nfp_dev_info_t *nfp_priv;
	static int ref_num;

	memset(txhdr, 0, FP_TX_HDR_LEN);

	if (tx_port < FP_MAX_PORTS) {
		txhdr->txport_map = (1 << tx_port);
		dev = fp_phy_addr[tx_port];
	}else{
		/* We don't know this port, cannot handle..  return */
		return 0;
	}

	nfp_priv = NFP_DEV_INFO(dev);

	if (nfp_priv->ptp_tx_skb) {
		dev_kfree_skb_any(nfp_priv->ptp_tx_skb);
		nfp_priv->ptp_tx_skb = NULL;
	}

	if (!(skb_shinfo(skb)->tx_flags & SKBTX_HW_TSTAMP)) {
		txhdr->ctrl = FP_TX_PKT_INJECT_EN;
		return 0;
	}
#ifdef LTC_EN
	if (skb->lt_enable == LTC_ENABLE_MAGICNUM) {
		txhdr->ctrl |= FP_TX_LAUNCH_TIME_VALID;
		txhdr->launch_time_nsec = skb->lt_nsec;
		txhdr->launch_time_sec = skb->lt_sec;
	}
#endif
	txhdr->ctrl = FP_TX_PKT_INJECT_EN;
	txhdr->ctrl |= FP_TX_PKT_PTP_EN;

	txhdr->ref_num = ref_num++;

	skb_shinfo(skb)->tx_flags |= SKBTX_IN_PROGRESS;
	nfp_priv->ptp_tx_skb = skb;
	nfp_priv->ptp_tx_start = jiffies;
	nfp_priv->ref_num = txhdr->ref_num;

#endif
	return 0;

}
/**********************************************************************
 * Function Name  : fp_add_prepend_tx_header
 * Description    :
 * Inputs         :
 * Parameters     : struct sk_buff *
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int fp_add_prepend_tx_header(struct sk_buff *skb)
{
	struct tx_header *txhdr;

	if( skb_headroom( skb ) < FP_TX_HDR_LEN ){
		FP_DEBUG( DBGP_FEAT_ERR, FP_LOG_EMERG,
			"%s: Not enough headroom for TX HDR\n", __func__ );
		return -1;
	}

	skb_push( skb, FP_TX_HDR_LEN );
	txhdr = ( struct tx_header* )skb->data;
	memset( txhdr, 0, FP_TX_HDR_LEN );

	if( ( !memcmp( &skb->cb[ CB_INJ_TX_FLAG1 ], control_block_injection_flag_values, CB_INJ_TX_FLAG_VALUES_LEN ) ) &&
		( ( uint8_t )skb->cb[ CB_INJ_TX_PORT ] < ( FP_MAX_PORTS - 1 ) ) ){

		txhdr->ctrl = FP_TX_PKT_INJECT_EN;
		txhdr->txport_map = 1 << skb->cb[ CB_INJ_TX_PORT ];
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_INFO, "TX INJECT PORT %u\n", ( uint8_t )skb->cb[ CB_INJ_TX_PORT ] );
	}

#ifdef LTC_EN
	if ( skb->lt_enable == LTC_ENABLE_MAGICNUM ){
		txhdr->ctrl |= FP_TX_LAUNCH_TIME_VALID;
		txhdr->launch_time_nsec = skb->lt_nsec;
		txhdr->launch_time_sec = skb->lt_sec;
	}
#endif
	
	txhdr->ref_num = 0;
	txhdr->rsvd1 = 0;

	return 0;
}

/**********************************************************************
 * Function Name  : fp_process_rx_header
 * Description    : check whether incoming packet is rx packet or egress
			time stamp report
 * Inputs         :
 * Parameters     : struct sk_buff *
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int fp_process_rx_header(struct sk_buff *skb)
{
	struct rx_header *rxhdr;
	struct egress_report *egrhdr;
	UCHAR rx_packet_valid = 0;
	UCHAR *ctrl = skb->data + 3;
	struct net_device *dev;
	int l2_special = 0;
	struct sk_buff  *new_skb = NULL;
	fp_dev_info_t   *logical_priv;
    fp_private_t    *real_priv;
    uint16_t network_hdr_offset = 0;

#if 0
#ifdef HAVE_PTP_1588_CLOCK
	nfp_dev_info_t *nfp_priv;
	u64 ns = 0;
#endif
#endif

	/* check whether incoming data is rxpacket or egress time stamp report */
	if (*ctrl & FP_RX_TS_REPORT_EN) {
		/* update rx packet is not valid */
		rx_packet_valid = 0;

		/* incoming packet is egress time stamp report */
		egrhdr = (struct egress_report *)skb->data;

#if 0
#ifdef HAVE_PTP_1588_CLOCK
		if (egrhdr->rxport_num < FP_MAX_PORTS) {
			dev = fp_phy_addr[egrhdr->rxport_num];
		}
		else{
			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_process_rx_header device unknown\n");
			dev_kfree_skb(skb);
			return -1;
		}

		ns = (egrhdr->egress_timestamp_sec * 1000000000ULL) + (egrhdr->egress_timestamp_nsec);
		nfp_priv = NFP_DEV_INFO(dev);

		if (nfp_priv->ref_num != egrhdr->ref_num) {

			FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "fp_process_rx_header: ref_num not matching\n");
			dev_kfree_skb(skb);
			return -1;
		}

		nfp_priv->ns = ns;
		npu_ptp_tx_handler(dev);
#endif /* HAVE_PTP_1588_CLOCK */
#endif
		dev_kfree_skb( skb );
	}else{
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: rcv normal packet\n", __func__ );
		/* update rx packet is valid */
		rx_packet_valid = 1;

		/* incoming packet is rx packet */
		rxhdr = ( struct rx_header* )skb->data;

		if( FP_PUNT_L2_SPECIAL == rxhdr->punt_reason ){
			l2_special = rxhdr->punt_reason;
		}

		if( rxhdr->rxport_num < FP_MAX_PORTS ){
			dev = fp_phy_addr[ rxhdr->rxport_num ];
		}else{
			dev_kfree_skb( skb );
			FP_DEBUG( DBGP_FEAT_ERR, FP_LOG_ERR, "%s: error port number in packet header!\n", __func__ );
			return -1;
		}
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: remove packet header\n", __func__ );
		skb_pull( skb, sizeof( struct rx_header ) );                               // remove header

		FP_DEBUG_PRINT_HEX_DUMP( DBGP_FEAT_HIF, FP_LOG_DEBUG, "RX_FROM_NET ", DUMP_PREFIX_OFFSET, 16, 1,
						skb->data, skb->len, true )
		skb->mac_len = ETH_HLEN;
		/* vlan tag processing */
		if( vlan_get_tag( skb, &skb->vlan_tci ) ){
			struct ethhdr *eh = ( struct ethhdr* )skb->data;
			FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: packet without vlan tag\n", __func__ );
			skb->protocol = eh->h_proto;
			if( fp_port[ rxhdr->rxport_num ].fallback_bd_id != fp_port[ HOST_PORT_NO ].fallback_bd_id ){
				skb->vlan_tci = (   fp_port[ rxhdr->rxport_num ].fallback_bd_id & VLAN_VID_MASK ) |
								( ( fp_port[ rxhdr->rxport_num ].def_pri << VLAN_PRIO_SHIFT ) & VLAN_PRIO_MASK ) |
								VLAN_TAG_PRESENT;
				skb->vlan_proto = htons( ETH_P_8021Q );
				FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_DEBUG, "%s: push VLAN tag in skb, VID: %d, prio %d\n",
							 __func__,
							(   skb->vlan_tci & VLAN_VID_MASK ),
							( ( skb->vlan_tci & VLAN_PRIO_MASK ) >> VLAN_PRIO_SHIFT ) );
			}else{
				skb->vlan_tci = 0;
			}
			network_hdr_offset = sizeof( struct ethhdr );
		}else{
			skb->protocol = htons( ETH_P_8021Q );
			skb->protocol = vlan_get_protocol( skb );
			skb->vlan_proto = htons( ETH_P_8021Q );
			skb->vlan_tci |= VLAN_TAG_PRESENT;
			memmove( skb->data + VLAN_HLEN, skb->data, ( ETH_ALEN << 1 ) );
			skb_pull( skb, VLAN_HLEN );	// remove vlan tag
			network_hdr_offset = sizeof( struct ethhdr );
			FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: vlan tci: %04X, protocol: %04X\n", __func__,
						( skb->vlan_tci & ( ~VLAN_TAG_PRESENT ) ), ntohs( skb->protocol ) );
		}
		skb_set_network_header( skb, network_hdr_offset );

		if( fp_logical_dev[ 0 ] ){
			logical_priv = netdev_priv( fp_logical_dev[ 0 ] );
        	real_priv  = netdev_priv(logical_priv->real_dev);
        	
			if( real_priv->host_system_loopback ){
				if( rxhdr->rxport_num != HIF_PORT_NUM ){
					new_skb = skb_copy( skb, GFP_ATOMIC );
					if( new_skb == NULL ){
						FP_DEBUG( DBGP_FEAT_ERR, FP_LOG_EMERG,
							"%s: out of memory allocating"
							"skb TX buffer\n", __func__ );
						return 0;
					}
					memcpy( &new_skb->cb[ CB_INJ_TX_FLAG1 ], control_block_injection_flag_values, CB_INJ_TX_FLAG_VALUES_LEN );
					new_skb->cb[ CB_INJ_TX_PORT ] = rxhdr->rxport_num;
					dev_queue_xmit( new_skb );
        		}
        		dev_kfree_skb(skb);
        		return 0;
			}
		}
	}
	if( rx_packet_valid ){
		fp_host_shim_entry(skb->data, skb->len, rxhdr);
		if( rxhdr->punt_reason == FP_PUNT_SA_MISS ){
			fp_tx_inject_packet( skb, rxhdr->rxport_num, 0 );
		}else if( rxhdr->punt_reason == FP_PUNT_SA_RELEARN ){
			fp_tx_inject_packet( skb, rxhdr->rxport_num, 1 );
		}
		//skb->protocol = eth_type_trans(skb, skb->dev);
		eth_type_trans( skb, skb->dev );
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: skb->protocol is %04X\n", __func__, ntohs( skb->protocol ) );

#if 0
#ifdef HAVE_PTP_1588_CLOCK
		if (unlikely(ntohs(skb->protocol) == 0x88F7)) {
			ns = (rxhdr->rx_timestamp_sec * 1000000000ULL) + (rxhdr->rx_timestamp_nsec);
			npu_ptp_rx_pktstamp(skb, ns);
			skb->dev = dev;
		}
#endif /* HAVE_PTP_1588_CLOCK */
#endif

		if (l2_special == FP_PUNT_L2_SPECIAL)
        {
			if (dev != NULL)
				skb->dev = dev;
		}
		FP_DEBUG( DBGP_FEAT_HIF, FP_LOG_DEBUG, "%s: push the packet to the network stack\n", __func__ );
		netif_rx( skb );
	}

	return 0;
}

/**********************************************************************
 * Function Name  : fp_tx_inject_packet
 * Description    : inject pkt to except incoming and host ports
 * Inputs         :
 * Parameters     : struct sk_buff *, int rx_port
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
void fp_tx_inject_packet( struct sk_buff *skb, int rx_port, int relearn_flag ){
	struct sk_buff *new_skb = NULL;
	int port_num = 0, port_list = 0, vlanid;
	int64_t action_entry;

	if( fp_baseAddr == NULL ){
		FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_EMERG, "%s: pci_mem base is NULL\n", __func__ );
		return;
	}

	if( skb_vlan_tag_present( skb ) ){
		vlanid = skb->vlan_tci & VLAN_VID_MASK;
	}else{
		vlanid = fp_port[ rx_port ].fallback_bd_id;
	}

    if( relearn_flag ){
		/* search mac table entry for destination port */
    	action_entry = hw_hash_table_search( skb->data, vlanid );
    	if( action_entry == -1 ){
    		/* mac table hasn't entry for destination port, search vlan table entry port list */
    		action_entry = wsp_vlan_entry_search( vlanid );
    		port_list = ( int32_t )action_entry;
    		if( action_entry == -1 ){
    			/* vlan table hasn't entry for this vlan, get GLOBAL_VLAN_ENTRY forward list */
				port_list = fp_global.fallback_bd_entry.forward_list;
			}
    	}else{
    		/* mac table has entry for destination port */
    		port_list = action_entry;
    	}
    }else{
		/*Getting vlan table entry port list*/
		action_entry = wsp_vlan_entry_search( vlanid );
		if( action_entry == -1 ){
			port_list = fp_global.fallback_bd_entry.forward_list;
		}else{
			port_list = action_entry;
		}
	}
	port_list &= FWD_PORT_LIST_MASK;

	FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "vlanid is %d, port_list is %x\n", vlanid, port_list );

	/* this condition should not occur, flood to all ports except iport */
	for( port_num = 0; port_num < FP_MAX_SWITCH_PORTS; port_num++ ){
		/* check for port bit set */
		if( ( port_list >> port_num ) & 1 ){
			/* fwd port shldn't be input port */
			if( ( rx_port != port_num ) && ( port_num != HIF_PORT_NUM ) ){
				/* allocate memory for sk_buff */
				new_skb = skb_copy( skb, GFP_ATOMIC );
				if( new_skb == NULL ){
					FP_DEBUG( DBGP_FEAT_ERR, FP_LOG_EMERG,
						"%s: Out of memory allocating"
						"skb TX buffer\n", __func__ );
					continue;
				}
				memcpy( &new_skb->cb[ CB_INJ_TX_FLAG1 ], control_block_injection_flag_values, CB_INJ_TX_FLAG_VALUES_LEN );
				new_skb->cb[ CB_INJ_TX_PORT ] = port_num;
				FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "%s: protocol %04x, vlan_tci %04x\n", __func__, ntohs( new_skb->protocol ), ( new_skb->vlan_tci & ( ~VLAN_CFI_MASK ) ) );
				dev_queue_xmit( new_skb );
			}
		}
	}
}
