/*************************************************************************/ /*!
*  @File
*  @Title          fp_drv_mcu.c
*  @Copyright      Copyright (c) Milandr. All Rights Reserved
*  @Description    Initializes EAC paramaters
*  @License        Dual MIT/GPLv2
*
*  The contents of this file are subject to the MIT license as set out below.
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  The above copyright notice and this permission notice shall be included in
*  all copies or substantial portions of the Software.
*
*  Alternatively, the contents of this file may be used under the terms of
*  the GNU General Public License Version 2 ("GPL") in which case the provisions
*  of GPL are applicable instead of those above.
*
*  If you wish to allow use of your version of this file only under the terms of
*  GPL, and not to allow others to use your version of this file under the terms
*  of the MIT license, indicate your decision by deleting the provisions above
*  and replace them with the notice and other provisions required by GPL as set
*  out in the file called "GPL-COPYING" included in this distribution. If you do
*  not delete the provisions above, a recipient may use your version of this file
*  under the terms of either the MIT license or GPL.
*
*  This License is also included in this distribution in the file called
*  "MIT-COPYING".
*
*  EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
*  PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
*  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
*  PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
*  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
*  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
*  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
*                                      Include Files
**********************************************************************/
//#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/slab.h>
#include <linux/kthread.h>
#include <linux/sched.h>

#include "../include/fp_library.h"
#include "fp_drv_mcu_tasks.h"
#include "fp_net_driver.h"
#include "sfp_copper_setup.h"

/**********************************************************************
*                                      #defines
**********************************************************************/
#define SFP_ETHERNET_PHY_ID_MASK 0xFFFFFFF0
#define SFP_ETHERNET_PHY_MARVELL_88E1111	0x01410cc0
/**********************************************************************
 *                              Global Declarations
 ***********************************************************************/
int32_t config_marvell_88e1111( uint8_t port, uint8_t i2c_device_addr );

struct sfp_copper_setup_s s_sfp_copper[ NUM_OF_OPTIC_PORTS ];
struct sfp_copper_thread_s thr[ NUM_OF_OPTIC_PORTS ];
struct completion fin[ NUM_OF_OPTIC_PORTS ];

void get_eth_phy_reg_i2c_callback( uint32_t data ){
	uint32_t port = ( data >> 16 ) & 0xFF;
	
	if( port >= NUM_OF_OPTIC_PORTS ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ALERT, "%s: PORT NUMBER wrong (%d)! This should never happen!\n", __func__, port );
	}else{
		spin_lock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
		s_sfp_copper[ port ].return_value = ( uint16_t )( data & 0xFFFF );
		s_sfp_copper[ port ].i2c_transaction_complete = 1;
		spin_unlock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: GET reg: 0x%08x\n", __func__, s_sfp_copper[ port ].return_value );
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: wake up sfp_copper_setup_thread_%d\n", __func__, port );
		wake_up_interruptible( &s_sfp_copper[ port ].sfp_copper_microtask_wait_queue );	//wake up process
	}
}

void set_eth_phy_reg_i2c_callback( uint32_t data ){
	uint32_t port = ( data >> 16 ) & 0xFF;
	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: port_%d\n", __func__, port );
	if( port >= NUM_OF_OPTIC_PORTS ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ALERT, "%s: PORT NUMBER wrong (%d)! This should never happen!\n", __func__, port );
	}else{
		spin_lock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
		s_sfp_copper[ port ].return_value = data;
		s_sfp_copper[ port ].i2c_transaction_complete = 1;
		spin_unlock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: wake up sfp_copper_setup_thread_%d\n", __func__, port );
		wake_up_interruptible( &s_sfp_copper[ port ].sfp_copper_microtask_wait_queue );	//wake up process
	}
}

int32_t get_reg_through_i2c_bus( uint8_t port_number, uint8_t device_addr, uint8_t reg_addr, uint16_t *data ){
	port_number &= 0xFF;
	device_addr &= 0xFF;
	reg_addr &= 0xFF;
	if( port_number >= NUM_OF_OPTIC_PORTS ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: invalid PORT NUMBER (%d), drop request!\n", __func__, port_number );
		return -2;
	}
	spin_lock( &s_sfp_copper[ port_number ].sfp_copper_spin_lock );
	s_sfp_copper[ port_number ].i2c_transaction_complete = 0;
	spin_unlock( &s_sfp_copper[ port_number ].sfp_copper_spin_lock );
	create_task( GET_ETH_PHY_REG_I2C, ( ( port_number )|( device_addr << 8 )|( reg_addr << 16 ) ), FP_NULL );
	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: wake event from callback\n", __func__ );
	while( !s_sfp_copper[ port_number ].i2c_transaction_complete ){
		wait_event_interruptible( s_sfp_copper[ port_number ].sfp_copper_microtask_wait_queue, s_sfp_copper[ port_number ].i2c_transaction_complete );	//sleep process
		if( kthread_should_stop() ){
			return -ERESTARTSYS;
		}
	}
	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: wakeup thread_%d, get value 0x%08X\n", __func__, port_number, s_sfp_copper[ port_number ].return_value );
	if(	s_sfp_copper[ port_number ].return_value & 0xFF000000 ){
		return -1;
	}
	*data = ( uint16_t )( s_sfp_copper[ port_number ].return_value & 0xFFFF );
	return 0;
}

int32_t set_reg_through_i2c_bus( uint8_t port_number, uint8_t device_addr, uint8_t reg_addr, uint16_t data ){
	port_number &= 0xFF;
	device_addr &= 0xFF;
	reg_addr &= 0xFF;
	if( port_number >= NUM_OF_OPTIC_PORTS ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: invalid PORT NUMBER (%d), drop request!\n", __func__, port_number );
		return -2;
	}
	s_sfp_copper[ port_number ].return_value = data;
	s_sfp_copper[ port_number ].i2c_transaction_complete = 0;
	create_task( SET_ETH_PHY_REG_I2C, ( ( port_number )|( device_addr << 8 )|( reg_addr << 16 ) ), &s_sfp_copper[ port_number ].return_value );
	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: wake event from callback\n", __func__ );
	while( !s_sfp_copper[ port_number ].i2c_transaction_complete ){
		wait_event_interruptible( s_sfp_copper[ port_number ].sfp_copper_microtask_wait_queue, s_sfp_copper[ port_number ].i2c_transaction_complete );	//sleep process
		if( kthread_should_stop() ){
			return -ERESTARTSYS;
		}
	}
	if( s_sfp_copper[ port_number ].return_value & 0xFF000000 ){
		return -1;
	}
	return 0;
}

int setup_sfp_copper_module( void *data ){
	uint32_t ephy_id = 0xFFFFFFFF, trials;
	uint16_t reg;
	uint8_t port = ( uint8_t )( ( uint64_t )data & 0xFF );
	uint8_t device_addr;

	if( port >= NUM_OF_OPTIC_PORTS ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ALERT, "try thread started to unsupported port %d, stop the thread!\n", port );
		return 0;
	}

	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "thread started to port %d\n", port );
	while( !kthread_should_stop() ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "New cycle of sfp_module_configure for port %d\n", port );
		while( !s_sfp_copper[ port ].has_job ){
			//sleep until has not jobs
			wait_event_interruptible( s_sfp_copper[ port ].sfp_copper_job_wait_queue, s_sfp_copper[ port ].has_job );	//sleep process
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: wake up port_%d\n", __func__, port );
			if( kthread_should_stop() || ( s_sfp_copper[ port ].has_job == 0xFF ) ){
				//need to finish thread!
				FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "has STOP thread request for port %d\n", port );
				goto exit_from_setup_sfp_copper_module;
			}
		}
		spin_lock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
		s_sfp_copper[ port ].has_job = 0;
		device_addr = ( uint8_t )s_sfp_copper[ port ].i2c_address;
		spin_unlock( &s_sfp_copper[ port ].sfp_copper_spin_lock );

		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "Has a new job for thread %d\n", port );
		trials = 0;
		do{
			if( get_reg_through_i2c_bus( port, device_addr, 2, &reg ) < 0 ){
				FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: can not get reg through i2c bus for port %d\n", __func__, port );
				trials++;
				continue;
			}
			ephy_id = reg << 16;
			if( get_reg_through_i2c_bus( port, device_addr, 3, &reg ) < 0 ){
				FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: can not get reg through i2c bus for port %d\n", __func__, port );
				trials++;
				continue;
			}
			ephy_id |= reg;
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: EPHY ID is 0x%08X!\n", __func__, ephy_id );
			break;
		}while( trials < 3 );

		if( ( ephy_id & SFP_ETHERNET_PHY_ID_MASK ) == SFP_ETHERNET_PHY_MARVELL_88E1111 ){
			if( !config_marvell_88e1111( port, device_addr ) ){
				sfp_copper_configured( port );
				//void update_port_information( uint8_t port, uint8_t type,  );
				//update_port_information( port );
				FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: MARVELL 88E1111 configured PASS!\n", __func__ );
			}else{
				FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: MARVELL 88E1111 configured FAILED!\n", __func__ );
			}
		}
	}
exit_from_setup_sfp_copper_module:	
	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "thread stopped to port %d\n", port );
	complete( fin + port );
	return 0;
}

void fp_sfp_copper_setup_thread_init( void ){
	uint32_t i;

	for( i = 0; i < NUM_OF_OPTIC_PORTS; i++ ){
		memset( &s_sfp_copper[ i ], 0, ( sizeof( struct sfp_copper_setup_s ) * NUM_OF_OPTIC_PORTS ) );
		spin_lock_init( &s_sfp_copper[ i ].sfp_copper_spin_lock );
		s_sfp_copper[ i ].i2c_transaction_complete = 0;
		s_sfp_copper[ i ].has_job = 0;
		s_sfp_copper[ i ].i2c_address = SFP_MODULE_I2C_EPHY_ADDR;
		init_waitqueue_head( &s_sfp_copper[ i ].sfp_copper_microtask_wait_queue );
		init_waitqueue_head( &s_sfp_copper[ i ].sfp_copper_job_wait_queue );
		init_completion( fin + i );
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: create kthread\n", __func__ );
		thr[ i ].sfp_copper_setup_thread = kthread_run( setup_sfp_copper_module, ( void* )( ( uint64_t )i ), "sfp_copper_setup_thread_%d", i );
		if( thr[ i ].sfp_copper_setup_thread == ERR_PTR(-ENOMEM) ){
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ALERT, "%s: create kthread for port %d FAILED\n", __func__, i );
			thr[ i ].sfp_copper_setup_thread = FP_NULL;
		}
	}
}

void flush_job( uint8_t port ){
	if( ( thr[ port ].sfp_copper_setup_thread != FP_NULL )&&
		( port < NUM_OF_OPTIC_PORTS ) ){
		spin_lock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
		s_sfp_copper[ port ].has_job = 0xFF;
		spin_unlock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
	}else{
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: can not flush the job or pointer to thread for port %d is FP_NULL!\n", __func__, port );
	}
}

void fp_sfp_copper_setup_thread_deinit( void ){
	uint32_t i;
	for( i = 0; i < NUM_OF_OPTIC_PORTS; i++ ){
		if( thr[ i ].sfp_copper_setup_thread != FP_NULL ){
			flush_job( i );			
			kthread_stop( thr[ i ].sfp_copper_setup_thread );
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: thread for port %d finished!\n", __func__, i );
		}else{
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: pointer to thread for port %d = FP_NULL!\n", __func__, i );
		}
	}
}

 uint32_t fp_sfp_copper_setup_job_start( uint32_t cmd ){
 	uint32_t port;
 	port = cmd & 0xFF;
 	if( port >= NUM_OF_OPTIC_PORTS ){
 		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_ERR, "%s: Invalid port number: %d\n", __func__, port );
 		return 1;
 	}
 	spin_lock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
 	s_sfp_copper[ port ].i2c_address = ( cmd & SFP_MODULE_I2C_DEV_ADDR_PARAM_MASK ) >>  SFP_MODULE_I2C_DEV_ADDR_PARAM_POS;
 	s_sfp_copper[ port ].has_job = 1;
 	spin_unlock( &s_sfp_copper[ port ].sfp_copper_spin_lock );
 	wake_up_interruptible( &s_sfp_copper[ port ].sfp_copper_job_wait_queue );	//wake up thread
 	return 0;
 }

int32_t config_marvell_88e1111( uint8_t port, uint8_t i2c_device_addr ){
	uint16_t reg;
	uint32_t trials = 0;

	FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: start: port %d, I2C Dev addr 0x%X\n", __func__, port, i2c_device_addr );
	do{
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: try configure: %d\n", __func__, ( trials + 1 ) );
		if( get_reg_through_i2c_bus( port, i2c_device_addr, 27, &reg ) ){
			trials++;
			continue;
		}	
		reg &= 0xFFF0;
		reg |= ( 1 << 15 )|( 1 << 12 )|( 4 << 0 );
		if( set_reg_through_i2c_bus( port, i2c_device_addr, 27, reg ) ){
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, 
				"set_reg_through_i2c_bus: Can not write data to Port %d, I2C Dev addr 0x%X, Register 27, data 0x%x\n",
				port, i2c_device_addr, reg );
			trials++;
			continue;
		}
		if( set_reg_through_i2c_bus( port, i2c_device_addr, 9, 0x0F00 ) ){
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, 
				"set_reg_through_i2c_bus: Can not write data to Port %d, I2C Dev addr 0x%X, Register 9, data 0x0F00\n",
				port, i2c_device_addr );
			trials++;
			continue;
		}
		if( set_reg_through_i2c_bus( port, i2c_device_addr, 0, 0x8140 ) ){
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, 
				"set_reg_through_i2c_bus: Can not write data to Port %d, I2C Dev addr 0x%X, Register 0, data 0x8140\n",
				port, i2c_device_addr );
			trials++;
			continue;
		}
		if( set_reg_through_i2c_bus( port, i2c_device_addr, 4, 0x01E1 ) ){
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, 
				"set_reg_through_i2c_bus: Can not write data to Port %d, I2C Dev addr 0x%X, Register 4, data 0x01E1\n",
				port, i2c_device_addr );
			trials++;
			continue;
		}
		if( set_reg_through_i2c_bus( port, i2c_device_addr, 0, 0x9140 ) ){
			FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, 
				"set_reg_through_i2c_bus: Can not write data to Port %d, I2C Dev addr 0x%X, Register 0, data 0x9140\n",
				port, i2c_device_addr );
			trials++;
			continue;
		}
		break;
	}while( trials < 3 );

	if( trials < 3 ){
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_DEBUG, "%s: PASS\n", __func__ );
		return 0;
	}else{
		FP_DEBUG( DBGP_FEAT_DRV_MCU, FP_LOG_INFO, "%s: FAILED\n", __func__ );
		return -1;
	}
}


