/*************************************************************************/ /*!
@File
@Title          fp_library.h
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    all user defined wrapper functions
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#ifndef _FP_LIBRARY_H_
#define _FP_LIBRARY_H_
#include <linux/kernel.h>
#include <stddef.h>
#include "fp_types.h"
#include "fp_global_addr_map.h"
#include "fp_macros.h"
#include "fp_pktbuff.h"
#include "fp_ether.h"
#ifdef __FP_OS__
#include <stddef.h>
#endif
#define FP_NULL (void *)0

#define FP_TRUE    0
#define FP_FALSE  -1

extern void fp_macLe2BeCopy(char *src, char *dst);
extern char fp_macLe2BeCmp(char *src, char *dst);

void *fp_memcpy(void *Dest , const void *Src, UINT uiSize);

void *fp_memset(void *s, int c, UINT uiSize);
UCHAR *fp_ether_ntoa(UCHAR *SourceMac);
void fp_convert_le2be(int8_t *, uint32_t);
extern int fp_entry(void);


/* HW register read function */
uint32_t fp_reg_read( uint32_t addr );
/* HW register write function */
void fp_reg_write( uint32_t addr, uint32_t value );

/* fpath trace levels */
#define FP_LOG_EMERG       0       /* system is unusable */
#define FP_LOG_ALERT       1       /* action must be taken immediately */
#define FP_LOG_CRIT        2       /* critical conditions */
#define FP_LOG_ERR         3       /* error conditions */
#define FP_LOG_WARNING     4       /* warning conditions */
#define FP_LOG_NOTICE      5       /* normal but significant condition */
#define FP_LOG_INFO        6       /* informational */
#define FP_LOG_DEBUG       7       /* debug-level messages */

/* fpath trace types */
#define FP_TRACE_INIT                0x1
#define FP_TRACE_CLEANUP             0x2
#define FP_TRACE_L2_PARSING          0x4
#define FP_TRACE_L3_PARSING          0x8
#define FP_TRACE_L4_PARSING          0x10
#define FP_TRACE_ROUTING             0x20
#define FP_TRACE_BRIDGING            0x40
#define FP_TRACE_NF_HOOKS            0x80
#define FP_TRACE_CONFIG              0x100
#define FP_TRACE_LINUX               0x200
#define FP_TRACE_OTHER               0x400
#define FP_TRACE_ALL                 0xFFFFFFFF
/* fpath trace macros and data structures */
typedef struct fpath_trace_s {

    UINT     trace_flags;
    UINT     trace_level;

} fpath_trace_t;

extern fpath_trace_t fpath_trace;

#define FPATH_TRACE(type, trace_level, fmt_str)

extern UCHAR *fp_baseAddr;
extern UCHAR *fp_baseAddr_bar4;
#ifdef __FP_OS__

#define FP_MAX_DBGP 32
#define FP_MAX_DBGP_LEN 32
extern UCHAR fp_dbgp_arr[FP_MAX_DBGP][FP_MAX_DBGP_LEN];
extern UINT fp_dbgp_flags;
extern UINT fp_trace_level;

enum fp_dbgp_features {
        DBGP_FEAT_HIF     = 0x1,
        DBGP_FEAT_BRIDGE  = 0x2,
        DBGP_FEAT_ROUTE   = 0x4,
        DBGP_FEAT_FWD_PE  = 0x8,
        DBGP_FEAT_TMU     = 0x10,
        DBGP_FEAT_CLI     = 0x20,
        DBGP_FEAT_L2CLI   = 0x40,
        DBGP_FEAT_L3CLI   = 0x80,
        DBGP_FEAT_L2API   = 0x100,
        DBGP_FEAT_L3API   = 0x200,
        DBGP_FEAT_IOCTL   = 0x400,
        DBGP_FEAT_LIB     = 0x800,
        DBGP_FEAT_ERR     = 0x1000,
        DBGP_FEAT_ETHTOOL = 0x2000,
        DBGP_FEAT_PROCFS  = 0x4000,
        DBGP_FEAT_DRV_MCU = 0x8000,

        /*
         * This is an execution time feature, so max number of allowed
         * groups are 0-31. You may add your features before this comment.
         */
        DBGP_FEAT_MAX = 0x80000000
};

#define FP_DEBUG(feat, level, msg, args...) if(feat & fp_dbgp_flags){ \
    int count = 0; \
    for (; count < 32; count++) { \
        if ((1 << count) & feat) { \
            break; \
        } \
    } \
    if (level <= fp_trace_level) \
    printk("%s: " msg,fp_dbgp_arr[count], ## args); \
};

#define FP_DEBUG_PRINT_HEX_DUMP( feat, level, msg, dump_flag, col, bytes, data, len, ascii ) \
    if( feat & fp_dbgp_flags ){ \
        int count = 0; \
        for ( ; count < 32; count++ ) { \
            if ( ( 1 << count ) & feat ) { \
                break; \
            } \
        } \
        if (level <= fp_trace_level) \
        print_hex_dump( KERN_NOTICE, msg, dump_flag, col, bytes, data, len, ascii ); \
    }
#else
#define FP_DEBUG(type, debug_level, msg, args...)
#define FP_DEBUG_PRINT_HEX_DUMP( feat, level, msg, dump_flag, col, bytes, data, len, ascii )
#endif /*__FP_OS__*/

#endif /*End of _FP_LIBRARY_H_ File */
