/*************************************************************************/ /*
@File
@Title          fp_switch_api.h
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    Structures for hw switch
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/*!
 * @mainpage NPU L2 Fast Path Control API
 *
 * Copyright (c) Milandr. All Rights Reserved
 *
 * The L2 Fast Path Control API is the primary method that the host uses
 * to configure the NPU L2 switch hardware. The API provides access to the
 * MAC table, the Bridge Domain table, and the switch ports.
 *
 * @file fp_switch_api.h
 * API for L2 Bridge Domain table and global configuration
 */

#ifndef  _FP_HW_SW_API_H_
#define  _FP_HW_SW_API_H_

#include "fp_port.h"

/**********************************************************************
 *                                      #defines
 **********************************************************************/
#if 0
#define BRENTRY_FWD_PORT_LIST_START_POS        0
#define BRENTRY_UNTAG_LIST_START_POS           8
#define BRENTRY_UCAST_HIT_ACT_START_POS        16
#define BRENTRY_MCAST_HIT_ACT_START_POS        19
#define BRENTRY_UCAST_MISS_ACT_START_POS       22
#define BRENTRY_MCAST_MISS_ACT_START_POS       25
#define BRENTRY_MSTP_HIT_ACT_START_POS         28
#define BRENTRY_RESERVED_BIT_MASK_START_POS    31
#endif

#define BRENTRY_FWDPORT                        1
#define BRENTRY_UNTAGLIST                      2

//#define BRENTRY_UCAST_HIT                      3
//#define BRENTRY_MCAST_HIT                      4
//#define BRENTRY_UCAST_MISS                     5
//#define BRENTRY_MCAST_MISS                     6
//#define BRENTRY_MSTP                           7

#define BRENTRY_UCAST_HIT                      0
#define BRENTRY_MCAST_HIT                      1
#define BRENTRY_UCAST_MISS                     2
#define BRENTRY_MCAST_MISS                     3
#define BRENTRY_MSTP                           4


#define PORT_BASE_ADDR                         0x400
#define GLOBAL_BR_ENTRY_ADDR                   0x48c

/*!
 Add an bridge domain entry into bridge domain hash table.

 @param   wsp_base_addr   base address of the WSP registers
 @param   vlanId  Value in the range of [0 - 4093]
 @param   fwdPortList     Max 8 ports, bit mapped to port from rt to lt [0-0xFF]
 @param   untagList       Max 8 ports, bit mapped to port from rt to lt [0-0xFF]
 @param   ucastHitAct     unicast hit action (see enum fwd_action_e)
 @param   mcastHitAct     multicast hit action (see enum fwd_action_e)
 @param   ucastMissAct    unicast miss action (see enum fwd_action_e)
 @param   mcastMissAct    multicast hit action (see enum fwd_action_e)
 @param   mstpAct         mstp action (see enum fwd_action_e)
 @return status code:       0=success, -1=failure
 */
extern int32_t fp_br_entry_add(uint32_t vlanId, uint32_t fwdPortList, uint32_t untagList, uint32_t ucastHitAct, uint32_t mcastHitAct, uint32_t ucastMissAct, uint32_t mcastMissAct, uint32_t mstpAct);

/*!
 Update ports in an existing bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId         [0 - 4093]
 @param PortList       Max 8 ports [0-0xFF], bit mapped to port from rt to lt
 @param UnTagList      Max 8 ports [0-0xFF], bit mapped to port from rt to lt [0-0xFF]
 @return status code:       0=success, -1=failure
 **/
extern int32_t fp_br_entry_add_ports(uint32_t vlanId, uint32_t portList, uint32_t untagList);


/*!
 Get ports list from bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId         [0 - 4093]
 @return PortList Max 17 ports [0-0x1FFFF], bit mapped to port from rt to lt
 **/
extern int32_t fp_br_entry_portlist_get(uint32_t vlanId);

/*!
 Search for a bridge entry from bridge domain.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId         [0 - 4093]
 @return bridge entry
 **/
extern int64_t fp_br_entry_search(uint32_t vlanId);


/*!
 Get port's untag list from bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId         [0 - 4093]
 @return UnTagList Max 8 ports [0-0xFF], bit mapped to port from rt to lt [0-0xFF]
 **/
extern int32_t fp_br_entry_untaglist_get(uint32_t vlanId);


/*!
 Change port's tag value.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId    [0 - 4093]
 @param portList       specifies bit mapped port num from [0 - 0x1FFFF]
 @param Tagvalue       mapped to each port tag/untag list from [0 - 0x1FFFF]
 @return status code:       0=success, -1=failure
 **/
extern int32_t fp_br_entry_port_tag_chg(uint32_t vlanId, uint32_t portList, uint32_t TagValue);

/*!
 Get fwd action entry from bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId   [0 - 4093]
 @return  bit-coded bridge domain action entry (see below)

 @code
   VLAN (BD) Hash table action entry map:
   =======================================
    [07:00]         - forward port list
    [15:08]         - untag list
    [18:16]         - ucast_hit_action
    [21:19]         - mcast_hit_action
    [24:22]         - ucast_miss_action
    [27:25]         - mcast_miss_action
    [30:28]         - mstp
    [31:31]         - Not used yet
   =======================================
 @endcode

 **/
extern int64_t fp_br_entry_fwd_action_get(uint32_t vlanId);

/*!
 Get port's config.

 @param wsp_base_addr  base address of the WSP registers
 @param portNo   specifies bit mapped port num from [0 - 0xFF]
 @param pPortInfo[out]  port configuration
 @return status code:       0=success, -1=failure
 **/
extern int32_t fp_port_config_get(uint8_t *dev, uint32_t portNo, struct port_s *pPortInfo);

/*!
 Get global (fallback) bridge domain entry's data.

 @param wsp_base_addr  base address of the WSP registers
 @param pGlobal_br_entry[out]    Returns global config structure
 @return status code:       0=success, -1=failure
 **/
extern int32_t fp_global_br_entry_config_get(struct global_s *pGlobal_br_entry);


/*!
 Change action fields of an existing bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId          [0 - 4093]
 @param field           BRENTRY_UCAST_HIT/BRENTRY_MCAST_HITBRENTRY_UCAST_MISS/BRENTRY_MCAST_MISS/BRENTRY_MSTP
 @param field_value     value should be in the range of [0- 5]
 @return status code   0=success, -1=failure
 **/
extern int32_t fp_br_entry_fwd_action_change(uint32_t vlanId, uint32_t field, uint32_t value);


/*!
 Set port configuration.

 @param wsp_base_addr  base address of the WSP registers
 @param portNo         specifies bit mapped port num from [0 - 0xFF]
 @param pPort          port configuration info
 @return status code   0=success, -1=failure
 **/
extern int32_t fp_port_config(uint32_t portNo, struct port_s *pPort);
extern int32_t fp_pvlan_config(int32_t add_config, uint32_t portNo, struct port_s *pPort);

/*!
 Configure global (fallback) bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param pGlobal_br_entry  global configuration info
 @return status code   0=success, -1=failure
 **/
extern int32_t fp_global_br_entry_config(struct global_s *pGlobal_br_entry);


/*!
 Delete a bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId        [0 - 4093]
 @return status code   0=success, -1=failure
 **/
extern int32_t fp_br_entry_del (uint32_t vlanId);


/*!
 Delete ports from an existing bridge domain entry.

 @param wsp_base_addr  base address of the WSP registers
 @param vlanId          [0 - 4093]
 @param portList        Max 8 ports [0-0xFF], bit mapped to port from rt to lt
 @return status code   0=success, -1=failure
 **/
extern int32_t fp_br_entry_del_ports(uint32_t vlanId, uint32_t portList);

/*!
 Configure static port initialization information.

 @param wsp_base_addr  base address of the WSP registers
 @param mac char pointer
 @param vlanId          [0 - 4093]
 @param fwd_ports       Max 8 ports [0-0xFF], bit mapped to port from rt to lt
 @param hit_action      [0 - 6]
 @param tc              [0 - 7]
 @return status code   0=success, -1=failure
 **/

int32_t fp_static_mac_entry_add(uint8_t *mac, uint32_t vlanid,
		uint32_t fwd_ports, uint32_t hit_action, uint32_t tc);

/*!
 Configure port initialization information.

 @param *dev char pointer
 @return int
 **/

extern int32_t fp_mac_entry_add(uint32_t phy_port, uint8_t *mac, uint32_t vlanid);

/*!
 Adds MAC entry into mac hash table

 @param wsp_base_addr base address of the WSP registers
 @param vlanId          [0 - 4093]
 @param phy_port          [0 - 1]
 @param mac               [0 - 255]
 @return status code 0=success, -1=failure
**/

extern int32_t fp_mac_entry_del(uint8_t *mac, uint32_t vlanid);
/*!
 Delete MAC entry from mac hash table

 @param wsp_base_addr base address of the WSP registers
 @param vlanId          [0 - 4093]
 @param mac               [0 - 255]
 @return status code 0=success, -1=failure
**/

extern int32_t fp_mac_entry_search(uint8_t *mac, uint32_t vlanid);
/*!
 Search MAC entry from mac hash table

 @param wsp_base_addr base address of the WSP registers
 @param vlanId          [0 - 4093]
 @param mac               [0 - 255]
 @return status code 0=success, -1=failure
**/

extern int32_t fp_mac_entry_update(uint32_t mac_entry, uint8_t *mac, uint32_t vlanid);
/*!
 Update MAC entry from mac hash table

 @param wsp_base_addr base address of the WSP registers
 @param vlanId          [0 - 4093]
 @param mac               [0 - 255]
 @param mac_entry               [0 - 4093]
 @return status code 0=success, -1=failure
**/

extern uint32_t fp_ltc_gettime(uint32_t dev_number, uint32_t ch_num);
/*!
 Getting ltc time difference between channel head

 @param wsp_base_addr base address of the WSP registers
 @param dev_number	[0-3]
 @param ch_num		[0-7]
 @return time diffvalue of time
**/

extern int32_t fp_ltc_hif_config_time(uint32_t devId, uint32_t value, uint32_t type);
/*!
 Config ltc time seconds and nano seconds

 @param devId          [0 - 3]
 @param value          [Unsigned int min to max value]
 @param type           [1 - 2]
 @return status code 0=success, -1=failure
**/

void fp_portcfg_init(uint8_t *dev);

#endif /*End of _FP_HW_SW_API_H_ */
