/*************************************************************************/ /*!
@File
@Title          fp_switch_api.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    contains switch API functions
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                      Include Files
 **********************************************************************/
 #include <fp_types.h>
 #include <fp_library.h>
 #include <fp_macros.h>
 #include <fp_port.h>
 #include <fp_switch_api.h>
 #include <switchApi_blocks_ctrlm.h>
 #include <fp_common_ctlm.h>
 #include <fp_switch.h>
 #include <hw_2field_hash_table.h>
 #include <hw_vlan_hash_table.h>
 #include <fp_ltc.h>
 #include <wsp_vlan_hash_table_L.h>

/**********************************************************************
 *                                      #defines
 **********************************************************************/


/**********************************************************************
 *                              Extern Declarations
***********************************************************************/

/* VLAN (BD) Hash table action entry map:*
*=======================================*
* [19:00]         - forward port list
* [39:20]         - untag list
* [42:40]         - ucast_hit_action
* [45:43]         - mcast_hit_action
* [48:46]         - ucast_miss_action
* [51:49]         - mcast_miss_action
* [54:52]         - mstp
* [64:55]         - Not used yet
*=======================================*/

/*******************************************************************
 * Function Name  : fp_br_entry_add
 * Description    : Adds an bridge entry into vlan hash table
 * Inputs         :
 * Parameters     : vlanId, fwdPortList, untagList, ucastHitAct, mcastHitAct,
 *                  ucastMissAct, mcastMissAct, mstpAct
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_add(uint32_t vlanId, uint32_t fwdPortList,
		uint32_t untagList, uint32_t ucastHitAct, uint32_t mcastHitAct,
		uint32_t ucastMissAct, uint32_t mcastMissAct, uint32_t mstpAct)
{
	uint32_t field_entries[5];
	int64_t action_entry = 0, i = 0;

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	for (i = 0; i < FP_MAX_PORTS; i++) {
		if (((fwdPortList >> i) & 0x1)) {
			if (!((FWD_PORT_LIST_MASK >> i) & 0x1)) {
				return SWITCHAPI_ERR_MSG_FWDPORTLIST;
			}
		}
	}

	/*untag list validation*/
	for (i = 0; i < FP_MAX_PORTS; i++) {
		if (((untagList >> i) & 0x1)) {
			if (!((FWD_PORT_LIST_MASK >> i) & 0x1)) {
				return SWITCHAPI_ERR_MSG_UNTAGLIST;
			}
		}
	}

	/*unicast hit action validation*/
	if (ucastHitAct < SWITCHAPI_MIN_UCASTHITACT || ucastHitAct > SWITCHAPI_MAX_UCASTHITACT) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_UCASTHITACT %d\n", __func__, ucastHitAct);
		return SWITCHAPI_ERR_MSG_UCASTHITACT;
	}

	/*multicast hit action validation*/
	if (mcastHitAct < SWITCHAPI_MIN_MCASTHITACT || mcastHitAct > SWITCHAPI_MAX_MCASTHITACT) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_MCASTHITACT %d\n", __func__, mcastHitAct);
		return SWITCHAPI_ERR_MSG_MCASTHITACT;
	}

	/*unicast miss action validation*/
	if (ucastMissAct < SWITCHAPI_MIN_UCASTMISSACT || ucastMissAct > SWITCHAPI_MAX_UCASTMISSACT) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_UCASTMISSACT %d\n", __func__, ucastMissAct);
		return SWITCHAPI_ERR_MSG_UCASTMISSACT;
	}

	/*multicast miss action validation*/
	if (mcastMissAct < SWITCHAPI_MIN_MCASTMISSACT || mcastMissAct > SWITCHAPI_MAX_MCASTMISSACT) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_MCASTMISSACT %d\n", __func__, mcastMissAct);
		return SWITCHAPI_ERR_MSG_MCASTMISSACT;
	}

	/*mstp action validation*/
	if (mstpAct < SWITCHAPI_MIN_MCASTMISSACT || mstpAct > SWITCHAPI_MAX_MCASTMISSACT) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_MSTPACT %d\n", __func__, mstpAct);
		return SWITCHAPI_ERR_MSG_MSTPACT;
	}

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"ADD:vlanId: %d, fwdPortList: 0x%x, untaglist: 0x%x\n",
			vlanId, fwdPortList, untagList);
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"ucastHit:%x, mcastHit:%x\n",
			ucastHitAct, mcastHitAct);
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"ucastMissAct:%x, mcastMissAct:%x, mstpAct:%x\n",
			ucastMissAct, mcastMissAct, mstpAct);
	field_entries[0] = vlanId;
	action_entry = (((fwdPortList << BRENTRY_FWD_PORT_LIST_START_POS) & BRENTRY_FWD_PORT_LIST_MASK) |
					(((uint64_t)untagList << BRENTRY_UNTAG_LIST_START_POS) & BRENTRY_UNTAG_LIST_MASK) |
					(((uint64_t)ucastHitAct << BRENTRY_UCAST_HIT_ACT_START_POS) & BRENTRY_UCAST_HIT_ACT_MASK) |
					(((uint64_t)mcastHitAct << BRENTRY_MCAST_HIT_ACT_START_POS) & BRENTRY_MCAST_HIT_ACT_MASK) |
					(((uint64_t)ucastMissAct << BRENTRY_UCAST_MISS_ACT_START_POS) & BRENTRY_UCAST_MISS_ACT_MASK) |
					(((uint64_t)mcastMissAct << BRENTRY_MCAST_MISS_ACT_START_POS) & BRENTRY_MCAST_MISS_ACT_MASK) |
					(((uint64_t)mstpAct << BRENTRY_MSTP_HIT_ACT_START_POS) & BRENTRY_MSTP_HIT_ACT_MASK)
				   );

	return wsp_vlan_entry_add(vlanId, action_entry);
}

/*******************************************************************
 * Function Name  : fp_br_entry_add_ports
 * Description    : Add ports to existing bridge entry
 * Inputs         :
 * Parameters     : vlanId, PortList, untagList
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_add_ports(uint32_t vlanId, uint32_t portList, uint32_t untagList)
{
	int64_t action_entry = 0, tmp_portlist, tmp_untagList;


	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	/*port list validation*/
	if (portList < SWITCHAPI_MIN_PORT_LIST || portList > SWITCHAPI_MAX_PORT_LIST) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_PORTLIST %d\n", __func__, portList);
		return SWITCHAPI_ERR_MSG_PORTLIST;
	}

	/*untag list validation*/
	if (untagList < SWITCHAPI_MIN_UNTAG_LIST || untagList > SWITCHAPI_MAX_UNTAG_LIST) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_UNTAGLIST %d\n", __func__, untagList);
		return SWITCHAPI_ERR_MSG_UNTAGLIST;
	}

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
		"ADDPORTS:vlanId: %d, fwdPortList: 0x%x, untaglist: 0x%x\n",
				vlanId, portList, untagList);

	action_entry = wsp_vlan_entry_search(vlanId);

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "action entry: 0x%llx\n", action_entry);
	if(action_entry == -1){
		return SWITCHAPI_ERR_MSG_NOENTRY;
	}

	tmp_portlist = portList;
	tmp_untagList = untagList;

	/* clear the target feilds in action entry */
	//action_entry = (INT64)(action_entry & ~(portList | (portList << BRENTRY_UNTAG_LIST_START_POS)));
	action_entry = (INT64)(action_entry & ~(tmp_portlist | (tmp_portlist << BRENTRY_UNTAG_LIST_START_POS)));

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "after clear action entry: 0x%llx\n", action_entry);

	/* update the target feilds */
	action_entry =	  action_entry |
					((tmp_portlist & FWD_PORT_LIST_MASK) << BRENTRY_FWD_PORT_LIST_START_POS ) |
					((tmp_untagList & FWD_PORT_LIST_MASK) << BRENTRY_UNTAG_LIST_START_POS );

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "modified action entry: 0x%llx\n", action_entry);

	return wsp_vlan_entry_update(vlanId, action_entry);

}

/*******************************************************************
 * Function Name  : fp_br_entry_portlist_get
 * Description    : Get ports list from bridge entry
 * Inputs         :
 * Parameters     : vlanId
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_portlist_get(uint32_t vlanId)
{
	int64_t action_entry = 0;

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	action_entry = wsp_vlan_entry_search(vlanId);
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "action_entry: 0x%llx\r\n", action_entry);

	if (action_entry == -1)
		return 0;

	action_entry = action_entry & FWD_PORT_LIST_MASK;
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "GETPORTS: vlanId: %x, fwdPortList: %x\r\n", vlanId, (UINT)action_entry);
	return (int32_t)action_entry;
}

/*******************************************************************
 * Function Name  : fp_br_entry_search
 * Description    : Get bridge entry from bridge table
 * Inputs         :
 * Parameters     : vlanId
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int64_t fp_br_entry_search(uint32_t vlanId)
{
	int64_t action_entry = 0;

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	action_entry = wsp_vlan_entry_search(vlanId);

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s: search entry: 0x%llX\n", __func__, action_entry);

	return action_entry;
}

/*******************************************************************
 * Function Name  : fp_br_entry_untaglist_get
 * Description    : Get port's untag list from bridge entry
 * Inputs         :
 * Parameters     : vlanId
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_untaglist_get(uint32_t vlanId)
{
	int64_t action_entry = 0;

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	action_entry = wsp_vlan_entry_search(vlanId);
	if(action_entry == -1){	//if entry not found
		action_entry = 0;
	}

	action_entry = ((action_entry >> BRENTRY_UNTAG_LIST_START_POS) & FWD_PORT_LIST_MASK);
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"GETUNTAGLIST:vlanId:%x, untaglist:%llx\r\n",
			vlanId, action_entry);
	return action_entry;
}

/*******************************************************************
 * Function Name  : fp_br_entry_port_tag_chg
 * Description    : change port's tag value
 * Inputs         :
 * Parameters     : vlanId, portNo and Tagvalue
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_port_tag_chg(uint32_t vlanId, uint32_t portList, uint32_t TagValue)
{
	int64_t action_entry = 0;
	int64_t tmp_portList, tmp_TagValue;

	action_entry = wsp_vlan_entry_search(vlanId);

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "PORTTAGCHG\n");
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "action entry:%llx\n", action_entry);
	/* CHK: Is the port is already in the fwd port list */
	if (portList & ~FWD_PORT_LIST_MASK){
		return -1;
	}
	/* clear the target feilds in action entry */
	tmp_portList = portList;
	tmp_TagValue = TagValue;

	action_entry = (action_entry & ~(tmp_portList << BRENTRY_UNTAG_LIST_START_POS));

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "after clear action entry:%llx\r\n", action_entry);
	/* update the target feilds */
	action_entry = action_entry | ((tmp_TagValue & tmp_portList) << BRENTRY_UNTAG_LIST_START_POS);

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "modified action entry:%llx\r\n", action_entry);

	return wsp_vlan_entry_update(vlanId, action_entry);
}

/*******************************************************************
 * Function Name  : fp_br_entry_fwd_action_get
 * Description    : Get fwd action entry from bridge entry
 * Inputs         :
 * Parameters     : vlanId
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int64_t fp_br_entry_fwd_action_get(uint32_t vlanId)
{
	int64_t action_entry = 0;

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID){
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "entered in fp_br_entry_fwd_action_get function\n");
	action_entry = wsp_vlan_entry_search(vlanId);
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "GETACT:vlanId:%x, fwdact:%llx \r\n", vlanId, action_entry);

	return action_entry;
}

/*******************************************************************
 * Function Name  : fp_port_config_get
 * Description    : To get port's config
 * Inputs         :
 * Parameters     : portNo, struct port_s *
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_port_config_get(uint8_t *dev,
				uint32_t portNo, struct port_s *pPortInfo)
{
	/* UINT port_addr = PORT_BASE_ADDR +
			(portNo * sizeof(struct port_s)); */
	int32_t ret = 0;

	/*port number validation*/
	if (portNo < SWITCHAPI_MIN_PORT_NO || portNo > SWITCHAPI_MAX_PORT_NO){
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_PORTNO %d\n", __func__, portNo);
		return SWITCHAPI_ERR_MSG_PORTNO;
	}

	portNo--;

	/* FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "GETPORTCFG:\r\n"); */
	if (pPortInfo != NULL){
		fp_memcpy(pPortInfo, &fp_port[portNo], sizeof(struct port_s));
		/* fp_read_hw_data(wsp_base_addr, (UCHAR *)pPortInfo,
					port_addr, sizeof(struct port_s)); */
		/* FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"portNo:%x, shutdown:%x, tpid:%x\r\n", portNo,
			pPortInfo->shutdown, pPortInfo->tpid); */
		/* FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"aft:%x, fallback_bd_id:%x, blockstate:%x\r\n",
			pPortInfo->aft, pPortInfo->fallback_bd_id,
			pPortInfo->block_state); */
	}else{
		ret = -1;
	}

	return ret;
}

/*******************************************************************
 * Function Name  : fp_global_br_entry_config_get
 * Description    : To get global bridge entry's data
 * Inputs         :
 * Parameters     : struct global_s *
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_global_br_entry_config_get(struct global_s *pGlobal_br_entry)
{
	/* UINT global_addr = GLOBAL_BR_ENTRY_ADDR; */
	int32_t ret = 0;

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "GETGLOBAL:\r\n");
	if (NULL != pGlobal_br_entry) {

		fp_memcpy(pGlobal_br_entry, &fp_global, sizeof(struct global_s));
		/* fp_read_hw_data(wsp_base_addr, (UCHAR *)pGlobal_br_entry, global_addr, sizeof(struct global_s)); */

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "In function fp_global_br_entry_config_get: l2splpunt: %d,\n\tfwd_list: 0x%X,\n\tuntag_list: 0x%X,\n\tmstpAct: 0x%X,\n\tUCAST_HIT_ACT: %d,\n\tMCAST_HIT_ACT: %d,\n\tUCAST_MISS_ACT: %d,\n\tMCAST_MISS_ACT: %d,\n\ttc: 0x%X\n",
												pGlobal_br_entry->l2_special_punt_enable,
												pGlobal_br_entry->fallback_bd_entry.forward_list,
												pGlobal_br_entry->fallback_bd_entry.untag_list,
												pGlobal_br_entry->fallback_bd_entry.mstp_action,
												pGlobal_br_entry->fallback_bd_entry.ucast_hit_action,
												pGlobal_br_entry->fallback_bd_entry.mcast_hit_action,
												pGlobal_br_entry->fallback_bd_entry.ucast_miss_action,
												pGlobal_br_entry->fallback_bd_entry.mcast_miss_action,
												pGlobal_br_entry->fallback_bd_entry.tc);
	} else {
		ret = -1;
	}
	return ret;
}


/*******************************************************************
 * Function Name  : fp_br_entry_fwd_action_change
 * Description    : Changes action feilds of an existing bridge entry
 * Inputs         :
 * Parameters     : vlanId, field, field_value
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_fwd_action_change(uint32_t vlanId, uint32_t field, uint32_t value)
{
	int64_t action_entry = 0;
	int32_t ret = 0;

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "FWD_ACT_CHG:\r\n");
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "vlan:%x,field:%x,value:%x\r\n",
							vlanId, field, value);

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	/*field validation*/
	if (field < SWITCHAPI_MIN_FIELD || field > SWITCHAPI_MAX_FIELD) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_FIELD %d\n", __func__, field);
		return SWITCHAPI_ERR_MSG_FIELD;
	}

	/*value validation*/
	if (value < SWITCHAPI_MIN_VALUE || value > SWITCHAPI_MAX_VALUE) {
 		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VALUE %d\n", __func__, value);
		return SWITCHAPI_ERR_MSG_VALUE;
	}

	action_entry = wsp_vlan_entry_search(vlanId);
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"before action entry:%llx\r\n", action_entry);
	if (action_entry != -1) {
		switch (field) {
			/* clear the targetted feilds and set the new values */
#if 0 /* These two are not actions */
		case BRENTRY_FWDPORT:
			action_entry = (action_entry & ~FWD_PORT_LIST_MASK) |
				(value << BRENTRY_FWD_PORT_LIST_START_POS);
		break;
		case BRENTRY_UNTAGLIST:
		action_entry = (action_entry & ~BRENTRY_UNTAG_LIST_MASK) |
				(value << BRENTRY_UNTAG_LIST_START_POS);
		break;
#endif
		case BRENTRY_UCAST_HIT:
			action_entry =
				(action_entry & ~BRENTRY_UCAST_HIT_ACT_MASK) |
				((uint64_t)value << BRENTRY_UCAST_HIT_ACT_START_POS);
			break;
		case BRENTRY_MCAST_HIT:
			action_entry =
				(action_entry & ~BRENTRY_MCAST_HIT_ACT_MASK) |
				((uint64_t)value  << BRENTRY_MCAST_HIT_ACT_START_POS);
			break;
		case BRENTRY_UCAST_MISS:
			action_entry =
				(action_entry & ~BRENTRY_UCAST_MISS_ACT_MASK) |
				((uint64_t)value << BRENTRY_UCAST_MISS_ACT_START_POS);
			break;
		case BRENTRY_MCAST_MISS:
			action_entry =
				(action_entry & ~BRENTRY_MCAST_MISS_ACT_MASK) |
				((uint64_t)value  << BRENTRY_MCAST_MISS_ACT_START_POS);
			break;
		case BRENTRY_MSTP:
			action_entry =
				(action_entry & ~BRENTRY_MSTP_HIT_ACT_MASK) |
				((uint64_t)value << BRENTRY_MSTP_HIT_ACT_START_POS);
			break;
		default:
			break;
		}
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
				"after action entry:%llx\r\n", action_entry);
		ret = wsp_vlan_entry_update(vlanId,	action_entry);
	}else{
		ret = -1;
	}

	return ret;
}

/* update vlan table entries with private vlan information */
void fp_pvlan_update_vlan( uint32_t add_config, uint32_t portNo, uint32_t vlanId,
					int32_t is_fwd_mem, int32_t is_untag_mem ){
	int64_t ret = -1;
	uint64_t action_entry, action;
	uint32_t fwd_list = 0, untag_list = 0;

	if( is_fwd_mem ){
		fwd_list = ( 1 << portNo );
	}
	if( is_untag_mem ){
		untag_list = ( 1 << portNo );
	}
	ret = wsp_vlan_entry_search( vlanId );
	if ( ret == -1 ){
		action_entry = ( ( ( ( uint64_t )untag_list ) << BRENTRY_UNTAG_LIST_START_POS ) | fwd_list );
		if ( add_config )
        {
			wsp_vlan_entry_add( vlanId, action_entry );
		}
	}
    else
    {
		action = ( ( ( ( uint64_t )untag_list ) << BRENTRY_UNTAG_LIST_START_POS ) | fwd_list );
		if ( add_config ){
			action_entry = ( ret | action );
		}
		else{
			action_entry = ( ret & ~action );
		}
		wsp_vlan_entry_update( vlanId, action_entry );
	}
}

/* Delete mac table entries(old config) when private vlan configuration is removed */
static int32_t fp_pvlan_del_mac_entries(uint32_t portNo, struct port_s *pPort)
{
	uint32_t action_entry = 0, fwd_list = 0, port_num = 0, pvid = 0, i;
    uint32_t hashidx = 0, vlan_valid = 0;
    uint32_t coll_ptr = 0;
    uint8_t mac[6];
    uint32_t data[4];

	for( hashidx = 0; hashidx <  HASH_CNT_PER_ITERATION; hashidx++ ){
start:
		coll_ptr = vlan_valid = 0;

        hw_hash_table_read_entry( data, hashidx );

		if( !( HASH_ENTRY_VALID_FLAG & data[ REG_MAC4 ] ) ){
			continue;
		}

		if( data[ REG_MAC4 ] & HASH_ENTRY_COLL_FLAG ){
			coll_ptr = ( data[ REG_MAC4 ] >> 8 ) & 0xFFFF;
		}

		if( HASH_ENTRY_PVLAN_FLAG & data[ REG_MAC3 ] ){
			action_entry = ((data[REG_MAC3] << 4) | (data[REG_MAC2] >> 29));
			pvid = ((data[REG_MAC2] & 0x1FFF0000) >> 16);
			fwd_list = action_entry & BRENTRY_FWD_PORT_LIST_MASK;

			for (i = 0; i < FP_MAX_SWITCH_PORTS; i++)
            {
				if ((FWD_PORT_LIST_MASK >> i) & 0x1)
                {
					if ((fwd_list >> i) & 0x1)
                    {
						port_num = i;
						break;
					}
				}
			}

			if ((portNo == port_num) && (pPort->pvlan_pri == pvid))
            {
                mac[0] = (data[REG_MAC1] >> 24) & 0xFF;
                mac[1] = (data[REG_MAC1] >> 16) & 0xFF;
                mac[2] = (data[REG_MAC1] >> 8) & 0xFF;
                mac[3] = data[REG_MAC1] & 0xFF;
                mac[4] = (data[REG_MAC2] >> 8) & 0xFF;
                mac[5] = data[REG_MAC2] & 0xFF;

                hw_hash_table_delete(mac, pvid);

				fp_rc_stats.hw_hash_entry_count--;

				hw_2field_pvlan_entries_delete(mac, pvid, action_entry);

				if (coll_ptr){
					goto start;
				}
			}
		}

		while (coll_ptr)
        {
            hw_hash_table_read_entry(data, coll_ptr);

			vlan_valid = 0;

			if (!(HASH_ENTRY_VALID_FLAG & data[REG_MAC4])){
				coll_ptr = 0;
				continue;
			}

			if (HASH_ENTRY_PVLAN_FLAG & data[REG_MAC3]){

				action_entry = ((data[REG_MAC3] << 4) | (data[REG_MAC2] >> 29));
				pvid = ((data[REG_MAC2] & 0x1FFF0000) >> 16);
				fwd_list = action_entry & BRENTRY_FWD_PORT_LIST_MASK;

				for (i = 0; i < FP_MAX_SWITCH_PORTS; i++){

					if ((FWD_PORT_LIST_MASK >> i) & 0x1)
                    {
						if ((fwd_list >> i) & 0x1)
                        {
							port_num = i;
							break;
						}
					}
				}

				if ((portNo == port_num) && (pPort->pvlan_pri == pvid)){

                    mac[0] = (data[REG_MAC1] >> 24) & 0xFF;
                    mac[1] = (data[REG_MAC1] >> 16) & 0xFF;
                    mac[2] = (data[REG_MAC1] >> 8) & 0xFF;
                    mac[3] = data[REG_MAC1] & 0xFF;
                    mac[4] = (data[REG_MAC2] >> 8) & 0xFF;
                    mac[5] = data[REG_MAC2] & 0xFF;

                    hw_hash_table_delete(mac, pvid);
					fp_rc_stats.hw_hash_entry_count--;
					hw_2field_pvlan_entries_delete(mac, pvid, action_entry);
				}
			}

			coll_ptr = 0;
			if (data[REG_MAC4] & HASH_ENTRY_COLL_FLAG){
				coll_ptr = (data[REG_MAC4] >> 8) & 0xFFFF;
			}

		}
	}

	return 0;
}

/* Private vlan configuration routine (Add/Delete) */
int32_t fp_pvlan_config( int32_t add_config, uint32_t portNo, struct port_s *pPort ){
	uint32_t addr1, addr2, addr1_val, addr2_val, index = 0, class;

	if( !pPort ){
		return 0;
	}

	if( ( portNo < SWITCHAPI_MIN_PORT_NO ) || ( portNo > SWITCHAPI_MAX_PORT_NO ) ) {
		FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_PORTNO %d\n", __func__, portNo );
		return SWITCHAPI_ERR_MSG_PORTNO;
	}
	portNo--;

	//fp_get_port_addr(0, portNo, &addr1, &addr2);

	// Update vlan configuration information
	/*
	switch (pPort->pvlan_type) {
		case ISOLATED:
			fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_pri, 1, 1);
			fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_sec[0], 0, 0);
			break;
		case COMMUNITY:
			fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_pri, 1, 1);
			fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_sec[0], 1, 1);
			break;
		case PROMISCUOUS:
			fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_pri, 1, 1);
			for (index = 0; index < pPort->num_sec; index++){
				fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_sec[index], 1, 1);
			}
			break;
		case INTERSWITCH:
			fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_pri, 1, 0);
			for (index = 0; index < pPort->num_sec; index++){
				fp_pvlan_update_vlan(add_config, portNo, pPort->pvlan_sec[index], 1, 0);
			}
			break;
		default:
			return 0;
	}
	*/

	// Update port configuration information
	if( add_config == 0 ) {
		pPort = &fp_port[ portNo ];

		switch( pPort->pvlan_type ){
			case ISOLATED:
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 1 );
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ 0 ], 0, 0 );
				break;
			case COMMUNITY:
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 1 );
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ 0 ], 1, 1 );
				break;
			case PROMISCUOUS:
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 1 );
				for( index = 0; index < pPort->num_sec; index++ ){
					fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ index ], 1, 1 );
				}
				break;
			case INTERSWITCH:
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 0 );
				for( index = 0; index < pPort->num_sec; index++ ){
					fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ index ], 1, 0 );
				}
				break;
			default:
				return 0;
		}
		/* Delete mac entries related to private vlan */
		fp_pvlan_del_mac_entries( portNo, pPort );

		/* Delete vlan entries related to private vlan, if all pvlan ports disable */
		/*
		dont_delete_pri_flag = 0;
		for(i = 0; i < FP_MAX_SWITCH_PORTS; i++){
			if(i != portNo){
				if( fp_port[i].pvlan_enable == 1){
					if(fp_port[i].pvlan_pri == pPort->pvlan_pri){
						dont_delete_flag = 1;
					}
				}
			}
		}
		if(!dont_delete_flag){
			fp_br_entry_del(pPort->pvlan_pri);
		}
		*/

		fp_port[ portNo ].pvlan_enable = 0;
		fp_port[ portNo ].num_sec = 0;
		fp_port[ portNo ].shutdown = 0;
		fp_port[ portNo ].tpid     = VLAN_TAGGED_FRAME;
		fp_port[ portNo ].aft      = ANY_TAGGING;
		fp_port[ portNo ].fallback_bd_id = DEF_PORT_FALLBACK_BD_ID;
		fp_port[ portNo ].block_state = FORWARDING;
		fp_port[ portNo ].csr_untag_from_btable = 0;

		for( class = 0; class < FP_CLASS_MAX_NUMBER; class++ ){
			fp_get_port_addr( class, portNo, &addr1, &addr2 );
			if( addr1 && addr2 ){

				fp_reg_write(	( uint32_t )addr1,
                        		( ( fp_port[ portNo ].fallback_bd_id << PORT_FALLBACK_BDID_START_POS ) & PORT_FALLBACK_BDID_MASK ) |
								( ( fp_port[ portNo ].tpid << PORT_TPID_START_POS ) & PORT_TPID_MASK ) );

				fp_reg_write(	( uint32_t )addr2,
								( ( fp_port[ portNo ].csr_untag_from_btable << PORT_UNTAG_FROM_BTABLE_START_POS ) & PORT_UNTAG_FROM_BTABLE_MASK ) |
								( ( fp_port[ portNo ].block_state << PORT_BLOCKSTATE_START_POS ) & PORT_BLOCKSTATE_MASK) |
								( ( fp_port[ portNo ].aft << PORT_AFT_START_POS ) & PORT_AFT_MASK) |
								( ( fp_port[ portNo ].shutdown << PORT_SHUTDOWN_START_POS ) & PORT_SHUTDOWN_MASK ) );
			}
		}
		return 0;
	}

	if( fp_port[portNo].pvlan_enable == 1 ){
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s: private vlan for port %d already configured! Need delete first and configure again.\n", __func__, portNo );
		return SWITCHAPI_ERR_MSG_FAILED_CFG;
	}

	switch( pPort->pvlan_type ){
		case ISOLATED:
			fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 1 );
			fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ 0 ], 0, 0 );
			break;
		case COMMUNITY:
			fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 1 );
			fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ 0 ], 1, 1 );
			break;
		case PROMISCUOUS:
			fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 1 );
			for( index = 0; index < pPort->num_sec; index++ ){
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ index ], 1, 1 );
			}
			break;
		case INTERSWITCH:
			fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_pri, 1, 0 );
			for( index = 0; index < pPort->num_sec; index++ ){
				fp_pvlan_update_vlan( add_config, portNo, pPort->pvlan_sec[ index ], 1, 0 );
			}
			break;
		default:
			return 0;
	}

	// Add Private Vlan Configuration
	fp_port[ portNo ].csr_untag_from_btable = 1;

	switch( pPort->pvlan_type ){
		case ISOLATED:
		case COMMUNITY:
			fp_port[ portNo ].aft = UNTAGGED_ONLY;
			fp_port[ portNo ].fallback_bd_id = pPort->pvlan_sec[ 0 ];
			break;
		case PROMISCUOUS:
			fp_port[ portNo ].aft = UNTAGGED_ONLY;
			fp_port[ portNo ].fallback_bd_id = pPort->pvlan_pri;
			break;
		case INTERSWITCH:
			fp_port[ portNo ].aft = TAGGED_ONLY;
			break;
		default:
			return 0;
	}

	fp_port[ portNo ].pvlan_enable = 1;
	fp_port[ portNo ].pvlan_type = pPort->pvlan_type;
	fp_port[ portNo ].pvlan_pri = pPort->pvlan_pri;
	fp_port[ portNo ].num_sec = pPort->num_sec;

	for( index = 0; index < pPort->num_sec; index++ ){
		fp_port[ portNo ].pvlan_sec[ index ] = pPort->pvlan_sec[ index ];
	}

	for( class = 0; class < FP_CLASS_MAX_NUMBER; class++ ){
		fp_get_port_addr( class, portNo, &addr1, &addr2 );
		if( addr1 && addr2 ){

			addr1_val = fp_reg_read( addr1 );
			addr2_val = fp_reg_read( addr2 );

			fp_reg_write(	( uint32_t )addr1,
							( ( fp_port[ portNo ].fallback_bd_id << PORT_FALLBACK_BDID_START_POS ) & PORT_FALLBACK_BDID_MASK ) |
							( addr1_val & PORT_TPID_MASK ) );

			fp_reg_write(	( uint32_t )addr2,
							( ( fp_port[ portNo ].csr_untag_from_btable << PORT_UNTAG_FROM_BTABLE_START_POS ) & PORT_UNTAG_FROM_BTABLE_MASK ) |
							( addr2_val & PORT_BLOCKSTATE_MASK) | ( ( fp_port[ portNo ].aft << PORT_AFT_START_POS ) & PORT_AFT_MASK ) |
							( addr2_val & PORT_SHUTDOWN_MASK ) );

		}
	}
	return 0;
}

/*******************************************************************
 * Function Name  : fp_port_config
 * Description    : Configures port
 * Inputs         :
 * Parameters     : portNo, struct posrt_s *
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_port_config(uint32_t portNo, struct port_s *pPort)
{
	int32_t ret = 0;
	uint32_t addr1, addr2, i = 0, class;

	/* UINT port_addr = PORT_BASE_ADDR +
			(portNo * sizeof(struct port_s)); */

	/*port number validation*/
	if (portNo < SWITCHAPI_MIN_PORT_NO || portNo > SWITCHAPI_MAX_PORT_NO) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_PORTNO %d\n", __func__, portNo);
		return SWITCHAPI_ERR_MSG_PORTNO;
	}

	portNo--;

	/* FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
				"PORTCFG: port_addr:%x\r\n", port_addr); */
	if ( (portNo < FP_MAX_SWITCH_PORTS) && pPort) {
		/* FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
				"portNo:%x, shutdown:%x, tpid:%x\r\n", portNo,
				pPort->shutdown, pPort->tpid); */
		/* FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
				"aft:%x, fallback_bd_id:%x, blockstate:%x\r\n",
				pPort->aft, pPort->fallback_bd_id,
				pPort->block_state); */

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
				"before memcpy\nshutdown = %x\ntpid = %x\n"
				"aft = %x\nfallback_bd_id = %x\n"
				"block_state = %x\ntrusted = %x\n"
				"tcsel_table = %x\npcp2tc_table = %x\n"
				"def_cfi = %x\ndef_pri = %x\ndef_tc = %x\n"
				"tc2cos_table = %x\npindex = %x\n"
				"csr_untag_from_btable = %x", pPort->shutdown,
				pPort->tpid, pPort->aft, pPort->fallback_bd_id,
				pPort->block_state, pPort->trusted,
				pPort->tcsel_table, pPort->pcp2tc_table,
				pPort->def_cfi, pPort->def_pri, pPort->def_tc,
				pPort->tc2cos_table, pPort->pindex,
				pPort->csr_untag_from_btable);

		fp_memcpy(&fp_port[portNo], pPort, sizeof(struct port_s));

		/* START */
		i = portNo;

		for(class = 0; class < FP_CLASS_MAX_NUMBER; class++){
			fp_get_port_addr(class, portNo, &addr1, &addr2);

			/* class hw port csr reg:
			*===============================================================*
			* res   | block_state | aft | shutdown | fallback_bd_id | tpid  *
			*===============================================================*
			* 20-bit| 4-bit       |4-bit| 1-bit    | 13-bit         |16-bit *
			*===============================================================*/

			if (addr1 && addr2) {

				fp_reg_write(	(uint32_t)addr1, 
								((fp_port[i].fallback_bd_id << PORT_FALLBACK_BDID_START_POS) & PORT_FALLBACK_BDID_MASK) | 
								((fp_port[i].tpid << PORT_TPID_START_POS) & PORT_TPID_MASK) );

				fp_reg_write( 	(uint32_t)addr2,
								((fp_port[i].csr_untag_from_btable << PORT_UNTAG_FROM_BTABLE_START_POS) & PORT_UNTAG_FROM_BTABLE_MASK) |
								((fp_port[i].block_state << PORT_BLOCKSTATE_START_POS) & PORT_BLOCKSTATE_MASK)        |
								((fp_port[i].aft << PORT_AFT_START_POS) & PORT_AFT_MASK) |
								((fp_port[i].shutdown << PORT_SHUTDOWN_START_POS) & PORT_SHUTDOWN_MASK) );

			}
		}

		/* END */

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"after memcpy\nshutdown = %d\ntpid = 0x%X\naft = %d\n"
			"fallback_bd_id = 0x%X\nblock_state = %d\ntrusted = %d\n"
			"tcsel_table = %d\npcp2tc_table = %d\ndef_cfi = %d\n"
			"def_pri = %d\ndef_tc = %d\ntc2cos_table = %d\n"
			"pindex = 0x%X\ncsr_untag_from_btable = %d",
			fp_port[portNo].shutdown, fp_port[portNo].tpid,
			fp_port[portNo].aft, fp_port[portNo].fallback_bd_id,
			fp_port[portNo].block_state, fp_port[portNo].trusted,
			fp_port[portNo].tcsel_table,
			fp_port[portNo].pcp2tc_table,
			fp_port[portNo].def_cfi, fp_port[portNo].def_pri,
			fp_port[portNo].def_tc, fp_port[portNo].tc2cos_table,
			fp_port[portNo].pindex,
			fp_port[portNo].csr_untag_from_btable);
	} else {
		ret = -1;
	}

	return ret;
}


/*******************************************************************
 * Function Name  : fp_global_br_entry_config
 * Description    : Configures GLobal bridge entry
 * Inputs         :
 * Parameters     : struct gloabl_s *
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_global_br_entry_config(struct global_s *pGlobal_br_entry)
{
	int32_t ret = 0;
	uint64_t gBrEntry = 0;

	/* UINT global_addr = GLOBAL_BR_ENTRY_ADDR; */
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "CFGGLOBAL:\r\n");

	if (NULL != pGlobal_br_entry) {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"before writing  l2_splPunt_enable = %x,"
			"fwdPortList = %x, untagList = %xucastHitAct = %x,"
			"mcastHitAct = %x,ucastMissAct = %xmcastMissAct = %x,"
			"mstpAct = %x\n\n",
			pGlobal_br_entry->l2_special_punt_enable,
			pGlobal_br_entry->fallback_bd_entry.forward_list,
			pGlobal_br_entry->fallback_bd_entry.untag_list,
			pGlobal_br_entry->fallback_bd_entry.ucast_hit_action,
			pGlobal_br_entry->fallback_bd_entry.mcast_hit_action,
			pGlobal_br_entry->fallback_bd_entry.ucast_miss_action,
			pGlobal_br_entry->fallback_bd_entry.mcast_miss_action,
			pGlobal_br_entry->fallback_bd_entry.mstp_action);

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"l2splpunt:%x, brentry:%llx\r\n",
			pGlobal_br_entry->l2_special_punt_enable,
			(uint64_t)(&pGlobal_br_entry->fallback_bd_entry) );

		fp_memcpy(&fp_global, pGlobal_br_entry, sizeof(struct global_s));

		/* fp_write_hw_data(wsp_base_addr, (UCHAR *)pGlobal_br_entry,
				global_addr, sizeof(struct global_s)); */

		gBrEntry = ((((uint64_t)fp_global.fallback_bd_entry.ucast_hit_action << BRENTRY_UCAST_HIT_ACT_START_POS) & BRENTRY_UCAST_HIT_ACT_MASK) |
					(((uint64_t)fp_global.fallback_bd_entry.ucast_miss_action << BRENTRY_UCAST_MISS_ACT_START_POS) & BRENTRY_UCAST_MISS_ACT_MASK) |
					(((uint64_t)fp_global.fallback_bd_entry.mcast_hit_action << BRENTRY_MCAST_HIT_ACT_START_POS) & BRENTRY_MCAST_HIT_ACT_MASK) |
					(((uint64_t)fp_global.fallback_bd_entry.mcast_miss_action << BRENTRY_MCAST_MISS_ACT_START_POS) & BRENTRY_MCAST_MISS_ACT_MASK)  |
					(((uint64_t)fp_global.fallback_bd_entry.untag_list << BRENTRY_UNTAG_LIST_START_POS) & BRENTRY_UNTAG_LIST_MASK) |
					(((uint64_t)fp_global.fallback_bd_entry.forward_list) & BRENTRY_FWD_PORT_LIST_MASK)); 

		gBrEntry = ((gBrEntry << 1) | (fp_global.l2_special_punt_enable & 0x1));
		/* RegWrite(wsp_base_addr, CLASS_HW_GLOBAL_CFG,
			((gBrEntry << 1) | GLOBAL_L2_SPL_PUNT_ACT)); */

		fp_reg_write( (uint32_t)CLASS_HW_GLOBAL_CFG, gBrEntry & 0xFFFFFFFF);
		fp_reg_write( (uint32_t)CLASS_HW_GLOBAL_CFG1, (gBrEntry >> 32));

		fp_reg_write( (uint32_t)CLASS2_HW_GLOBAL_CFG, gBrEntry & 0xFFFFFFFF);
		fp_reg_write( (uint32_t)CLASS2_HW_GLOBAL_CFG1, (gBrEntry >> 32));


	} else {
		ret = -1;
	}

	return ret;
}

/*******************************************************************
 * Function Name  : fp_br_entry_del
 * Description    : Deletes an bridge entry
 * Inputs         :
 * Parameters     : vlanId
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_del(uint32_t vlanId)
{
	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID)
    {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "DELENTRY\r\n");
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "vlanId: %x\r\n", vlanId);

	return wsp_vlan_entry_delete(vlanId);
}

/*******************************************************************
 * Function Name  : fp_br_entry_del_ports
 * Description    : Deletes ports from an existing bridge entry
 * Inputs         :
 * Parameters     : vlanId
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
int32_t fp_br_entry_del_ports(uint32_t vlanId, uint32_t portList)
{
	int64_t action_entry = 0;

	/*vlan id validation*/
	if (vlanId < SWITCHAPI_MIN_VLAN_ID || vlanId > SWITCHAPI_MAX_VLAN_ID)
    {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_VLANID %d\n", __func__, vlanId);
		return SWITCHAPI_ERR_MSG_VLANID;
	}

	/*port list validation*/
	if (portList < SWITCHAPI_MIN_PORT_LIST || portList > SWITCHAPI_MAX_PORT_LIST)
    {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s:SWITCHAPI_ERR_MSG_PORTLIST %d\n", __func__, portList);
		return SWITCHAPI_ERR_MSG_PORTLIST;
	}

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "DELPORTS:\r\n");
	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "vlanId: %x, PortList: %x\r\n", vlanId, portList);

	action_entry = wsp_vlan_entry_search(vlanId);
	if (action_entry == -1){
		return 0;
	}

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "act entry before:%llx\r\n", action_entry);

	action_entry = action_entry & (~((portList & FWD_PORT_LIST_MASK) |
					((portList & FWD_PORT_LIST_MASK) << BRENTRY_UNTAG_LIST_START_POS)) );

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "act entry after:%llx\r\n", action_entry);

	return wsp_vlan_entry_update(vlanId, action_entry);
}

/*******************************************************************
 * Function Name  : fp_static_mac_entry_add
 * Description    : Adds MAC entry into mac hash table
 * Inputs         :
 * Parameters     : phy_port, mac, vlanid
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/

int32_t fp_static_mac_entry_add(uint8_t *mac, uint32_t vlanid,
		uint32_t fwd_ports, uint32_t hit_action, uint32_t tc)
{
	int32_t ret = -1;
	uint32_t action_entry = 0;

	/* check for for already entry existence */
    ret = hw_hash_table_search(mac, vlanid);

    if (ret == -1){    /* entry not found */
		/* fill action_entry's fresh bit */
		action_entry = FRESH_FLAG_SET | MAC_2F_ENTRY_STATIC_MASK;

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
				"%s: Search Fail before add\n", __func__);
		/* Add SMAC entry with action entry & valid fields */

		/* update action entry with port info on which it arrived */
		action_entry |= ((fwd_ports << MAC_2F_ENTRY_FWD_PORT_LIST_START_POS) & FWD_PORT_LIST_MASK);
//		action_entry |= ( ( ( 1 << fwd_ports ) << MAC_2F_ENTRY_FWD_PORT_LIST_START_POS ) & FWD_PORT_LIST_MASK );
		/* update TC field */
		action_entry |= ( ( tc << MAC_2F_ENTRY_TC_START_POS ) & MAC_2F_ENTRY_TC_MASK );
		/* update hit action */
		action_entry |= ( hit_action << MAC_2F_ENTRY_FWD_ACT_START_POS ) & MAC_2F_ENTRY_FWD_ACT_MASK;

		/* add mac entry */
		ret = hw_hash_table_add(mac, vlanid, action_entry);
		if (ret == -1){
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "hash table add Fail\n");
		}else{
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "entry added successfully\n");
		}
	}else{  /* entry found, but not a static one then overwrite */
		/* mac hash entry found */
		if ( (  ret & MAC_2F_ENTRY_STATIC_MASK) &&
			 (( ret & MAC_2F_ENTRY_FWD_PORT_LIST_MASK) == fwd_ports) && 
			 (((ret & MAC_2F_ENTRY_TC_MASK) >> MAC_2F_ENTRY_TC_START_POS) == tc) && 
			 (((ret & MAC_2F_ENTRY_FWD_ACT_MASK) >> MAC_2F_ENTRY_FWD_ACT_START_POS) == hit_action) ){
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO,
			"%s:entry found already, not able to add\n", __func__);
			return ret;
		}

		/* non static entry, update with static data */

		/* fill action_entry's fresh bit */
		action_entry = FRESH_FLAG_SET | MAC_2F_ENTRY_STATIC_MASK;

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s: Search Pass! Entry needs to update\n", __func__);
		/* Add SMAC entry with action entry & valid fields */

		/* update action entry with port info on which it arrived */
		//action_entry |= ( ( ( 1 << fwd_ports ) & FWD_PORT_LIST_MASK ) << MAC_2F_ENTRY_FWD_PORT_LIST_START_POS );
		action_entry |= ( ( fwd_ports & FWD_PORT_LIST_MASK ) << MAC_2F_ENTRY_FWD_PORT_LIST_START_POS);
		/* update hit action */
		action_entry |= ( hit_action << MAC_2F_ENTRY_FWD_ACT_START_POS ) & MAC_2F_ENTRY_FWD_ACT_MASK;
		/* update TC field */
		action_entry |= ( ( tc << MAC_2F_ENTRY_TC_START_POS ) & MAC_2F_ENTRY_TC_MASK);

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s: action_entry %x\n", __func__, action_entry);
        /* add mac entry */
		ret = hw_hash_table_update(mac, vlanid, action_entry);
		if (ret == -1){
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "hash table update Fail\n");
		}else{
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "entry update successfully\n");
		}
	}
	return ret;
}

/*******************************************************************
 * Function Name  : fp_mac_entry_add
 * Description    : Adds MAC entry into mac hash table
 * Inputs         :
 * Parameters     : phy_port, mac, vlanid
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/

int32_t fp_mac_entry_add(uint32_t phy_port, uint8_t *mac, uint32_t vlanid)
{
	int64_t ret = -1;
	uint32_t fwd_action = 0;
	uint32_t action_entry = 0;

	/* check for for already entry existence */
	ret = hw_hash_table_search(mac, vlanid);
    if (ret == -1)                                                                                  /* entry not found */
    {
		/* fill action_entry's fresh bit */
		action_entry = FRESH_FLAG_SET;

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s: Search Fail before add\n", __func__);
		/* Add SMAC entry with action entry & valid fields */

		/* update action entry with port info on which it arrived */
		action_entry |= ((0x1 << phy_port) & FWD_PORT_LIST_MASK);

		/* Get bridge action entry */
		ret = wsp_vlan_entry_search(vlanid);
		if (ret == -1)
        {
			/* Un handeled bridge entry, take fwd action of fp_global br entry */
			fwd_action = fp_global.fallback_bd_entry.ucast_hit_action;
		}
        else
        {
			/* Get bridge action entry's unicast hit fwd action */
            fwd_action = (ret & BRENTRY_UCAST_HIT_ACT_MASK) >> BRENTRY_UCAST_HIT_ACT_START_POS;
		}

		/* Update mac action entry's fwd action */
		action_entry = (action_entry & ~MAC_2F_ENTRY_FWD_ACT_MASK) |
			(fwd_action << MAC_2F_ENTRY_FWD_ACT_START_POS);

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s: action_entry %x\n",
							__func__, action_entry);
		/* add mac entry */
		ret = hw_hash_table_add(mac, vlanid, action_entry);
		if (ret == -1)
        {
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "hash table add Fail\n");
		}
        else
        {
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "entry added successfully\n");
		}
	}
    else                                                                                            /* entry found */
    {
		/* mac hash entry found */
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s:entry found already, not able to add\n", __func__);
	}

	return ret;
}


/*******************************************************************
 * Function Name  : fp_mac_entry_delete
 * Description    : Delete MAC entry from mac hash table
 * Inputs         :
 * Parameters     : mac, vlanid
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/

int32_t fp_mac_entry_del(uint8_t *mac, uint32_t vlanid)
{
	int32_t ret = -1;

	/* check for for already entry existence */
	ret = hw_hash_table_search(mac, vlanid);
	if (ret == -1)                                                                                     /* entry not found */
    {
		/* fill action_entry's fresh bit */
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s: Entry not found to delete\n", __func__);
	}
    else                                                                                               /* entry found */
    {
		/* mac hash entry found */
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s: Entry found to delete\n", __func__);
		hw_hash_table_delete(mac, vlanid);
		/* check for entry after deletion */
		ret = hw_hash_table_search(mac, vlanid);
		if (ret == -1)
        {
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s: Entry deleted\n", __func__);
			fp_rc_stats.hw_hash_entry_count--;
		}
        else                                                                                        /* entry not deleted */
        {
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "%s: Entry not deleted\n", __func__);
		}
	}
	return 0;
}

/*******************************************************************
 * Function Name  : fp_mac_entry_search
 * Description    : search MAC entry from mac hash table
 * Inputs         :
 * Parameters     : mac, vlanid
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/

int32_t fp_mac_entry_search(uint8_t *mac, uint32_t vlanid)
{
	int32_t ret = -1;

	/* check for for already entry existence */
	ret = hw_hash_table_search( mac, vlanid );
	FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_DEBUG, "%s: ret = %x",  __func__, ret );
    FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_DEBUG, "%s: Entry %sfound\n", __func__, ( ret == -1 )?"not ":"" );

	return ret;
}

/*******************************************************************
 * Function Name  : fp_mac_entry_update
 * Description    : Update MAC entry into mac hash table
 * Inputs         :
 * Parameters     : mac_entry, mac, vlanid
 * Outputs        :
 * Parameters     : -
 * Returns        : int
 ********************************************************************/

int32_t fp_mac_entry_update(uint32_t mac_entry,	uint8_t *mac, uint32_t vlanid)
{
	int32_t ret = -1;

    /* check for for already entry existence */
	ret = hw_hash_table_search(mac, vlanid);
    if (-1 == ret)                                                                                  /* entry not found */
    {
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_EMERG,	"%s: Entry not found to update\n", __func__);
	}
    else                                                                                            /* entry found */
    {
		/* mac hash entry found */

		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s: Entry found to update\n", __func__);
		/* Add SMAC entry with action entry & valid fields */
		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "%s: mac_entry %x\n", __func__, mac_entry);
		/* update mac entry */
		ret = hw_hash_table_update(mac, vlanid, mac_entry);
		if (ret == -1)
        {
			FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_ERR, "hash table update Fail\n");
		}
	}
	return ret;
}
/*******************************************************************
 * Function Name  : fp_ltc_gettime
 * Description    : Getting ltc time difference between channel head
 *                : queue and current time
 * Inputs         :
 * Parameters     : device number, channel number
 * Outputs        : difftime
 * Parameters     : -
 * Returns        : int
 ********************************************************************/
uint32_t fp_ltc_gettime(uint32_t dev_number, uint32_t ch_num)
{
	uint32_t ch_time = 0, cur_time = 0;

	if (!((0 < ch_num) && (ch_num <= 5)))
		return SWITCHAPI_ERR_MSG_LTC_TIMEDIFF;

	if (!((0 < dev_number) && (dev_number <= 5)))
		return SWITCHAPI_ERR_MSG_LTC_TIMEDIFF;

	ch_num--;
	dev_number--;

	if (dev_number == HIF0) {
		ch_time = fp_reg_read((HIF0_LTC_CH0_TIME_SEC + ch_num * LTC_CHANNEL_NUM));
		cur_time = fp_reg_read(LTC_TSU_TIME_SEC);
		return ch_time - cur_time;

	} else if (dev_number == HIF1) {
		ch_time = fp_reg_read((HIF1_LTC_CH0_TIME_SEC + ch_num * LTC_CHANNEL_NUM));
		cur_time = fp_reg_read(LTC_TSU_TIME_SEC);
		return ch_time - cur_time;

	} else {
		return 0;
	}
		return 0;
}
/*******************************************************************
 *  * Function Name  : fp_ltc_hif_config_time
 *  * Description    : Config ltc time seconds and nano seconds
 *  * Inputs         :
 *  * Parameters     : device number, value, type
 *  * Outputs        :
 *  * Parameters     : -
 *  * Returns        : int
 ********************************************************************/
int32_t
fp_ltc_hif_config_time(uint32_t devId, uint32_t value, uint32_t type)
{
	if (type == LTC_HIF_SECONDS)
		fp_reg_write( (uint32_t)(LTC_HIF1_SECONDS_VAL + (LTC_REG_LEN * (devId-LTC_DEV_VAL))), value);
	else
		fp_reg_write( (uint32_t)(LTC_HIF1_NANOSECONDS_VAL + (LTC_REG_LEN * devId-LTC_DEV_VAL)), value);
	return 0;
}

