/*************************************************************************/ /*!
@File
@Title          hw_2field_hash_table.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    routines for mac hash add, del, update and search
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/*********************************************************************
*                           Include files
*********************************************************************/
 #include <fp_library.h>
 #include <fp_switch.h>
 #include <hw_2field_hash_table.h>
 #include <linux/spinlock.h>

#define DEFINE_SPINLOCK(x)    spinlock_t x = __SPIN_LOCK_UNLOCKED(x)
DEFINE_SPINLOCK(lockHashMac);

static const unsigned char act[8][20] = {"ACT_FORWARD", "ACT_FLOOD", "ACT_PUNT", "ACT_DISCARD", "ACT_OVERRIDE", "ACT_FWD_MASK", "ACT_COS_DISCARD", "UNKNOWN"};

/**********************************************************************
* Function Name  : hw_hash_table_search
* Description    : Search Mac Hash Table with given Mac Address
* Inputs
* Parameters   : - Mac Address and VLAN ID is Optional
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Values if Search Found
* Changes        :
************************************************************************/
int32_t hw_hash_table_search( uint8_t *mac, uint16_t vlaId ){
    int32_t entry = -1;
    uint32_t counter, status;

    spin_lock_bh( &lockHashMac );

    /* initialize MAC regs */
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG ), ( mac[ 0 ] << 24 ) | ( mac[ 1 ] << 16 ) | ( mac[ 2 ] << 8 ) | mac[ 3 ] );
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG ), ( ( vlaId << 16 ) & 0xFFFF0000 ) | ( mac[ 4 ] << 8 ) | mac[ 5 ] );
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG ), 0x00 );
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG ), 0x00 );

    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_CMD_REG ), CMD_SEARCH | BIT_SET_MAC_2F_MACTBL | BIT_SET_VLAN_2F_MACTBL );

    counter = 0;
    do{
        status = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_STATUS_REG );
    }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

    if( unlikely( !( status & STAT_CMD_DONE ) ) ){
        FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash search cmd\n", __func__ );
    }
	/* Clear Status Reg */
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_STATUS_REG ), STAT_REG_CLR );

	/*Bridge entry match done check*/
	if( status & STAT_ENTR_MATCH ){
		/* SEARCH FOUND */
		entry = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_ENTRY_REG );
		entry &= MAC_2F_ACTION_ENTRY_MASK;
		FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_INFO, "%s: entry = %x\n", __func__, entry );
        /* Clear req entry reg */
        fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_ENTRY_REG ), 0x0 );
	}
    spin_unlock_bh( &lockHashMac );
    return entry;
}

/**********************************************************************
* Function Name  : hw_hash_table_add
* Description    : Add given Mac Entry into MAC HASH TABLE
* Inputs
* Parameters   : - Mac Address and bridge entry pointer
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Add Status
* Changes        :
************************************************************************/
int32_t hw_hash_table_add(uint8_t *mac, uint16_t vlaId, uint32_t brentry)
{
    uint32_t  counter, status, classNum, classBase;
    int32_t result = 0;

    spin_lock_bh( &lockHashMac );

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ ){
        if( classNum == 1 ){
            classBase = CLASS_HW_BASE_ADDR;
        }else{
            classBase = CLASS2_HW_BASE_ADDR;
        }

        /* initialize MAC regs */
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG, ( ( mac[ 0 ] << 24 ) | ( mac[ 1 ] << 16 ) | ( mac[ 2 ] << 8 ) | mac[ 3 ] ) );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG, ( ( ( vlaId << 16 ) & 0xFFFF0000 ) | ( mac[ 4 ] << 8 ) | mac[5] ) );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG, 0x00 );
        /* fill ENTRY reg */
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_ENTRY_REG, brentry );

        /* Issue ADD cmd */
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_CMD_REG, ( CMD_ADD | BIT_SET_MAC_2F_MACTBL | BIT_SET_VLAN_2F_MACTBL ) );

        counter = 0;
        do{
            status = fp_reg_read( classBase + CLASS_DAMACHASH_HOST_STATUS_REG );
        }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

        if( unlikely( !( status & STAT_CMD_DONE ) ) ){
            FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash add cmd, class %d\n", __func__, classNum );
            result = -1;
        }

        /* Clear STATUS reg by writing all 1's */
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_STATUS_REG, STAT_REG_CLR );

        if( ( status & STAT_CMD_DONE ) && ( classNum == 1 ) ){
            fp_rc_stats.hw_hash_entry_count++;
        }
    }
    spin_unlock_bh( &lockHashMac );
	return result;
}

/**********************************************************************
* Function Name  : hw_hash_table_delete
* Description    : Delete given Mac Entry from MAC HASH TABLE
* Inputs
* Parameters   : - Mac Address
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Delete Status
* Changes        :
**********************************************************************/
void hw_hash_table_delete(uint8_t *mac, uint16_t vlanId)
{
    uint32_t  counter, status, classNum, classBase;

    spin_lock_bh( &lockHashMac );

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ ){
        if( classNum == 1 ){
            classBase = CLASS_HW_BASE_ADDR;
        }else{
            classBase = CLASS2_HW_BASE_ADDR;
        }

        FP_DEBUG( DBGP_FEAT_BRIDGE, FP_LOG_INFO, "HASH Del mac = %x:%x:%x:%x:%x:%x vlanID = %d. Class%d\r\n",
                 mac[ 0 ], mac[ 1 ], mac[ 2 ], mac[ 3 ], mac[ 4 ], mac[ 5 ], vlanId, classNum );

        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG, ( ( mac[ 0 ] << 24 ) | ( mac[ 1 ] << 16 ) | ( mac[ 2 ] << 8 ) | mac[ 3 ] ) );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG, ( ( ( vlanId << 16 ) & 0xFFFF0000 ) | ( mac[ 4 ] << 8 ) | mac[5] ) );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG, 0x0 );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG, 0x0 );

        /* Issue DEL cmd */
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_CMD_REG, ( CMD_DEL | BIT_SET_MAC_2F_MACTBL | BIT_SET_VLAN_2F_MACTBL ) );

        counter = 0;
        do{
            status = fp_reg_read( classBase + CLASS_DAMACHASH_HOST_STATUS_REG );
        }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

        if( unlikely( !( status & STAT_CMD_DONE ) ) ){
            FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash delete cmd\n", __func__ );
        }
        /* Clear STATUS reg by writing all 1's */
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_STATUS_REG, STAT_REG_CLR );
    }
    spin_unlock_bh( &lockHashMac );
}
/**********************************************************************
* Function Name  : hw_hash_table_update
* Description    : Update given BRENTRY PTR into HASH MAC TABLE
* Inputs
* Parameters   : - Mac Address and brentry
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Update Status
* Changes        :
**********************************************************************/
int32_t hw_hash_table_update(uint8_t *mac, uint16_t vlanId, uint32_t brentry)
{
    uint32_t  counter, status, classNum, classBase;
    int32_t result = 0;

    spin_lock_bh( &lockHashMac );

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ )
    {
        if( classNum == 1 ){
            classBase = CLASS_HW_BASE_ADDR;
        }else{
            classBase = CLASS2_HW_BASE_ADDR;
        }

        /* initialize MAC regs */
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG, mac[0] << 24 | mac[1] << 16 | mac[2] << 8 | mac[3]);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG, ((vlanId << 16) & 0xFFFF0000) | mac[4] << 8 | mac[5]);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_ENTRY_REG, brentry);

        /* Issue UPDATE cmd */
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_CMD_REG, CMD_UPDATE | BIT_SET_MAC_2F_MACTBL | BIT_SET_VLAN_2F_MACTBL);

        counter = 0;
        do{
            status = fp_reg_read(classBase + CLASS_DAMACHASH_HOST_STATUS_REG);
        }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

        if( unlikely( !( status & STAT_CMD_DONE ) ) ){
            FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash update cmd, class %d\n", __func__, classNum );
            result = -1;
        }

        /* Clear STATUS reg by writing all 1's */
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_STATUS_REG, STAT_REG_CLR);
    }

    spin_unlock_bh( &lockHashMac );

    return 0;
}


/**********************************************************************
* Function Name  : hw_hash_table_init
* Description    : Initialize MAC HASH TABLE & VLAN HASH TABLE
*                  Initialize the hash collision space
*                  chain the collision space
*                  init the free head, list pointers and no of entries
*                  All initialize done through register programming only
************************************************************************/
void hw_hash_table_init( void )
{
    uint32_t counter, status, i, classNum, classBase, collisionPtr;
    int32_t result = 0;

    /* 2-field Mac table format: Table Depth - 8192(hash 4096 + collison 4096)
    *=====================================================================================*
    * valid bits |  col_ptr  |   portNO  | field_valids | action entry |  vlanid |	MAC   *
    *=====================================================================================*
    *    4-bit   |  16-bit   |   4-bit   | 	  8-bit     |    31-bit    |  13-bit | 48-bit *
    *  123 - 120 | 119 - 104 | 103 - 100 | 	 99 - 92	| 	91 - 61	   | 60 - 48 | 47 - 0 *
    *=====================================================================================*/
    /* CLASS_DACAMHASH_HOST_MAC_ADDR1_REG has MAC entry bits:	31 - 0	*/
    /* CLASS_DACAMHASH_HOST_MAC_ADDR2_REG has MAC entry bits:	63 - 32	*/
    /* CLASS_DACAMHASH_HOST_MAC_ADDR3_REG has MAC entry bits:	95 - 64	*/
    /* CLASS_DACAMHASH_HOST_MAC_ADDR4_REG has MAC entry bits:	127 - 96*/

    spin_lock_bh( &lockHashMac );

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ )
    {
        if( classNum == 1 ){
            classBase = CLASS_HW_BASE_ADDR;
        }else{
            classBase = CLASS2_HW_BASE_ADDR;
        }

        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_STATUS_REG, STAT_REG_CLR );
        fp_reg_write( classBase + CLASS_DAMACHASH_HOST_CMD_REG, CMD_INIT );

        counter = 0;
        do
        {
            status = fp_reg_read( classBase + CLASS_DAMACHASH_HOST_STATUS_REG );
        }while( ( status != ( STAT_INIT_DONE | STAT_CMD_DONE ) )  && ( counter++ < 100 ) );

        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_STATUS_REG, STAT_REG_CLR);

        if( unlikely( status != ( STAT_INIT_DONE | STAT_CMD_DONE ) ) ){
            FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash init cmd, class %d\n", __func__, classNum );
            result = -1;
        }

        //if( ( counter == 100 ) & ( status != ( STAT_INIT_DONE | STAT_CMD_DONE ) ) )
        //    FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "MAC HASH TABLE INIT FAILED! HW CLASS%d\n", classNum);

        /* initialize mac addresses */
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR5_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_ENTRY_REG, 0x0);

        for( i = 0; i < TWO_FIELD_HASH_ENTRIES; i++ )
        {
            if( i == ( TWO_FIELD_HASH_ENTRIES - 1 ) ){
                collisionPtr = 0x40000 << 8;
            }else{
                collisionPtr = ( 0x41001 + i ) << 8;
            }

            fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG, collisionPtr);
            fp_reg_write(classBase + CLASS_DAMACHASH_HOST_CMD_REG, ( ( ( TWO_FIELD_HASH_ENTRIES + i ) << 16 ) | CMD_MEM_WRITE ));

            counter = 0;
            do
            {
                status = fp_reg_read( classBase + CLASS_DAMACHASH_HOST_STATUS_REG );
            }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

            //if( ( counter == 100 ) & !( status & STAT_CMD_DONE ) )
            //    FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "MAC HASH TABLE INIT COLLISION SPACE FAILED!! HW CLASS%d\n", classNum);
            if( unlikely( !( status & STAT_CMD_DONE ) ) ){
                FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash write cmd at collision space init, class %d\n", __func__, classNum );
                result = -1;
            }
        }

        fp_reg_write(classBase + FREE_LIST_HEAD_PTR, TWO_FIELD_INIT_HEAD_PTR);
        fp_reg_write(classBase + FREE_LIST_TAIL_PTR, TWO_FIELD_INIT_TAIL_PTR);
        fp_reg_write(classBase + FREE_LIST_ENTRIES, TWO_FIELD_COLL_ENTRIES);

        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_STATUS_REG, STAT_REG_CLR);
        fp_reg_write(classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG, 0x00);
    }
    spin_unlock_bh( &lockHashMac );

    if( !result ){
	   FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_INFO, "%s: MAC HASH TABLES INIT DONE\n", __func__ );
    }
}

/* Read one entry data */
void hw_hash_table_read_entry( uint32_t *entry_data, uint16_t entry )
{
    uint32_t counter, status;

    spin_lock_bh( &lockHashMac );

    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_CMD_REG ), (entry << 16) | CMD_MEM_READ );

    counter = 0;
    do
    {
        status = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_STATUS_REG );
    }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

    if( unlikely( !( status & STAT_CMD_DONE ) ) ){
        FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash read cmd\n", __func__ );
    }
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_STATUS_REG ), STAT_REG_CLR );

    if( status & STAT_CMD_DONE )
    {
        entry_data[REG_MAC1] = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG );
        entry_data[REG_MAC2] = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG );
        entry_data[REG_MAC3] = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG );
        entry_data[REG_MAC4] = fp_reg_read( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG );
    }
    else
    {
        entry_data[REG_MAC1] = 0;
        entry_data[REG_MAC2] = 0;
        entry_data[REG_MAC3] = 0;
        entry_data[REG_MAC4] = 0;
    }
    spin_unlock_bh( &lockHashMac );
}

/* Write one entry data */
void hw_hash_table_write_entry( uint32_t *entry_data, uint16_t entry )
{
    uint32_t classNum, counter, status, classBase;

    spin_lock_bh( &lockHashMac );

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ )
    {
        if( classNum == 1 )
            classBase = CLASS_HW_BASE_ADDR;
        else
            classBase = CLASS2_HW_BASE_ADDR;

        fp_reg_write( ( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR1_REG ), entry_data[REG_MAC1] );
        fp_reg_write( ( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR2_REG ), entry_data[REG_MAC2] );
        fp_reg_write( ( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR3_REG ), entry_data[REG_MAC3] );
        fp_reg_write( ( classBase + CLASS_DAMACHASH_HOST_MAC_ADDR4_REG ), entry_data[REG_MAC4] );
        fp_reg_write( ( classBase + CLASS_DAMACHASH_HOST_CMD_REG ), ( entry << 16 ) | CMD_MEM_WRITE );

        counter = 0;
        do
        {
            status = fp_reg_read( classBase + CLASS_DAMACHASH_HOST_STATUS_REG );
        }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

        if( unlikely( !( status & STAT_CMD_DONE ) ) ){
            FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete mac hash write cmd, class %d\n", __func__, classNum );
        }
        fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAMACHASH_HOST_STATUS_REG ), STAT_REG_CLR );
    }
    spin_unlock_bh( &lockHashMac );
}

/* Read function for procFS */
int hw_hash_table_read( uint8_t *buff )
{
    static uint16_t pos = 0, record = 0;
    uint32_t data[4], tmp, dataLen, counter;

    FP_DEBUG(DBGP_FEAT_PROCFS, FP_LOG_DEBUG, "hw_hash_table_read start read\n");

    data[0] = 0;
    data[1] = 0;
    data[2] = 0;
    data[3] = 0;
    dataLen = 0;

    while( pos < 8192 )
    {
        if( !pos ){
            FP_DEBUG(DBGP_FEAT_PROCFS, FP_LOG_DEBUG, "hw_hash_table_read read hash mac table start\n");
        }

        hw_hash_table_read_entry( data, pos );

        pos++;

        if( data[REG_MAC4] & HASH_ENTRY_VALID_FLAG )
        {
            record++;
            dataLen += sprintf( buff, "Record %d:\n", record );
            dataLen += sprintf( &buff[dataLen], "VLAN ID:   %d\nMAC:       ", ( data[REG_MAC2] >> 16 ) & 0xFFF );
            dataLen += sprintf( &buff[dataLen], "%02x:", ( data[REG_MAC1] >> 24 ) & 0xFF );
            dataLen += sprintf( &buff[dataLen], "%02x:", ( data[REG_MAC1] >> 16 ) & 0xFF );
            dataLen += sprintf( &buff[dataLen], "%02x:", ( data[REG_MAC1] >> 8 ) & 0xFF );
            dataLen += sprintf( &buff[dataLen], "%02x:", data[REG_MAC1] & 0xFF );
            dataLen += sprintf( &buff[dataLen], "%02x:", ( data[REG_MAC2] >> 8 ) & 0xFF );
            dataLen += sprintf( &buff[dataLen], "%02x\n", data[REG_MAC2] & 0xFF );
            if( data[REG_MAC3] & HASH_ENTRY_FRESH_FLAG ){
                dataLen += sprintf( &buff[dataLen], "Fresh bit: Set\n" );
            }
            else{
                dataLen += sprintf( &buff[dataLen], "Fresh bit: Reset\n" );
            }

            if( data[REG_MAC3] & HASH_ENTRY_STATIC_FLAG ){
                dataLen += sprintf( &buff[dataLen], "Type:      Static\n" );
            }
            else{
                dataLen += sprintf( &buff[dataLen], "Type:      Dynamic\n" );
            }

            dataLen += sprintf( &buff[dataLen], "Action:    %s\n", act[(data[REG_MAC3] >> 20) & 0x07] );

            tmp = ((data[REG_MAC3] << 3 ) | (data[REG_MAC2] >> 29)) & 0x1FFFF;

            counter = 1;
            while( tmp != 1 )
            {
                tmp = tmp >> 1;
                counter++;
            }
            dataLen += sprintf( &buff[dataLen], "Port:      %d\n\n", counter );
            return dataLen;
        }
    }

    pos = 0;
    record = 0;
    FP_DEBUG(DBGP_FEAT_PROCFS, FP_LOG_DEBUG, "hw_hash_table_read read hash mac table end\n");
    return 0;
}
/*********************** End of File **********************/
