/*************************************************************************/ /*!
@File
@Title          hw_vlan_hash_table.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    routines for vlan hash add, del, update and search
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/*********************************************************************
*                           Include files
*********************************************************************/
#include <linux/delay.h>
#include <linux/spinlock.h>
#include "wsp_vlan_hash_table_L.h"

#define CLASS_BUS_ACCESS_ADDR               (CLASS_HW_BASE_ADDR + 0x228)
#define CLASS_BUS_ACCESS_WDATA              (CLASS_HW_BASE_ADDR + 0x22c)
#define CLASS_BUS_ACCESS_RDATA              (CLASS_HW_BASE_ADDR + 0x230)

#define  INDIRECT_VLAN_CMD_REG        	     0x2050000
#define  INDIRECT_VLAN_MAC1_ADDR_REG   	    (INDIRECT_VLAN_CMD_REG + 0x04)
#define  INDIRECT_VLAN_MAC2_ADDR_REG        (INDIRECT_VLAN_CMD_REG + 0x08)
#define  INDIRECT_VLAN_MAC3_ADDR_REG        (INDIRECT_VLAN_CMD_REG + 0x0c)
#define  INDIRECT_VLAN_MAC4_ADDR_REG        (INDIRECT_VLAN_CMD_REG + 0x10)
#define  INDIRECT_VLAN_MAC5_ADDR_REG        (INDIRECT_VLAN_CMD_REG + 0x14)
#define  INDIRECT_VLAN_ENTRY_REG            (INDIRECT_VLAN_CMD_REG + 0x18)
#define  INDIRECT_VLAN_STATUS_REG           (INDIRECT_VLAN_CMD_REG + 0x1c)
#define  INDIRECT_VLAN_DIRECT_REG           (INDIRECT_VLAN_CMD_REG + 0x20)

#define DEFINE_SPINLOCK(x)    spinlock_t x = __SPIN_LOCK_UNLOCKED(x)
DEFINE_SPINLOCK(lockHashVlan);

static const uint8_t act[8][20] = {"ACT_FORWARD", "ACT_FLOOD", "ACT_PUNT", "ACT_DISCARD", "ACT_OVERRIDE", "ACT_FWD_MASK", "ACT_COS_DISCARD", "UNKNOWN"};

uint32_t CSR_REG_READ_INDIRECT(uint32_t offset)
{
    fp_reg_write(CLASS_BUS_ACCESS_ADDR, offset);
    return fp_reg_read(CLASS_BUS_ACCESS_RDATA);
}

void CSR_REG_WRITE_INDIRECT(uint32_t offset, uint32_t value)
{
    fp_reg_write(CLASS_BUS_ACCESS_WDATA, value);
    fp_reg_write(CLASS_BUS_ACCESS_ADDR, offset | 0x80000000);
}

/**********************************************************************
* Function Name  : wsp_vlan_entry_search
* Description    : Search vlan grp Table with given vlanid
* Inputs
* Parameters   : - vlanid
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Values if Search Found
* Changes        :
************************************************************************/
int64_t wsp_vlan_entry_search( uint16_t vlanid )
{
	uint32_t /*match_done = 0,*/ cmd_done = 0, field_valids = BIT_SET_VLAN_1F_VLANTBL, counter = 0;
	uint32_t entry1 = 0, entry2 = 0;
	int64_t entry = -1;

    spin_lock_bh( &lockHashVlan );

    CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_MAC1_ADDR_REG, vlanid & VLANID_MASK );
    CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_MAC2_ADDR_REG, 0 );
    CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_MAC3_ADDR_REG, 0 );
    CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_MAC4_ADDR_REG, 0 );
    CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_MAC5_ADDR_REG, 0 );

	/* write the search command */
	CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_CMD_REG, ( CMD_SEARCH | field_valids ) );
	FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_INFO, "CLASS_1 CMDREG %x\n", ( CMD_SEARCH | field_valids ) );

	cmd_done = CSR_REG_READ_INDIRECT( ( uint32_t )INDIRECT_VLAN_STATUS_REG );
	/*While loop for command done chek*/
	while ( !( cmd_done & STAT_CMD_DONE ) && ( counter++ < 100 ) ) {
		cmd_done = CSR_REG_READ_INDIRECT( ( uint32_t )INDIRECT_VLAN_STATUS_REG );
	}

    if( unlikely( !( cmd_done & STAT_CMD_DONE ) ) ){
        FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete vlan hash search cmd\n", __func__ );
    }
	/* Match done status is also from STATUS REG */
	//match_done = cmd_done;

	/* Clear Status Reg */
	CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_STATUS_REG, STAT_REG_CLR );

	/*Bridge entry match done check*/
	if( cmd_done & STAT_ENTR_MATCH ){
		/* SEARCH FOUND */
		entry1 = CSR_REG_READ_INDIRECT( ( uint32_t )INDIRECT_VLAN_ENTRY_REG );
		entry2 = CSR_REG_READ_INDIRECT( ( uint32_t )INDIRECT_VLAN_DIRECT_REG );

		entry = ( ( ( uint64_t )entry2 << 32 ) | entry1 );
		/* Clear req entry reg */
		CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_ENTRY_REG, 0x0 );
		CSR_REG_WRITE_INDIRECT( ( uint32_t )INDIRECT_VLAN_DIRECT_REG, 0x0 );
	}

    spin_unlock_bh( &lockHashVlan );
	return entry;
}


/**********************************************************************
* Function Name  : wsp_vlan_entry_add
* Description    : Add given Mac Entry into MAC HASH TABLE
* Inputs
* Parameters   : - Mac Address port no and action entry pointer
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Add Status
* Changes        :
************************************************************************/
int32_t wsp_vlan_entry_add( uint16_t vlanid, uint64_t action_entry ){
    uint32_t status, classNum, classBase, entry1, entry2;
    int32_t retval = 0;

	entry1 = ( action_entry & 0xFFFFFFFF );
	entry2 = ( action_entry >> 32 );

	FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_INFO, "ADD VID:%x\n", vlanid );

	/* vlan group table format *
	*======================================================================*
	* valid bits | col_ptr | portNo | field_valids | action entry | vlan   *
	*======================================================================*
	* 4-bit      | 16-bit  | 4-bit  | 8-bit        | 55-bit       | 12-bit *
	*=====================================================================*/

	/* VLAN (BD) Hash table action entry map:*
	*=======================================*
	* [19:0]        - forward port list
	* [39:20]       - untag list
    * [42:40]		- ucast_hit_action
	* [45:43]       - mcast_hit_action
	* [48:46]       - ucast_miss_action
	* [51:49]       - mcast_miss_action
	* [54-52]       - Not used yet
	*=======================================*/
    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ ){
        if( classNum == 1 ){
            classBase = CLASS_HW_BASE_ADDR;
        }else{
            classBase = CLASS2_HW_BASE_ADDR;
        }

        spin_lock_bh( &lockHashVlan );

        /* Write entries corresponding MAC REG */
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG, vlanid & VLANID_MASK );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR5_REG, 0x00 );

        /* fill ENTRY reg */
        FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_INFO, "VLAN ENTRY %llx HW CLASS%d\n", action_entry & BRENTRY_ACTION_ENTRY_MASK, classNum );

        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_ENTRY_REG, entry1 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_DIRECT, entry2 );

        /* Description of ADD CMD REG:
        * 3:0   - CMD
        * 7:4   - Reserved
        * 8     - Vlan Field Valid
        * 15:9 - Reserved
        * 31:16 - Mem Index / portno[16:19]
        */

        /* Issue ADD cmd */
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_CMD_REG, ( CMD_ADD | BIT_SET_VLAN_1F_VLANTBL ) );
        FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_INFO, "VLAN TABLE ADD %x HW CLASS%d\r\n", ( CMD_ADD | BIT_SET_VLAN_1F_VLANTBL ), classNum );

        /* Description of STAUTS CMD REG:
        * 0 - req1_cmd_done
        * 1 - req1_sig_entry_not_found
        * 2 - req1_sig_init_done
        * 3 - req1_sig_entry_added
        * 4 - req1_match
        * 31:5 - reserved.
        */

        /* read STATUS reg - cmd done */
        status = fp_reg_read( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG );

        while( !( status & STAT_CMD_DONE ) ) {
            status = fp_reg_read( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG );
        }

        /* Clear STATUS reg by writing all 1's */
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG, STAT_REG_CLR );

        spin_unlock_bh( &lockHashVlan );

        if( !( status & ( STAT_CMD_DONE | STAT_ENTR_ADD ) ) ){
            FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_EMERG,	"VLAN ADD CMD TO TABLE FAILED. HW CLASS%d\n", classNum );
            retval--;
        }
    }

	return retval;
}

/**********************************************************************
* Function Name  : wsp_vlan_entry_delete
* Description    : Delete given vlan Entry from VLAN HASH TABLE
* Inputs
* Parameters   : - vlan id
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Delete Status
* Changes        :
**********************************************************************/
int32_t wsp_vlan_entry_delete(uint16_t vlanid)
{
    uint32_t status, classNum, classBase;
	int32_t retval = 0;

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ )
    {
        if( classNum == 1 )
            classBase = CLASS_HW_BASE_ADDR;
        else
            classBase = CLASS2_HW_BASE_ADDR;

        spin_lock_bh( &lockHashVlan );

        /* Write entries corresponding MAC REG */
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG, vlanid & VLANID_MASK);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG, 0x00);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG, 0x00);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG, 0x00);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR5_REG, 0x00);

        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_CMD_REG, CMD_DEL | BIT_SET_VLAN_1F_VLANTBL);

        /* read STATUS reg - cmd done */
        status = fp_reg_read(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG);

        while (!(status & STAT_CMD_DONE)) {
            status = fp_reg_read(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG);
        }

        /* Clear STATUS reg by writing all 1's */
    	fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG, STAT_REG_CLR);

    	if (!(status & (STAT_CMD_DONE | STAT_ENTR_ADD)))
        {
    		FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_EMERG,	"VLAN DEL CMD FROM TABLE FAILED. HW CLASS%d\n", classNum);
    		retval--;
    	}

        spin_unlock_bh( &lockHashVlan );
    }

	return retval;
}

/**********************************************************************
* Function Name  : wsp_vlan_entry_update
* Description    : Update given BRENTRY PTR into HASH MAC TABLE
* Inputs
* Parameters   : - Mac Address and brentry
* Globals      : -
* Outputs        :
*  Parameters  : -
* Returns      :   Returns Entry Update Status
* Changes        :
**********************************************************************/
int32_t wsp_vlan_entry_update(uint16_t vlanid, uint64_t brentry){
    uint32_t status, classNum, classBase, entry1, entry2;
	int32_t retval = 0;

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ ){
        if( classNum == 1 ){
            classBase = CLASS_HW_BASE_ADDR;
        }else{
            classBase = CLASS2_HW_BASE_ADDR;
        }

        spin_lock_bh( &lockHashVlan );

        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG, vlanid & VLANID_MASK );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG, 0x00 );
        fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR5_REG, 0x00 );

        entry1 = ( brentry & 0xFFFFFFFF );
        entry2 = ( brentry >> 32 );

        /* fill ENTRY reg if valid */
    	fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_ENTRY_REG, entry1 );
    	fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_DIRECT, entry2 );

    	/* Issue UPDATE cmd */
    	fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_CMD_REG, ( CMD_UPDATE | BIT_SET_VLAN_1F_VLANTBL ) );

    	/* read STATUS reg - cmd done */
    	status = fp_reg_read( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG );

    	while( !( status & STAT_CMD_DONE ) ) {
    		status = fp_reg_read( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG );
    	}

    	/* Clear STATUS reg by writing all 1's */
    	fp_reg_write( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG, STAT_REG_CLR );

    	if( !( status & STAT_ENTR_ADD ) ){
    		FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_EMERG, "VLAN UPDATE CMD TO TABLE FAILED. HW CLASS%d\n", classNum );
    		/* Entry Not Added */
    		retval--;
    	}
        spin_unlock_bh( &lockHashVlan );
    }
	return retval;
}


/**********************************************************************
* Function Name  : hw_vlan_hash_table_init
* Description    : Initialize VLAN HASH TABLE
*                  Initialize the hash collision space
*                  chain the collision space
*                  init the free head, list pointers and no of entries
*                  All initialize done through register programming only
************************************************************************/
void wsp_vlan_hash_table_init( void )
{
    uint32_t counter, status, i, classNum, collisionPtr, classBase;

	/* vlan group table format *
	*======================================================================*
	* valid bits | col_ptr | portNO  | field_valids | action entry | vlan  *
	*======================================================================*
	* 4-bit      | 16-bit  | 4-bit   | 8-bit        | 55-bit       | 13-bit*
	*   99 - 96  | 95 - 80 | 79 - 76 |	 75 - 68	|	67 - 13	   | 12 - 0*
	*======================================================================*/
	/*	CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG	has VLAN entry bits:	31 - 0	*/
	/*	CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG	has VLAN entry bits:	63 - 32	*/
	/*	CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG	has VLAN entry bits:	95 - 64	*/
	/*	CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG	has VLAN entry bits:	99 - 96	*/

    for( classNum = 1; classNum <= FP_CLASS_MAX_NUMBER; classNum++ )
    {
        if( classNum == 1 )
            classBase = CLASS_HW_BASE_ADDR;
        else
            classBase = CLASS2_HW_BASE_ADDR;

        spin_lock_bh( &lockHashVlan );

        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG, STAT_REG_CLR);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_CMD_REG, CMD_INIT);

        counter = 0;
        do
        {
            status = fp_reg_read(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG);
        }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG, STAT_REG_CLR);                        // Очистка статусного регистра

        if( ( !( status & STAT_CMD_DONE ) ) && ( counter == 100 ) )
            FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "VLAN HASH TABLE INIT FAILED! HW CLASS%d\n", classNum);

        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR5_REG, 0x0);

        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_ENTRY_REG, 0x0);
        fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_DIRECT, 0x0);

        for( i = 0; i < 64; i++ )
        {
            if( i == 63 )
                collisionPtr = 0x40000;
            else
                collisionPtr = 0x40041 + i;

            fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG, ( ( collisionPtr & 0xFFFF ) << 16 ));
            fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG, ( ( collisionPtr >> 16 ) & 0xF ));
            fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_CMD_REG, CMD_MEM_WRITE | ( ( VLAN_HASH_ENTRIES + i ) << 16 ));

            counter = 0;
            do
            {
                status = fp_reg_read(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG);
            }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

            fp_reg_write(classBase + CLASS_DAVLANHASH_HOST_STATUS_REG, STAT_REG_CLR);                    // Очистка статусного регистра

            if( ( !( status & STAT_CMD_DONE ) ) && ( counter == 100 ) )
                FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "VLAN HASH TABLE COLLISION SPACE INIT FAILED! HW CLASS%d\n", classNum);
        }

        fp_reg_write(classBase + FREE_VLAN_LIST_HEAD_PTR, VLAN_INIT_HEAD_PTR);
        fp_reg_write(classBase + FREE_VLAN_LIST_TAIL_PTR, VLAN_INIT_TAIL_PTR);
        fp_reg_write(classBase + FREE_VLAN_LIST_ENTRIES, VLAN_FREE_LIST_ENTRIES);

        spin_unlock_bh( &lockHashVlan );
    }

	FP_DEBUG(DBGP_FEAT_L2API, FP_LOG_INFO, "VLAN HASH TABLES INIT DONE\n");
}

/* Read one entry from VLAN table */
void wsp_read_entry_vlan_hash_table( uint32_t *entry_data, uint16_t entry_number ){
    uint32_t counter = 0, status;

    entry_number &= 0x7F;
    spin_lock_bh( &lockHashVlan );
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_CMD_REG ), ( ( entry_number << 16 ) | CMD_MEM_READ ) );
    do{
        status = fp_reg_read( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_STATUS_REG ) );
    }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

    if( unlikely( !( status & STAT_CMD_DONE ) && ( counter == 100 ) ) ){
        FP_DEBUG( DBGP_FEAT_L2API, FP_LOG_CRIT, "%s: ERROR: can not complete vlan hash read cmd\n", __func__ );
    }
    fp_reg_write( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_STATUS_REG ), STAT_REG_CLR );

    entry_data[ 0 ] = fp_reg_read( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG ) );
    entry_data[ 1 ] = fp_reg_read( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG ) );
    entry_data[ 2 ] = fp_reg_read( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG ) );
    entry_data[ 3 ] = fp_reg_read( ( CLASS_HW_BASE_ADDR + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG ) );

    spin_unlock_bh( &lockHashVlan );
}

/* Read function for procFS */
uint32_t wsp_vlan_hash_table_read(uint8_t *buff)
{
    static uint16_t pos;

    uint32_t counter, status, classBase = CLASS_HW_BASE_ADDR;
    uint32_t data[4], tmp, dataLen;

    while( pos < 128 ){
        spin_lock_bh( &lockHashVlan );

        fp_reg_write( ( classBase + CLASS_DAVLANHASH_HOST_CMD_REG ), ( ( pos << 16 ) | CMD_MEM_READ ) );
        pos++;

        counter = 0;
        do{
            status = fp_reg_read( ( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG ) );
        }while( !( status & STAT_CMD_DONE ) && ( counter++ < 100 ) );

        fp_reg_write( ( classBase + CLASS_DAVLANHASH_HOST_STATUS_REG ), STAT_REG_CLR );

        data[ REG_MAC1 ] = fp_reg_read( ( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG ) );
        if( data[ REG_MAC1 ] )
        {
            data[ REG_MAC1 ] = fp_reg_read( ( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR1_REG ) );
            data[ REG_MAC2 ] = fp_reg_read( ( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR2_REG ) );
            data[ REG_MAC3 ] = fp_reg_read( ( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR3_REG ) );
            data[ REG_MAC4 ] = fp_reg_read( ( classBase + CLASS_DAVLANHASH_HOST_MAC_ADDR4_REG ) );

            spin_unlock_bh( &lockHashVlan );

            tmp = data[ REG_MAC1 ] & 0xFFF;
            dataLen = sprintf( buff, "VLAN ID:   %d\n", tmp );
            tmp = ( ( data[ REG_MAC1 ] >> 13 )|( ( data[ REG_MAC2 ] & 1 ) << 20 ) ) & 0xFFFFF;
            dataLen += sprintf( &buff[ dataLen ], "Forward list:  0x%05X\n", tmp );
            tmp = ( data[ REG_MAC2 ] >> 1 ) & 0xFFFFF;
            dataLen += sprintf( &buff[ dataLen ], "Untag list:    0x%05X\n", tmp );
            tmp = ( data[ REG_MAC2 ] >> 21 ) & 7;
            dataLen += sprintf( &buff[ dataLen ], "UnicastHit:    %s\n", act[ tmp ] );
            tmp = ( data[ REG_MAC2 ] >> 27 ) & 7;
            dataLen += sprintf( &buff[ dataLen ], "UnicastMiss:   %s\n", act[ tmp ] );
            tmp = ( data[ REG_MAC2 ] >> 24 ) & 7;
            dataLen += sprintf( &buff[ dataLen ], "MulticastHit:  %s\n", act[ tmp ] );
            tmp = ( ( data[ REG_MAC2 ] >> 30 ) & 7 )|( ( data[ REG_MAC3 ] & 1 ) << 2 );
            dataLen += sprintf( &buff[ dataLen ], "MulticastMiss: %s\n", act[ tmp ] );
            tmp = ( data[ REG_MAC3 ] >> 1 ) & 7;
            dataLen += sprintf( &buff[ dataLen ], "Mstp:          %s\n", act[ tmp ] );
            tmp = ( data[ REG_MAC3 ] >> 4 ) & 0xFF;
            dataLen += sprintf( &buff[ dataLen ], "True field:    0x%02X\n", tmp );
            tmp = ( data[ REG_MAC3 ] >> 12 ) & 0xF;
            dataLen += sprintf( &buff[ dataLen ], "Port field:    0x%X\n", tmp );
            tmp = ( data[ REG_MAC3 ] >> 16 ) & 0xFFFF;
            dataLen += sprintf( &buff[ dataLen ], "Collision ptr: 0x%04X\n", tmp );
            tmp = data[ REG_MAC4 ] & 0xF;
            dataLen += sprintf( &buff[ dataLen ], "Flags:         0x%X\n\n", tmp );

            return dataLen;
        }
        else{
            spin_unlock_bh( &lockHashVlan );
        }
    }
    pos = 0;

    return 0;
}
