/*************************************************************************/ /*!
@File
@Title          fp_library.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    library functions for fast path module
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                     Include Files
 **********************************************************************/
#include <fp_library.h>

#ifdef __FP_OS__
#include <linux/module.h>
#include <linux/slab.h>
#endif

/**********************************************************************
 * Function Name  : fp_free
 * Description    : copy the data byte by byte with given len
 * Inputs
 *   Parameters   : char *, char *, int
 * Outputs        :
 *   Parameters   : char * - copied data into dst pointer
 *   Returns      : -
 * Changes        :
 ********************************************************************/
void
fp_bcopy(register char *src, register char *dst, int len)
{
	if (src == FP_NULL || dst == FP_NULL)
		return;

	if (dst < src) {
		while (len--)
			*dst++ = *src++;
	} else {
		char *lasts = src + (len - 1);
		char *lastd = dst + (len - 1);

		while (len--)
			*(char *)lastd-- = *(char *)lasts--;
	}
}

/**********************************************************************
 * Function Name  : fp_memcpy
 * Description    : helper routine for fp_bcopy
 * Inputs
 *   Parameters   : char *, char *, int
 * Outputs        :
 *   Parameters   : char * - copied data into dst pointer
 *   Returns      : void *
 * Changes        :
 ********************************************************************/
void *
fp_memcpy(void *dst , const void *src, unsigned int len)
{
	fp_bcopy((char *)src, (char *)dst, len);

	return dst;
}

/**********************************************************************
 * Function Name  : fp_memset
 * Description    : fills the given len of bytes contained by val in dest
 * Inputs
 *   Parameters   : void *, int, int
 * Outputs        :
 *   Parameters   : void *
 *   Returns      : returns filled data pointer
 * Changes        :
 ********************************************************************/
void *
fp_memset(void *dest, int val, unsigned int len)
{
	register unsigned char *ptr = (unsigned char *)dest;

	while (len-- > 0)
		*ptr++ = val;
	return dest;
}

/**********************************************************************
 * Function Name  : fp_SwapBytes
 * Description    : swap the first n no of bytes pointed by pv
 * Inputs
 *   Parameters   : void *, unsigned int
 * Outputs        :
 *   Parameters   : void *
 *   Returns      : -
 * Changes        :
 ********************************************************************/
void fp_SwapBytes(void *pv, unsigned int n)
{
	char *p = pv;
	unsigned int  lo, hi;
	for (lo = 0, hi = n-1; hi > lo; lo++, hi--) {
		char tmp = p[lo];
		p[lo] = p[hi];
		p[hi] = tmp;
	}
}

#define fp_SWAP(x) fp_SwapBytes(&x, sizeof(x));

/**********************************************************************
 * Function Name  : fp_convert_le2be
 * Description    : convert little-endian to big-endian
 * Inputs
 *   Parameters   : char *, unsigned int
 * Outputs        :
 *   Parameters   : char *
 *   Returns      : -
 * Changes        :
 ********************************************************************/
void
fp_convert_le2be(int8_t *data, uint32_t size)
{
	int32_t i, cnt;
	uint32_t *data_ptr = (uint32_t *)data;

	cnt  = size >> 2;
	if (size & 0x3)
		cnt++;
	for (i = 0; i < cnt; i++)
		fp_SwapBytes(&data_ptr[i], 4);
	return;
}

#ifdef __FP_OS__
/**********************************************************************
 * Function Name  : SwapBytes
 * Description    :
 * Inputs
 *   Parameters   : void *, size in bytes
 * Outputs        :
 *   Parameters   : void * - swaped bytes
 *   Returns      : -
 * Changes        :
 *********************************************************************/
void SwapBytes(void *pv, size_t n)
{
	int8_t *p = pv;
	size_t lo, hi;
	for (lo = 0, hi = n - 1; hi > lo; lo++, hi--) {
		int8_t tmp = p[lo];
		p[lo] = p[hi];
		p[hi] = tmp;
	}
}

#define SWAP(x) SwapBytes(&x, sizeof(x));
/**********************************************************************
 * Function Name  : convert_le2be_64
 * Description    : convert given data from little-endian to big-endian
 * Inputs           with 32b bit word swap
 *   Parameters   : char * , int
 * Outputs        :
 *   Parameters   : char *
 *   Returns      :
 * Changes        :
 *********************************************************************/
void
convert_le2be_64(int8_t *data, uint32_t size)
{
	int i, cnt;
	uint64_t  *data_ptr = (uint64_t *)data;

	cnt  = size >> 3;
	if (size & 0x7)
		cnt++;
	/* swap the lsb bytes as msb vice-versa */
	for (i = 0; i < cnt; i++)
		SwapBytes(&data_ptr[i], 8);
	return;
}

/**********************************************************************
 * Function Name  : convert_le2be
 * Description    : convert given data from little-endian to big-endian
 * Inputs
 *   Parameters   : char * , int
 * Outputs        :
 *   Parameters   : char *
 *   Returns      :
 * Changes        :
 *********************************************************************/
void
convert_le2be(int8_t *data, uint32_t size)
{
	int i, cnt;
	uint32_t *data_ptr = (uint32_t *)data;

	cnt  = size >> 2;
	if (size & 0x3)
		cnt++;
	/* swap the lsb bytes as msb vice-versa */
	for (i = 0; i < cnt; i++)
		SwapBytes(&data_ptr[i], 4);
	return;
}
#endif
/*****************************************************************
 * Function Name  : RegRead()
 * Description    : Hardware register read function return register
 *                  read value for given address.
 * Inputs
 *   Parameters   : UINT
 * Outputs        :
 *   Parameters   : -
 *   Returns      : UINT
 * Changes        :
 ****************************************************************/
inline uint32_t fp_reg_read( uint32_t addr )
{
	volatile uint32_t read_val;
	volatile uint32_t *reg_addr;

    if( ( fp_baseAddr == NULL ) || ( fp_baseAddr_bar4 == NULL ) )
    {
        FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
        return 0xFFFFFFFF;
    }

    if( ( addr >= 0x800000 ) && ( addr < 0x880000 ) )
    {
        addr -= 0x800000;
        reg_addr  = ( uint32_t * )( CBUS_ADDR_BASE + fp_baseAddr_bar4 + addr );
    }
    else if( ( addr < 0x800000 ) && ( addr > (BMU1_BASE_ADDR - 1) ) )
    {
        reg_addr  = ( uint32_t * )( CBUS_ADDR_BASE + fp_baseAddr + addr );
    }
    else
    {
        FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: read out of memory range\n", __func__);
        return 0xFFFFFFFF;
    }

	read_val  = *reg_addr;
    return read_val;
}
/******************************************************************
 * Function Name  : RegWrite()
 * Description    : Hardware register write function Write given value
 *                  in given address.
 * Inputs
 *   Parameters   : UINT, UINT
 * Outputs        :
 *   Parameters   : -
 *   Returns      : -
 * Changes        :
 *****************************************************************/
inline void fp_reg_write( uint32_t addr, uint32_t value )
{
    volatile uint32_t *reg_addr;

    if( ( fp_baseAddr == NULL ) || ( fp_baseAddr_bar4 == NULL ) )
    {
        FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: pci_mem base is NULL\n", __func__);
    }

    if( ( addr >= 0x800000 ) && ( addr < 0x880000 ) )
    {
        addr -= 0x800000;
        reg_addr  = ( uint32_t * )( CBUS_ADDR_BASE + fp_baseAddr_bar4 + addr );
        *reg_addr = (uint32_t) value;
    }
    else if( ( addr >= BMU1_BASE_ADDR ) && ( addr < 0x800000 ) )
    {
        reg_addr  = ( uint32_t * )( CBUS_ADDR_BASE + fp_baseAddr + addr );
        *reg_addr = (uint32_t) value;
    }
    else
    {
        FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: write out of memory range, addr 0x%x, val 0x%x\n", __func__, addr, value);
    }
}
