/*************************************************************************/ /*!
@File
@Title          fp_poolalloc.c
@Copyright      Copyright (c) Milandr. All Rights Reserved
@Description    bufferpool initialization and management (allocation/free)
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

/**********************************************************************
 *                                     Include Files
 **********************************************************************/
#include <fp_library.h>
#include <fp_types.h>
#include <fp_poolalloc.h>

#ifdef __FP_OS__
#include <linux/slab.h>
#include <linux/gfp.h>
#endif

#include "fp_pci_pe.h"

/**********************************************************************
 *                                      #defines                      *
 **********************************************************************/
#define MAGIC_NUM       0xdeadc0de

#ifdef __FP_OS__
#define POOLPANIC()      printk("\nPANIC: pool alloc error\n");
#endif

/**********************************************************************
 *                              Extern Declarations                   *
 **********************************************************************/
#ifdef __FP_OS__
extern POOLHEAD_PTR  gRtFlowPoolHead;
#else
extern POOLHEAD_PTR  gBridgePoolHead,gVlanPoolHead;
extern POOLHEAD  fp_brPoolHead,fp_vlanPoolHead;
#ifdef HW_SIM
POOLHEAD  rtPoolHead;
POOLHEAD_PTR pHead = FP_NULL;
#endif
#endif

/*******************************************************************************
 * Function name:       poolInit
 * Input Parameters:    Num, Size, Align
 * Output Parameters:   None
 * Result:              Pointer to PoolHead data structure
 * Description:
 *                      This function allocates a big chunk of memory. The
 *                      initial part is allocated for the PoolHead structre and
 *                      rest of the part is used for the linked list of the
 *                      buffers of size "Size" each. The starting location for
 *                      the buffer linked list is aligned at the "Align" byte
 *                      boundary. The Magic number is stored in the first 4
 *                      bytes of each buffer to detect the multiple free
 *                      operations on the same buffer. The value of "Align"
 *                      parameter should be multiple of 4.
 ******************************************************************************/


POOLHEAD_PTR
poolInit (UINT Num, UINT Size, int Align, UINT type)
{
    int iCount = 0;
    UINT iSize = 0;
#ifndef HW_SIM
    POOLHEAD_PTR pHead = FP_NULL;
#endif
    char *Ptr = FP_NULL;
    volatile POOLITEM_PTR Last = FP_NULL;
    volatile POOLITEM_PTR Tmp = FP_NULL;

#ifdef __FP_OS__
    /* allocate the pool head */
    pHead = kmalloc(sizeof (PoolHead_t), GFP_KERNEL);
    if (pHead == NULL) {
        FPATH_TRACE(FP_TRACE_INIT, FP_LOG_DEBUG,
                    ("%s: couldn't allocate pool head\n", __FUNCTION__));
        return NULL;
    }

    iSize = Size * Num;

#ifdef _FEATURE_RTP_RELAY_
    if(type == RTP_SOCK_ENTRIES_TABLE)
    {
       Ptr = kmalloc(iSize,GFP_KERNEL);
    }
    else
#endif
    {
        Ptr = fp_pci_malloc(iSize, type);
    }
#else /*Else of _FP_OS_*/
#if ((defined (FP_SWITCH_MODE)) || (defined (FP_RT_SW_MODE)))

 if(type == VLAN_ENTRIES_TABLE)  {
#ifdef DEBUG_HW
    FP_DEBUG(DBGP_FEAT_LIB, FP_LOG_INFO, "J\n");
#endif
     pHead = &fp_vlanPoolHead;
 } else if(type == BR_ENTRIES_TABLE){
     pHead = &fp_brPoolHead;
 }
#endif /*End of FP_SWITCH_MODE and FP_SW_RT_MODE*/

#if ((defined (HW_SIM)) && (defined (FP_ROUTE_MODE)))
     pHead = &rtPoolHead;
#endif /*End of HW_SIM  and FP_ROUTE_MODE*/

    for(j=1 ; j <= Num; j=j+1)  /* As per PE coding guide lines removing '*' */
        iSize = iSize + Size;
    Ptr = fp_pci_malloc(iSize, type);
#endif /*End of _FP_OS_*/

    FPATH_TRACE(FP_TRACE_INIT, FP_LOG_DEBUG,
                ("%s: pool buffer ptr = %p\n", __FUNCTION__, Ptr));

    if (Ptr == FP_NULL) {
        FPATH_TRACE(FP_TRACE_INIT, FP_LOG_DEBUG,
                    ("%s: error allocating pool buffer", __FUNCTION__));
        return FP_NULL;
    }

    /* pool head buffer */
    pHead->Front = Ptr;
    pHead->Start = Ptr;
    iCount = 0;
    /* prepare a linked list of elements of size "Size". */
    for ( Tmp = (POOLITEM_PTR)Ptr; iCount < Num; iCount++) {
        Last = Tmp;

        if (Tmp == FP_NULL) {
            FPATH_TRACE(FP_TRACE_INIT, FP_LOG_ERR, ("%s: PANIC", __FUNCTION__));
            break;
        }
        /*To avoid route fetch pe crash -added res field in poolitem*/
        Tmp->space = 0;
        Tmp->res = 0;
        Tmp->MagicNum = MAGIC_NUM;
#ifdef DEBUG_SIM
        DbgRegWrite(Tmp,peid);
        DbgRegWrite(Tmp->MagicNum,peid);
#endif
        //FP_DEBUG(DBGP_FEAT_LIB, FP_LOG_INFO, "MAGIC %x tmp:%x\n",Tmp->MagicNum,PCI_TO_PE((unsigned int)Tmp));
        Tmp->Ptr = (((char*)Tmp) + Size);
#ifdef HW_SIM
        Tmp = (PoolItem_t *)Tmp->Ptr;
#else
        Tmp = (POOLITEM_PTR)((POOLITEM_PTR)Tmp)->Ptr;
#endif


    }/*End of pool memory intialisation FOR-loop*/

    Last->Ptr = FP_NULL;
    pHead->Rear = (char*)Last;
    pHead->Free = Num;
    pHead->PoolSize = Num;
    FPATH_TRACE(FP_TRACE_INIT, FP_LOG_DEBUG,
                ("%s: free = %d, pool size = %d\n",
                 __FUNCTION__, pHead->Free, pHead->PoolSize));

    return pHead;
} /* poolInit() */

/*******************************************************************************
 * Function name:       poolAlloc
 * Input Parameters:    Pointer to PoolHead data structure
 * Output Parameters:   None
 * Result:              Pointer to the allocated buffer
 * Description:
 *                      This function allocates a buffer from the free buffer
 *                      pool. For major errors it does the Panic operation as
 *                      defined by the macro POOLPANIC()
 ******************************************************************************/
char *
poolAlloc (POOLHEAD_PTR pHead)
{
    char *New = FP_NULL;
#ifdef HW_SIM
    PoolItem_t *Tmp = FP_NULL;
#else
    POOLITEM_PTR Tmp =FP_NULL;
#endif
    if (pHead == FP_NULL) {
#ifndef __FP_OS__
#ifdef DEBUG_HW
        FP_DEBUG(DBGP_FEAT_LIB, FP_LOG_EMERG, "PHN\n");
#endif
#endif
        return FP_NULL;
    }
    /* check to see if pool is empty */
    if (pHead->Free == 0) {
#ifndef __FP_OS__
#ifdef DEBUG_HW
       FP_DEBUG(DBGP_FEAT_LIB, FP_LOG_EMERG, "PR\n");
#endif
#endif
       return FP_NULL;
    }

    if (pHead->Front != FP_NULL) {
        New = pHead->Front;
        /* clear off the MAGIC number before allocating */
#ifdef HW_SIM
       Tmp = (PoolItem_t *)New;
#ifdef DEBUG_SIM
       DbgRegWrite(0x5555,peid);
       DbgRegWrite(Tmp->space,peid);
       DbgRegWrite(Tmp->res,peid);
       DbgRegWrite(Tmp->MagicNum,peid);
       DbgRegWrite(Tmp->Ptr,peid);
       DbgRegWrite(0x6666,peid);
       DbgRegWrite(New,peid);
       DbgRegWrite(Tmp->MagicNum,peid);
#endif
       if (Tmp->MagicNum != MAGIC_NUM) {
#ifdef DEBUG_SIM
            DbgRegWrite(0xDEAD,peid);
#endif
            return FP_NULL;
       }
#else  /*Else of HW_SIM*/
        Tmp = (POOLITEM_PTR)New;
        if(((POOLITEM_PTR)Tmp)->MagicNum != MAGIC_NUM) {
            FPATH_TRACE(FP_TRACE_INIT, FP_LOG_ERR,
                        ("%s: trying to allocate an allocated buffer\n",
                         __FUNCTION__));
            return FP_NULL;
        }

#endif /*End of HW_SIM*/
        Tmp->space = 0;
        Tmp->res = 0;
        Tmp->MagicNum = 0;
        pHead->Front = Tmp->Ptr;
        Tmp->Ptr = FP_NULL;
        pHead->Free--;
    } else {  /* Should never come here */
        FPATH_TRACE(FP_TRACE_INIT, FP_LOG_ERR,
                    ("%s: error while allocating the pool element\n",
                     __FUNCTION__));
        return FP_NULL;
    }
    return New;
}

/*******************************************************************************
 * Function name:       pooFree
 * Input Parameters:    ptr to the buffer, ptr to poolhead
 * Output Parameters:   None
 * Result:              None
 * Description:
 *                      This function frees the given buffer. It basically puts
 *                      back the buffer into the free buffer pool.
 *                      For major errors it does the Panic operation as defined
 *                      by the macro POOLPANIC()
 ******************************************************************************/
int
poolFree (char *pDel, POOLHEAD_PTR pHead)
{
    if (pHead == FP_NULL) {
        FPATH_TRACE(FP_TRACE_OTHER, FP_LOG_ERR, ("%s: PANIC\n", __FUNCTION__));
        return -1;
    }

    /* check to see if pool is full */
    if (pHead->Free == pHead->PoolSize) {
        FPATH_TRACE(FP_TRACE_CLEANUP, FP_LOG_ERR,
                    ("%s: buffer pool full\n", __FUNCTION__));
        return -1;
    }

    if (pDel == FP_NULL) {
        FPATH_TRACE(FP_TRACE_CLEANUP, FP_LOG_WARNING,
                    ("%s: trying to free a NULL ptr\n", __FUNCTION__));
        return -1;
    }

    if (((POOLITEM_PTR)pDel)->MagicNum == MAGIC_NUM) {
        FPATH_TRACE(FP_TRACE_CLEANUP, FP_LOG_ERR,
                    ("%s: trying to free a freed buffer\n", __FUNCTION__));
        return -1;
    }

    if (pHead->Free == 0) {
        pHead->Front = pDel;
    } else {
        ((POOLITEM_PTR)pHead->Rear)->Ptr = pDel;
    }

    pHead->Rear = pDel;
    /* put the MAGIC number back */
    ((POOLITEM_PTR)pHead->Rear)->space = 0;
    ((POOLITEM_PTR)pHead->Rear)->res   = 0;
    ((POOLITEM_PTR)pHead->Rear)->MagicNum = MAGIC_NUM;
    ((POOLITEM_PTR)pHead->Rear)->Ptr = FP_NULL;
    pHead->Free++;

#ifndef __FP_OS__
   #ifdef DEBUG_HW
    FP_DEBUG(DBGP_FEAT_LIB, FP_LOG_INFO, "PF\n");
   #endif
#else
    //FP_DEBUG(DBGP_FEAT_LIB, FP_LOG_INFO, "FR:%x size:%x,delp:%x\n", pHead->Free,pHead->PoolSize,pDel);
#endif
    return 0;
}

/*******************************************************************************
 * Function name:       poolCount
 * Input Parameters:    Pointer to PoolHead data structure
 * Output Parameters:   None
 * Result:              The number of buffers left in the pool
 * Description:
 *                      This function returns a count of the number of buffers
 *                      in the pool.
 ******************************************************************************/
UINT
poolCount (POOLHEAD_PTR pHead)
{
    FPATH_TRACE(FP_TRACE_INIT, FP_LOG_DEBUG,
                ("total available memory = %d\n", pHead->Free));
    FPATH_TRACE(FP_TRACE_INIT, FP_LOG_DEBUG,
                ("total pool memory = %d\n", pHead->PoolSize));
    return 0;
}

/*******************************************************************************
 * Function name:       poolDelete
 * Input Parameters:    Pointer to PoolHead data structure
 * Output Parameters:   None
 * Result:              success or failure
 * Description:
 *                      This function frees the entire pool
 ******************************************************************************/
UINT
poolDelete (POOLHEAD_PTR pHead, UINT type)
{
#ifdef _FP_OS__
    /* will need to add back when we have multiple tables */
    fp_pci_free((char *)pHead->Start, type);
    kfree(pHead);
#endif
    return 0;
}
