#include "phy_dp83867.h"
#include "fp_library.h"
#include <linux/ethtool.h>

/* ----------------------------------- Настройка микросхем PHY ---------------------------------- */
void phy_dp83867_init( uint8_t ucMdioPort, uint8_t ucMdioAdr, uint16_t speed )
{
    uint32_t tmp;
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CTRL, DP83867_SW_RESET );   // Software reset

    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_PHYCR, DP83867_PHYCR_FIFO_DEPTH_4_B_NIB << DP83867_PHYCR_FIFO_DEPTH_SHIFT | 1 << 11 );

    /* --- Задание режима работы светодиодов --- */
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_LEDCR1,
                                        ( DP83867_LED_TX_ACT << DP83867_LED0_POS ) |
                                        ( DP83867_LED_RX_ACT << DP83867_LED1_POS ) );    

    tmp = fp_mdio_read_phy(ucMdioPort, ucMdioAdr, DP83867_PHYCR);
    tmp |= 3 << 5;  //Enable automatic crossover (Auto-MDIX)
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_PHYCR, tmp); 

    tmp = phy_dp83867_mmd_read( ucMdioPort, ucMdioAdr, DP83867_CFG4 );
    tmp &= ~( 1 << 7 ); //disable advertised EEE
    phy_dp83867_mmd_write( ucMdioPort, ucMdioAdr, DP83867_CFG4, tmp );

    phy_dp83867_speed_set( ucMdioPort, ucMdioAdr, 0, 1, 1 );
    //phy_dp83867_restart_aneg( ucMdioPort, ucMdioAdr );

    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "init phy_dp83867 end\n");

    //phy_dp83867_sgmii_loopback_enable(ucMdioPort, ucMdioAdr);
}



/*
Set Auto-Negotiation flow control parameters
params:
    ucMdioPort  - mdio port (MDIO_PORT_1 or MDIO_PORT_2)
    ucMdioAdr   - ethernet phy address at mdio bus
    param       - flow control parameters (0, 1, 2, 3) for reg_addr 0x0004 (Auto-Negotiation Advertisement Register)
                  0 - flow control not advertisement 
                  1 - pause bit Advertisement
                  2 - asm_dir bit advertisement
                  3 - pause and asm_dir bits sets
*/
uint32_t phy_dp83867_set_fc_parameters( uint8_t ucMdioPort, uint8_t ucMdioAdr, uint32_t param){
    uint32_t tmp;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR );

    tmp &= ~(3 << 10);
    param &= 3;
    tmp |= (param << 10);

    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR, tmp );
    return 0;
}

/*
Get Auto-Negotiation Advertisement value from register Auto-Negotiation Advertisement Register (0x0004)
return values:
    0 - flow control not advertisement
    1 - pause flow control advertisement
    2 - asymmectrical pause flow control advertisement (only generate pause frames)
    3 - both pause and asymmectrical pause flow control advertisement
*/
uint32_t phy_dp83867_is_flow_control_advertised( uint8_t ucMdioPort, uint8_t ucMdioAdr ){
    uint32_t tmp = 0;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR );

    return ( ( tmp & ( 3 << 10 ) ) >> 10 );
}

/*
Get Auto-Negotiation Advertisement value from register Auto-Negotiation Link Partner Ability Register (0x0005)
return values:
    0 - flow control not supported by Link Partner
    1 - pause flow control supported  by Link Partner
    2 - asymmectrical pause flow control supported by Link Partner (only generate pause frames)
    3 - both pause and asymmectrical pause flow control by Link Partner 
*/
uint32_t phy_dp83867_is_flow_control_link_part( uint8_t ucMdioPort, uint8_t ucMdioAdr ){
    uint32_t tmp = 0;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANLPAR );

    return ( ( tmp & ( 3 << 10 ) ) >> 10 );
}

/*
Get Auto-Negotiation duplex resolved value from register PHY Status Register (0x0011)
return values:
    0 - Auto-Negotiation enabled and has not completed
    1 - Auto-Negotiation is disabled, setup half duplex
    2 - Auto-Negotiation is disabled, setup full duplex
    3 - Auto-Negotiation enabled, half duplex
    4 - Auto-Negotiation enabled, full duplex
*/
uint32_t phy_dp83867_get_duplex( uint8_t ucMdioPort, uint8_t ucMdioAdr ){
    
    uint32_t tmp = 0;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );

    if( !( tmp & (1 << 12) ) ){ //Auto-Negotiation is disabled
        if( tmp & (1 << 8) ){   //setup full duplex
            return 2;
        }else{  //setup half duplex
            return 1;
        }
    }else{
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_PHYSTS );
        
        if( !( tmp & (1 << 11) ) ){ //Auto-Negotiation enabled and has not completed
            return 0;
        }

        if( !( tmp & (1 << 13) ) ){ //Half duplex
            return 3;
        } 

        //full duplex
        return 4;
    }
}

/*
Check settings for flow control
return values:
    0 - flow control not supported by Local Host or Link Partner
    1 - pause flow control supported by Local Host and Link Partner
    2 - asymmectrical pause flow control supported by Local Host (tx only)
    3 - asymmectrical pause flow control supported by Local Host (rx only)
*/
uint32_t phy_dp83867_is_flow_control_available( uint8_t ucMdioPort, uint8_t ucMdioAdr ){

    uint32_t tmp, local_advertised, remote_advertised;

    tmp = phy_dp83867_get_duplex( ucMdioPort, ucMdioAdr );
    if( (tmp == 0) || (tmp  == 1) || (tmp == 3) ){ //flow control (PAUSE) available only in full duplex mode of Ethernet PHY
        return 0;
    }

    local_advertised = phy_dp83867_is_flow_control_advertised( ucMdioPort, ucMdioAdr );
    remote_advertised = phy_dp83867_is_flow_control_link_part( ucMdioPort, ucMdioAdr );

    switch(local_advertised){
        case 0:
            return 0;       //flow control not available

        case 1:
            if(remote_advertised & 1){
                return 1;   //pause flow control abailable (tx & rx)
            }else{
                return 0;
            }
            break;

        case 2:
            if( (remote_advertised & 3) == 3 ){
                return 2;   //asymmectrical pause flow control supported by Local Host (tx only)
            }else{
                return 0;   //flow control not available
            }
            break;

        case 3:
            if( !(remote_advertised) ){
                return 0;   //flow control not available
            }
            if( remote_advertised & 1 ){
                return 1;   //pause flow control abailable (tx & rx)
            }else{
                return 3;   //asymmectrical pause flow control supported by Local Host (rx only)
            }
            break;

        default:
            FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "%s: should never come here!\n", __func__);
            return 0;
            break;
    }
}


void phy_dp83867_speed_set( uint8_t ucMdioPort, uint8_t ucMdioAdr, uint16_t speed, uint8_t duplex, uint8_t autoneg )
{
    u32 tmp, bmcr;

    bmcr = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );
    if( autoneg ){
        bmcr |= ( 1 << 12 );    //autoneg enable 
    }else{
        bmcr &= ~( 1 << 12 );   //autoneg disable
    }

    switch ( speed ){
    case 10:
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR );
        tmp &= 0x7E1F;
        bmcr &= ~( ( 1 << 6 )|( 1 << 13 ) );    //speed 10 Mbit/s
        if( duplex ){
            //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0x0100 );
            bmcr |= 0x100;          //full duplex
            tmp |= (1 << 6);
        }else{
            //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0x0000 );
            bmcr &= ~( 1 << 8 );    //half duplex
            tmp |= (1 << 5);
        }
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, bmcr );
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR, tmp );
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1 );
        tmp &= ~( ( 1 << 9 )|( 1 << 8 ) );	                                       // 1000BASE-T Full-Duplex Not Advertise, 1000BASE-T Half-Duplex Not Advirtes
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1, tmp );
        break;
    case 100:
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR );
        tmp &= 0x7E1F;
        /*set speed 100 Mbit/s */
        bmcr &= ~( 1 << 6 );
        bmcr |= ( 1 << 13 );
        if( duplex ){
            //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0xA100 );
            bmcr |= 0x100;          //full duplex
            tmp |= ( 1 << 8 );
        }else{
            //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0xA000 );
            bmcr &= ~( 1 << 8 );    //half duplex
            tmp |= ( 1 << 7 );
        }
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, bmcr );
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR, tmp );
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1 );
        tmp &= ~( ( 1 << 9 )|( 1 << 8 ) );	                                       // 1000BASE-T Full-Duplex Not Advertise, 1000BASE-T Half-Duplex Not Advirtes
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1, tmp );
        break;
    case 1000:
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1 );
        tmp &= ~( ( 1 << 9 )|( 1 << 8 ) );
        /* set speed 1000 Mbit/s */
        bmcr &= ~( 1 << 13 );
        bmcr |= ( 1 << 6 );
        if( duplex ){
            //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0x0140 );
            bmcr |= ( 1 << 8 );          //full duplex
            tmp |= ( 1 << 9 );
        }else{
            //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0x0040 );
            bmcr &= ~( 1 << 8 );    //half duplex
            tmp |= ( 1 << 8 );
        }
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, bmcr );
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR, 0x0001 );
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1, tmp );
        break;
    default:                                                                   // Autoneg on
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, 0x1140 );
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR );
        tmp |= ( 1 << 8 )|( 1 << 7 )|( 1 << 6 )|( 1 << 5 )|( 1 << 0 );
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR, tmp );
        tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1 );
        tmp |= 3 << 8;
        fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1, tmp );
        break;
    }
    //tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );
    //fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, tmp | 0x8000 );

    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CTRL, 0x4000 );   
}

void phy_dp83867_restart_aneg( uint8_t ucMdioPort, uint8_t ucMdioAdr )
{
    uint32_t tmp;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, tmp | 0x0200 );
}

uint32_t phy_dp83867_get_link_part_abil( uint8_t ucMdioPort, uint8_t ucMdioAdr )
{
    uint32_t advertising = 0, tmp;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANLPAR );

    if( tmp & ( 1 << 5 ) ){
        advertising |= ADVERTISED_10baseT_Half;
    }
    if( tmp & ( 1 << 6 ) ){
        advertising |= ADVERTISED_10baseT_Full;
    }
    if( tmp & ( 1 << 7 ) ){
        advertising |= ADVERTISED_100baseT_Half;
    }
    if( tmp & ( 1 << 8 ) ){
        advertising |= ADVERTISED_100baseT_Full;
    }
    if( tmp & ( 1 << 10 ) ){
        advertising |= ADVERTISED_Pause;
    }
    if( tmp & ( 1 << 11 ) ){
        advertising |= ADVERTISED_Asym_Pause;
    }

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_STS1 );

    if( tmp & ( 1 << 10 ) )
        advertising |= ADVERTISED_1000baseT_Half;
    if( tmp & ( 1 << 11 ) )
        advertising |= ADVERTISED_1000baseT_Full;

    return advertising;
}

uint32_t phy_dp83867_get_mdi_stat( uint8_t ucMdioPort, uint8_t ucMdioAdr )
{
    uint32_t tmp;

    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_PHYSTS );

    if( tmp & ( 1 << 8 ) || tmp & ( 1 << 9 ) )                                  // ??? 8, 9 bit
        return ETH_TP_MDI_X;
    else
        return ETH_TP_MDI;
}

void phy_dp83867_led_drive( uint8_t ucMdioPort, uint8_t ucMdioAdr, uint8_t ucState )
{
    switch( ucState )
    {
        case DP83867_LEDS_ON:
            fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_LEDCR2, 0x4477 );
        break;
        case DP83867_LEDS_OFF:
            fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_LEDCR2, 0x4455 );
        break;
        default:                                                                // Leds control by events
            fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_LEDCR2, 0x4444 );
        break;
    }
}

uint32_t phy_dp83867_mmd_read(uint8_t ucMdioPort, uint8_t ucMdioAdr, uint16_t reg_addr){

    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_REGCR, 0x001F);
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_ADDAR, reg_addr);
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_REGCR, 0x401F);

    return ( fp_mdio_read_phy(ucMdioPort, ucMdioAdr, DP83867_ADDAR) & 0xFFFF );
}


void  phy_dp83867_mmd_write(uint8_t ucMdioPort, uint8_t ucMdioAdr, uint16_t reg_addr, uint16_t write_value){
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_REGCR, 0x001F);
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_ADDAR, reg_addr);
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_REGCR, 0x401F);
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_ADDAR, write_value);
}

void phy_dp83867_sgmii_loopback_enable( uint8_t ucMdioPort, uint8_t ucMdioAdr){
    uint16_t tmp;

    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "Start init of sgmii loopback of DP83867\n");
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_BISCR, 0x0001);    //pcs loopback on
    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "wr DP83867_BISCR 0001\n");

    phy_dp83867_mmd_write(ucMdioPort, ucMdioAdr, DP83867_LOOPCR, 0xE720);
    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "wr DP83867_LOOPCR 0001\n");

    tmp = fp_mdio_read_phy(ucMdioPort, ucMdioAdr, DP83867_BMCR);
    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "rd from DP83867_BMCR %04X\n", tmp);

    tmp &= ~((1 << 13)|(1 << 12));  //aneg disable, 1G
    tmp |= (1 << 14)|(1 << 8)|(1 << 6); //loopback enable, full duplex, 1G
    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_BMCR, tmp);    //
    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "wr DP83867_BMCR %04X\n", tmp);

    fp_mdio_write_phy(ucMdioPort, ucMdioAdr, DP83867_CTRL, (1 << 14) );    //sw_restart
    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "wr DP83867_CTRL 0x4000\n");

    FP_DEBUG(DBGP_FEAT_HIF, FP_LOG_INFO, "init pcs loopback of dp83867 end\n");
}

void phy_dp83867_set_power_down( uint8_t ucMdioPort, uint8_t ucMdioAdr ){
    uint16_t tmp;
    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );
    tmp |= ( 1 << 11 );
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, tmp );
}

void phy_dp83867_set_power_up( uint8_t ucMdioPort, uint8_t ucMdioAdr ){
    uint16_t tmp;
    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );
    tmp &= ~( 1 << 11 );
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, tmp );
}

void phy_dp83867_set_advertise_speed_duplex( uint8_t ucMdioPort, uint8_t ucMdioAdr, uint32_t advertise ){
    uint16_t adv;
    adv = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR );
    adv &= ~( 0xF << 5 );
    adv |=  ( ( advertise & ADVERTISED_10baseT_Half )  << 5 )|( ( advertise & ADVERTISED_10baseT_Full )  << 5 )|
            ( ( advertise & ADVERTISED_100baseT_Half ) << 5 )|( ( advertise & ADVERTISED_100baseT_Full ) << 5 );
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_ANAR, adv );

    adv = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1 );
    adv &= ~( 3 << 8 );
    adv |= ( ( advertise & ADVERTISED_1000baseT_Half ) << 4 )|( ( advertise & ADVERTISED_1000baseT_Full ) << 4 );
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_CFG1, adv );
}

void phy_dp83867_autoneg_enable( uint8_t ucMdioPort, uint8_t ucMdioAdr ){
    uint16_t tmp;
    tmp = fp_mdio_read_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR );
    tmp |= ( 1 << 12 );
    fp_mdio_write_phy( ucMdioPort, ucMdioAdr, DP83867_BMCR, tmp );
}
